@@ -17,22 +17,23 @@
  * Rspamd fuzzy storage server
  */
 
-#include <libserver/rspamd_control.h>
-#include <src/libutil/map_private.h>
 #include "config.h"
 #include "util.h"
 #include "rspamd.h"
 #include "map.h"
+#include "map_helpers.h"
 #include "fuzzy_wire.h"
 #include "fuzzy_backend.h"
 #include "ottery.h"
+#include "ref.h"
+#include "xxhash.h"
 #include "libserver/worker_util.h"
 #include "libserver/rspamd_control.h"
 #include "libcryptobox/cryptobox.h"
 #include "libcryptobox/keypairs_cache.h"
 #include "libcryptobox/keypair.h"
-#include "ref.h"
-#include "xxhash.h"
+#include "libserver/rspamd_control.h"
+#include "libutil/map_private.h"
 #include "libutil/hash.h"
 #include "libutil/http_private.h"
 #include "unix-std.h"
@@ -58,10 +59,6 @@ static const gchar *local_db_name = "local";
         session->name, session->uid, \
         G_STRFUNC, \
         __VA_ARGS__)
-#define msg_debug_fuzzy_update(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        session->name, session->uid, \
-        G_STRFUNC, \
-        __VA_ARGS__)
 #define msg_err_fuzzy_collection(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
         "fuzzy_collection", session->uid, \
         G_STRFUNC, \
@@ -74,10 +71,6 @@ static const gchar *local_db_name = "local";
        "fuzzy_collection", session->uid, \
         G_STRFUNC, \
         __VA_ARGS__)
-#define msg_debug_fuzzy_collection(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        "fuzzy_collection", session->uid, \
-        G_STRFUNC, \
-        __VA_ARGS__)
 
 /* Init functions */
 gpointer init_fuzzy (struct rspamd_config *cfg);
@@ -136,8 +129,10 @@ struct rspamd_fuzzy_storage_ctx {
 	char *hashfile;
 	gdouble expire;
 	gdouble sync_timeout;
-	radix_compressed_t *update_ips;
-	radix_compressed_t *master_ips;
+	struct rspamd_radix_map_helper *update_ips;
+	struct rspamd_radix_map_helper *master_ips;
+	struct rspamd_radix_map_helper *blocked_ips;
+
 	struct rspamd_cryptobox_keypair *sync_keypair;
 	struct rspamd_cryptobox_pubkey *master_key;
 	struct timeval master_io_tv;
@@ -145,6 +140,8 @@ struct rspamd_fuzzy_storage_ctx {
 	GPtrArray *mirrors;
 	const ucl_object_t *update_map;
 	const ucl_object_t *masters_map;
+	const ucl_object_t *blocked_map;
+
 	GHashTable *master_flags;
 	guint keypair_cache_size;
 	gint peer_fd;
@@ -163,7 +160,7 @@ struct rspamd_fuzzy_storage_ctx {
 	struct rspamd_keypair_cache *keypair_cache;
 	rspamd_lru_hash_t *errors_ips;
 	struct rspamd_fuzzy_backend *backend;
-	GQueue *updates_pending;
+	GArray *updates_pending;
 	guint updates_failed;
 	guint updates_maxfail;
 	guint32 collection_id;
@@ -233,19 +230,31 @@ struct fuzzy_master_update_session {
 static void rspamd_fuzzy_write_reply (struct fuzzy_session *session);
 
 static gboolean
-rspamd_fuzzy_check_client (struct fuzzy_session *session)
+rspamd_fuzzy_check_client (struct fuzzy_session *session, gboolean is_write)
 {
-	if (session->ctx->update_ips != NULL) {
-		if (radix_find_compressed_addr (session->ctx->update_ips,
-				session->addr) == RADIX_NO_VALUE) {
+	if (session->ctx->blocked_ips != NULL) {
+		if (rspamd_match_radix_map_addr (session->ctx->blocked_ips,
+				session->addr) != NULL) {
 			return FALSE;
 		}
-		else {
-			return TRUE;
-		}
 	}
 
-	return FALSE;
+	if (is_write) {
+		if (session->ctx->update_ips != NULL) {
+			if (rspamd_match_radix_map_addr (session->ctx->update_ips,
+					session->addr) == NULL) {
+				return FALSE;
+			}
+			else {
+				return TRUE;
+			}
+		}
+
+		return FALSE;
+	}
+
+	/* Non write */
+	return TRUE;
 }
 
 static void
@@ -296,7 +305,7 @@ fuzzy_mirror_close_connection (struct fuzzy_slave_connection *conn)
 
 		close (conn->sock);
 
-		g_slice_free1 (sizeof (*conn), conn);
+		g_free (conn);
 	}
 }
 
@@ -311,7 +320,6 @@ static void
 fuzzy_mirror_updates_version_cb (guint64 rev64, void *ud)
 {
 	struct rspamd_fuzzy_updates_cbdata *cbdata = ud;
-	GList *cur;
 	struct fuzzy_peer_cmd *io_cmd;
 	guint32 rev32 = rev64, len;
 	const gchar *p;
@@ -321,17 +329,18 @@ fuzzy_mirror_updates_version_cb (guint64 rev64, void *ud)
 	struct rspamd_http_message *msg;
 	struct rspamd_fuzzy_mirror *m;
 	struct timeval tv;
+	guint i;
 
 	conn = cbdata->conn;
 	ctx = cbdata->ctx;
 	msg = cbdata->msg;
 	m = cbdata->m;
-	g_slice_free1 (sizeof (*cbdata), cbdata);
+	g_free (cbdata);
 	rev32 = GUINT32_TO_LE (rev32);
 	len = sizeof (guint32) * 2; /* revision + last chunk */
 
-	for (cur = ctx->updates_pending->head; cur != NULL; cur = g_list_next (cur)) {
-		io_cmd = cur->data;
+	for (i = 0; i < ctx->updates_pending->len; i ++) {
+		io_cmd = &g_array_index (ctx->updates_pending, struct fuzzy_peer_cmd, i);
 
 		if (io_cmd->is_shingle) {
 			len += sizeof (guint32) + sizeof (guint32) +
@@ -347,8 +356,8 @@ fuzzy_mirror_updates_version_cb (guint64 rev64, void *ud)
 	reply = rspamd_fstring_append (reply, (const char *)&rev32,
 			sizeof (rev32));
 
-	for (cur = ctx->updates_pending->head; cur != NULL; cur = g_list_next (cur)) {
-		io_cmd = cur->data;
+	for (i = 0; i < ctx->updates_pending->len; i ++) {
+		io_cmd = &g_array_index (ctx->updates_pending, struct fuzzy_peer_cmd, i);
 
 		if (io_cmd->is_shingle) {
 			len = sizeof (guint32) +
@@ -386,7 +395,7 @@ fuzzy_mirror_updates_to_http (struct rspamd_fuzzy_mirror *m,
 
 	struct rspamd_fuzzy_updates_cbdata *cbdata;
 
-	cbdata = g_slice_alloc (sizeof (*cbdata));
+	cbdata = g_malloc (sizeof (*cbdata));
 	cbdata->ctx = ctx;
 	cbdata->msg = msg;
 	cbdata->conn = conn;
@@ -427,7 +436,7 @@ rspamd_fuzzy_send_update_mirror (struct rspamd_fuzzy_storage_ctx *ctx,
 	struct fuzzy_slave_connection *conn;
 	struct rspamd_http_message *msg;
 
-	conn = g_slice_alloc0 (sizeof (*conn));
+	conn = g_malloc0 (sizeof (*conn));
 	conn->up = rspamd_upstream_get (m->u,
 			RSPAMD_UPSTREAM_MASTER_SLAVE, NULL, 0);
 	conn->mirror = m;
@@ -503,8 +512,6 @@ rspamd_fuzzy_updates_cb (gboolean success, void *ud)
 	guint i;
 	struct rspamd_fuzzy_storage_ctx *ctx;
 	const gchar *source;
-	GList *cur;
-	struct fuzzy_peer_cmd *io_cmd;
 
 	ctx = cbdata->ctx;
 	source = cbdata->source;
@@ -512,7 +519,7 @@ rspamd_fuzzy_updates_cb (gboolean success, void *ud)
 	if (success) {
 		rspamd_fuzzy_backend_count (ctx->backend, fuzzy_count_callback, ctx);
 
-		if (g_queue_get_length (ctx->updates_pending) > 0) {
+		if (ctx->updates_pending->len > 0) {
 			for (i = 0; i < ctx->mirrors->len; i ++) {
 				m = g_ptr_array_index (ctx->mirrors, i);
 
@@ -521,15 +528,7 @@ rspamd_fuzzy_updates_cb (gboolean success, void *ud)
 		}
 
 		/* Clear updates */
-		cur = ctx->updates_pending->head;
-
-		while (cur) {
-			io_cmd = cur->data;
-			g_slice_free1 (sizeof (*io_cmd), io_cmd);
-			cur = g_list_next (cur);
-		}
-
-		g_queue_clear (ctx->updates_pending);
+		ctx->updates_pending->len = 0;
 		rspamd_fuzzy_backend_version (ctx->backend, source,
 				fuzzy_update_version_callback, g_strdup (source));
 		ctx->updates_failed = 0;
@@ -538,23 +537,15 @@ rspamd_fuzzy_updates_cb (gboolean success, void *ud)
 		if (++ctx->updates_failed > ctx->updates_maxfail) {
 			msg_err ("cannot commit update transaction to fuzzy backend, discard "
 					"%ud updates after %d retries",
-					g_queue_get_length (ctx->updates_pending),
+					ctx->updates_pending->len,
 					ctx->updates_maxfail);
 			ctx->updates_failed = 0;
-			cur = ctx->updates_pending->head;
-
-			while (cur) {
-				io_cmd = cur->data;
-				g_slice_free1 (sizeof (*io_cmd), io_cmd);
-				cur = g_list_next (cur);
-			}
-
-			g_queue_clear (ctx->updates_pending);
+			ctx->updates_pending->len = 0;
 		}
 		else {
 			msg_err ("cannot commit update transaction to fuzzy backend, "
 					"%ud updates are still pending, %d updates left",
-					g_queue_get_length (ctx->updates_pending),
+					ctx->updates_pending->len,
 					ctx->updates_maxfail - ctx->updates_failed);
 		}
 	}
@@ -570,7 +561,7 @@ rspamd_fuzzy_updates_cb (gboolean success, void *ud)
 	}
 
 	g_free (cbdata->source);
-	g_slice_free1 (sizeof (*cbdata), cbdata);
+	g_free (cbdata);
 }
 
 static void
@@ -580,9 +571,8 @@ rspamd_fuzzy_process_updates_queue (struct rspamd_fuzzy_storage_ctx *ctx,
 
 	struct rspamd_updates_cbdata *cbdata;
 
-	if (ctx->updates_pending &&
-			(forced || g_queue_get_length (ctx->updates_pending) > 0)) {
-		cbdata = g_slice_alloc (sizeof (*cbdata));
+	if ((forced ||ctx->updates_pending->len > 0)) {
+		cbdata = g_malloc (sizeof (*cbdata));
 		cbdata->ctx = ctx;
 		cbdata->source = g_strdup (source);
 		rspamd_fuzzy_backend_process_updates (ctx->backend, ctx->updates_pending,
@@ -610,11 +600,23 @@ rspamd_fuzzy_write_reply (struct fuzzy_session *session)
 				session->cmd_type == CMD_ENCRYPTED_SHINGLE) {
 		/* Encrypted reply */
 		data = &session->reply;
-		len = sizeof (session->reply);
+
+		if (session->epoch > RSPAMD_FUZZY_EPOCH10) {
+			len = sizeof (session->reply);
+		}
+		else {
+			len = sizeof (session->reply.hdr) + sizeof (session->reply.rep.v1);
+		}
 	}
 	else {
 		data = &session->reply.rep;
-		len = sizeof (session->reply.rep);
+
+		if (session->epoch > RSPAMD_FUZZY_EPOCH10) {
+			len = sizeof (session->reply.rep);
+		}
+		else {
+			len = sizeof (session->reply.rep.v1);
+		}
 	}
 
 	r = rspamd_inet_address_sendto (session->fd, data, len, 0,
@@ -648,7 +650,7 @@ fuzzy_peer_send_io (gint fd, gshort what, gpointer d)
 	}
 
 	event_del (&up_req->io_ev);
-	g_slice_free1 (sizeof (*up_req), up_req);
+	g_free (up_req);
 }
 
 static void
@@ -718,26 +720,40 @@ rspamd_fuzzy_make_reply (struct rspamd_fuzzy_cmd *cmd,
 		struct fuzzy_session *session,
 		gboolean encrypted, gboolean is_shingle)
 {
-	if (cmd) {
-		result->tag = cmd->tag;
+	gsize len;
 
+	if (cmd) {
+		result->v1.tag = cmd->tag;
 		memcpy (&session->reply.rep, result, sizeof (*result));
 
 		rspamd_fuzzy_update_stats (session->ctx,
 				session->epoch,
-				result->prob > 0.5,
+				result->v1.prob > 0.5,
 				is_shingle,
 				session->key_stat,
 				session->ip_stat,
 				cmd->cmd,
-				result->value);
+				result->v1.value);
 
 		if (encrypted) {
 			/* We need also to encrypt reply */
 			ottery_rand_bytes (session->reply.hdr.nonce,
 					sizeof (session->reply.hdr.nonce));
+
+			/*
+			 * For old replies we need to encrypt just old part, otherwise
+			 * decryption would fail due to mac verification mistake
+			 */
+
+			if (session->epoch > RSPAMD_FUZZY_EPOCH10) {
+				len = sizeof (session->reply.rep);
+			}
+			else {
+				len = sizeof (session->reply.rep.v1);
+			}
+
 			rspamd_cryptobox_encrypt_nm_inplace ((guchar *)&session->reply.rep,
-					sizeof (session->reply.rep),
+					len,
 					session->reply.hdr.nonce,
 					session->nm,
 					session->reply.hdr.mac,
@@ -784,7 +800,7 @@ rspamd_fuzzy_process_command (struct fuzzy_session *session)
 	gboolean encrypted = FALSE, is_shingle = FALSE;
 	struct rspamd_fuzzy_cmd *cmd = NULL;
 	struct rspamd_fuzzy_reply result;
-	struct fuzzy_peer_cmd *up_cmd;
+	struct fuzzy_peer_cmd up_cmd;
 	struct fuzzy_peer_request *up_req;
 	struct fuzzy_key_stat *ip_stat = NULL;
 	gchar hexbuf[rspamd_cryptobox_HASHBYTES * 2 + 1];
@@ -813,19 +829,27 @@ rspamd_fuzzy_process_command (struct fuzzy_session *session)
 		encrypted = TRUE;
 		is_shingle = TRUE;
 		break;
+	default:
+		msg_err ("invalid command type: %d", session->cmd_type);
+		return;
 	}
 
+	memset (&result, 0, sizeof (result));
+	memcpy (result.digest, cmd->digest, sizeof (result.digest));
+	result.v1.flag = cmd->flag;
+	result.v1.tag = cmd->tag;
+
 	if (G_UNLIKELY (cmd == NULL || up_len == 0)) {
-		result.value = 500;
-		result.prob = 0.0;
+		result.v1.value = 500;
+		result.v1.prob = 0.0;
 		rspamd_fuzzy_make_reply (cmd, &result, session, encrypted, is_shingle);
 		return;
 	}
 
 	if (session->ctx->encrypted_only && !encrypted) {
 		/* Do not accept unencrypted commands */
-		result.value = 403;
-		result.prob = 0.0;
+		result.v1.value = 403;
+		result.v1.prob = 0.0;
 		rspamd_fuzzy_make_reply (cmd, &result, session, encrypted, is_shingle);
 		return;
 	}
@@ -836,7 +860,7 @@ rspamd_fuzzy_process_command (struct fuzzy_session *session)
 
 		if (ip_stat == NULL) {
 			naddr = rspamd_inet_address_copy (session->addr);
-			ip_stat = g_slice_alloc0 (sizeof (*ip_stat));
+			ip_stat = g_malloc0 (sizeof (*ip_stat));
 			rspamd_lru_hash_insert (session->key_stat->last_ips,
 					naddr, ip_stat, -1, 0);
 		}
@@ -844,46 +868,50 @@ rspamd_fuzzy_process_command (struct fuzzy_session *session)
 		session->ip_stat = ip_stat;
 	}
 
-	result.flag = cmd->flag;
-
 	if (cmd->cmd == FUZZY_CHECK) {
-		if (G_UNLIKELY (session->ctx->collection_mode)) {
-			result.prob = 0;
-			result.value = 500;
-			result.flag = 0;
-			rspamd_fuzzy_make_reply (cmd, &result, session, encrypted, is_shingle);
+		if (rspamd_fuzzy_check_client (session, FALSE)) {
+			if (G_UNLIKELY (session->ctx->collection_mode)) {
+				result.v1.prob = 0;
+				result.v1.value = 500;
+				result.v1.flag = 0;
+				rspamd_fuzzy_make_reply (cmd, &result, session, encrypted,
+						is_shingle);
+			} else {
+				REF_RETAIN (session);
+				rspamd_fuzzy_backend_check (session->ctx->backend, cmd,
+						rspamd_fuzzy_check_callback, session);
+			}
 		}
 		else {
-			REF_RETAIN (session);
-			rspamd_fuzzy_backend_check (session->ctx->backend, cmd,
-					rspamd_fuzzy_check_callback, session);
+			result.v1.value = 403;
+			result.v1.prob = 0.0;
 		}
 	}
 	else if (cmd->cmd == FUZZY_STAT) {
 		if (G_UNLIKELY (session->ctx->collection_mode)) {
-			result.prob = 0;
-			result.value = 500;
-			result.flag = 0;
+			result.v1.prob = 0;
+			result.v1.value = 500;
+			result.v1.flag = 0;
 			rspamd_fuzzy_make_reply (cmd, &result, session, encrypted, is_shingle);
 		}
 		else {
-			result.prob = 1.0;
-			result.value = 0;
-			result.flag = session->ctx->stat.fuzzy_hashes;
+			result.v1.prob = 1.0;
+			result.v1.value = 0;
+			result.v1.flag = session->ctx->stat.fuzzy_hashes;
 			rspamd_fuzzy_make_reply (cmd, &result, session, encrypted, is_shingle);
 		}
 	}
 	else {
-		if (rspamd_fuzzy_check_client (session)) {
+		if (rspamd_fuzzy_check_client (session, TRUE)) {
 			/* Check whitelist */
-			if (session->ctx->skip_hashes) {
-				rspamd_encode_hex_buf (cmd->digest, sizeof (cmd->cmd),
+			if (session->ctx->skip_hashes && cmd->cmd == FUZZY_WRITE) {
+				rspamd_encode_hex_buf (cmd->digest, sizeof (cmd->digest),
 					hexbuf, sizeof (hexbuf) - 1);
 				hexbuf[sizeof (hexbuf) - 1] = '\0';
 
 				if (g_hash_table_lookup (session->ctx->skip_hashes, hexbuf)) {
-					result.value = 401;
-					result.prob = 0.0;
+					result.v1.value = 401;
+					result.v1.prob = 0.0;
 
 					goto reply;
 				}
@@ -891,17 +919,16 @@ rspamd_fuzzy_process_command (struct fuzzy_session *session)
 
 			if (session->worker->index == 0 || session->ctx->peer_fd == -1) {
 				/* Just add to the queue */
-				up_cmd = g_slice_alloc0 (sizeof (*up_cmd));
-				up_cmd->is_shingle = is_shingle;
+				up_cmd.is_shingle = is_shingle;
 				ptr = is_shingle ?
-						(gpointer)&up_cmd->cmd.shingle :
-						(gpointer)&up_cmd->cmd.normal;
+						(gpointer)&up_cmd.cmd.shingle :
+						(gpointer)&up_cmd.cmd.normal;
 				memcpy (ptr, cmd, up_len);
-				g_queue_push_tail (session->ctx->updates_pending, up_cmd);
+				g_array_append_val (session->ctx->updates_pending, up_cmd);
 			}
 			else {
 				/* We need to send request to the peer */
-				up_req = g_slice_alloc0 (sizeof (*up_req));
+				up_req = g_malloc0 (sizeof (*up_req));
 				up_req->cmd.is_shingle = is_shingle;
 				ptr = is_shingle ?
 						(gpointer)&up_req->cmd.cmd.shingle :
@@ -913,12 +940,12 @@ rspamd_fuzzy_process_command (struct fuzzy_session *session)
 				event_add (&up_req->io_ev, NULL);
 			}
 
-			result.value = 0;
-			result.prob = 1.0;
+			result.v1.value = 0;
+			result.v1.prob = 1.0;
 		}
 		else {
-			result.value = 403;
-			result.prob = 0.0;
+			result.v1.value = 403;
+			result.v1.prob = 0.0;
 		}
 reply:
 		rspamd_fuzzy_make_reply (cmd, &result, session, encrypted, is_shingle);
@@ -932,6 +959,18 @@ rspamd_fuzzy_command_valid (struct rspamd_fuzzy_cmd *cmd, gint r)
 	enum rspamd_fuzzy_epoch ret = RSPAMD_FUZZY_EPOCH_MAX;
 
 	switch (cmd->version) {
+	case 4:
+		if (cmd->shingles_count > 0) {
+			if (r == sizeof (struct rspamd_fuzzy_shingle_cmd)) {
+				ret = RSPAMD_FUZZY_EPOCH11;
+			}
+		}
+		else {
+			if (r == sizeof (*cmd)) {
+				ret = RSPAMD_FUZZY_EPOCH11;
+			}
+		}
+		break;
 	case 3:
 		if (cmd->shingles_count > 0) {
 			if (r == sizeof (struct rspamd_fuzzy_shingle_cmd)) {
@@ -1112,13 +1151,13 @@ rspamd_fuzzy_mirror_process_update (struct fuzzy_master_update_session *session,
 	gsize remain;
 	gint32 revision;
 	guint32 len = 0, cnt = 0;
-	struct fuzzy_peer_cmd cmd, *pcmd;
+	struct fuzzy_peer_cmd cmd;
 	enum {
 		read_len = 0,
 		read_data,
 		finish_processing
 	} state = read_len;
-	GList *updates = NULL, *cur;
+
 	gpointer flag_ptr;
 
 	/*
@@ -1205,9 +1244,20 @@ rspamd_fuzzy_mirror_process_update (struct fuzzy_master_update_session *session,
 				goto err;
 			}
 
-			pcmd = g_slice_alloc (sizeof (cmd));
-			memcpy (pcmd, &cmd, len);
-			updates = g_list_prepend (updates, pcmd);
+			if (cmd.is_shingle) {
+				if ((flag_ptr = g_hash_table_lookup (session->ctx->master_flags,
+						GUINT_TO_POINTER (cmd.cmd.shingle.basic.flag))) != NULL) {
+					cmd.cmd.shingle.basic.flag = GPOINTER_TO_UINT (flag_ptr);
+				}
+			}
+			else {
+				if ((flag_ptr = g_hash_table_lookup (session->ctx->master_flags,
+						GUINT_TO_POINTER (cmd.cmd.normal.flag))) != NULL) {
+					cmd.cmd.normal.flag = GPOINTER_TO_UINT (flag_ptr);
+				}
+			}
+
+			g_array_append_val (session->ctx->updates_pending, cmd);
 
 			p += len;
 			remain -= len;
@@ -1222,27 +1272,6 @@ rspamd_fuzzy_mirror_process_update (struct fuzzy_master_update_session *session,
 		}
 	}
 
-	/* Insert elements to the updates from head */
-	for (cur = updates; cur != NULL; cur = g_list_next (cur)) {
-		pcmd = cur->data;
-
-		if (pcmd->is_shingle) {
-			if ((flag_ptr = g_hash_table_lookup (session->ctx->master_flags,
-					GUINT_TO_POINTER (pcmd->cmd.shingle.basic.flag))) != NULL) {
-				pcmd->cmd.shingle.basic.flag = GPOINTER_TO_UINT (flag_ptr);
-			}
-		}
-		else {
-			if ((flag_ptr = g_hash_table_lookup (session->ctx->master_flags,
-					GUINT_TO_POINTER (pcmd->cmd.normal.flag))) != NULL) {
-				pcmd->cmd.normal.flag = GPOINTER_TO_UINT (flag_ptr);
-			}
-		}
-
-
-		g_queue_push_head (session->ctx->updates_pending, cur->data);
-		cur->data = NULL;
-	}
 
 	rspamd_fuzzy_process_updates_queue (session->ctx, session->src, TRUE);
 	msg_info_fuzzy_update ("processed updates from the master %s, "
@@ -1252,16 +1281,7 @@ rspamd_fuzzy_mirror_process_update (struct fuzzy_master_update_session *session,
 			cnt, revision, our_rev);
 
 err:
-	if (updates) {
-		/* We still need to clear queue */
-		for (cur = updates; cur != NULL; cur = g_list_next (cur)) {
-			if (cur->data) {
-				g_slice_free1 (sizeof (cmd), cur->data);
-			}
-		}
-
-		g_list_free (updates);
-	}
+	return;
 }
 
 
@@ -1273,7 +1293,7 @@ fuzzy_session_destroy (gpointer d)
 	rspamd_inet_address_free (session->addr);
 	rspamd_explicit_memzero (session->nm, sizeof (session->nm));
 	session->worker->nconns--;
-	g_slice_free1 (sizeof (*session), session);
+	g_free (session);
 }
 
 static void
@@ -1288,7 +1308,7 @@ rspamd_fuzzy_mirror_session_destroy (struct fuzzy_master_update_session *session
 		if (session->psrc) {
 			g_free (session->psrc);
 		}
-		g_slice_free1 (sizeof (*session), session);
+		g_free (session);
 	}
 }
 
@@ -1428,9 +1448,7 @@ rspamd_fuzzy_collection_finish_handler (struct rspamd_http_connection_entry *con
 
 
 	rspamd_inet_address_free (session->from_addr);
-
-
-	g_slice_free1 (sizeof (struct rspamd_fuzzy_collection_session), session);
+	g_free (session);
 }
 
 void
@@ -1514,7 +1532,7 @@ rspamd_fuzzy_collection_data (struct rspamd_http_connection_entry *conn_ent,
 	struct rspamd_fuzzy_collection_session *session = conn_ent->ud;
 	const rspamd_ftok_t *sign_header;
 	struct rspamd_fuzzy_storage_ctx *ctx;
-	GList *cur;
+	guint i;
 	struct fuzzy_peer_cmd *io_cmd;
 	rspamd_fstring_t *reply;
 	GError *err = NULL;
@@ -1565,7 +1583,6 @@ rspamd_fuzzy_collection_data (struct rspamd_http_connection_entry *conn_ent,
 	ottery_rand_bytes (ctx->cookie, sizeof (ctx->cookie));
 
 	/* Send&Clear updates */
-	cur = ctx->updates_pending->head;
 	reply = rspamd_fstring_sized_new (8192);
 	/*
 	 * Message format:
@@ -1578,10 +1595,9 @@ rspamd_fuzzy_collection_data (struct rspamd_http_connection_entry *conn_ent,
 	 */
 	reply = rspamd_fstring_append (reply, (const gchar *)&ctx->collection_id,
 					sizeof (ctx->collection_id));
-	cur = ctx->updates_pending->head;
 
-	while (cur) {
-		io_cmd = cur->data;
+	for (i = 0; i < ctx->updates_pending->len; i ++) {
+		io_cmd = &g_array_index (ctx->updates_pending, struct fuzzy_peer_cmd, i);
 
 		if (io_cmd->is_shingle) {
 			cmdlen = sizeof (io_cmd->cmd.shingle) + sizeof (guint32);
@@ -1596,9 +1612,7 @@ rspamd_fuzzy_collection_data (struct rspamd_http_connection_entry *conn_ent,
 				sizeof (cmdlen));
 		reply = rspamd_fstring_append (reply, (const gchar *)io_cmd,
 				cmdlen);
-		g_slice_free1 (sizeof (*io_cmd), io_cmd);
 		nupdates ++;
-		cur = g_list_next (cur);
 	}
 
 	msg_info_fuzzy_collection ("collection %d done, send %d updates",
@@ -1608,7 +1622,7 @@ rspamd_fuzzy_collection_data (struct rspamd_http_connection_entry *conn_ent,
 	reply = rspamd_fstring_append (reply, (const gchar *)&cmdlen,
 			sizeof (cmdlen));
 
-	g_queue_clear (ctx->updates_pending);
+	ctx->updates_pending->len = 0;
 	/* Clear failed attempts counter */
 	ctx->updates_failed = 0;
 	ctx->collection_id ++;
@@ -1648,7 +1662,7 @@ accept_fuzzy_collection_socket (gint fd, short what, void *arg)
 		return;
 	}
 
-	session = g_slice_alloc0 (sizeof (*session));
+	session = g_malloc0 (sizeof (*session));
 	session->ctx = ctx;
 	session->worker = worker;
 	rspamd_random_hex (session->uid, sizeof (session->uid) - 1);
@@ -1665,31 +1679,21 @@ static void
 rspamd_fuzzy_collection_periodic (gint fd, gshort what, gpointer ud)
 {
 	struct rspamd_fuzzy_storage_ctx *ctx = ud;
-	GList *cur;
-	struct fuzzy_peer_cmd *io_cmd;
 
 	if (++ctx->updates_failed > ctx->updates_maxfail) {
 		msg_err ("cannot store more data in workqueue, discard "
 				"%ud updates after %d missed collection points",
-				g_queue_get_length (ctx->updates_pending),
+				ctx->updates_pending->len,
 				ctx->updates_maxfail);
 		ctx->updates_failed = 0;
-		cur = ctx->updates_pending->head;
-
-		while (cur) {
-			io_cmd = cur->data;
-			g_slice_free1 (sizeof (*io_cmd), io_cmd);
-			cur = g_list_next (cur);
-		}
-
-		g_queue_clear (ctx->updates_pending);
+		ctx->updates_pending->len = 0;
 		/* Regenerate cookie */
 		ottery_rand_bytes (ctx->cookie, sizeof (ctx->cookie));
 	}
 	else {
 		msg_err ("fuzzy data has not been collected in time, "
 				"%ud updates are still pending, %d updates left",
-				g_queue_get_length (ctx->updates_pending),
+				ctx->updates_pending->len,
 				ctx->updates_maxfail - ctx->updates_failed);
 	}
 
@@ -1735,7 +1739,7 @@ accept_fuzzy_mirror_socket (gint fd, short what, void *arg)
 
 		return;
 	}
-	else if (radix_find_compressed_addr (ctx->master_ips, addr) == RADIX_NO_VALUE) {
+	else if (rspamd_match_radix_map_addr (ctx->master_ips, addr) == NULL) {
 		msg_err ("deny update request from %s",
 				rspamd_inet_address_to_string (addr));
 		rspamd_inet_address_free (addr);
@@ -1753,7 +1757,7 @@ accept_fuzzy_mirror_socket (gint fd, short what, void *arg)
 		return;
 	}
 
-	session = g_slice_alloc0 (sizeof (*session));
+	session = g_malloc0 (sizeof (*session));
 	session->name = rspamd_inet_address_to_string (addr);
 	rspamd_random_hex (session->uid, sizeof (session->uid) - 1);
 	session->uid[sizeof (session->uid) - 1] = '\0';
@@ -1818,7 +1822,7 @@ accept_fuzzy_socket (gint fd, short what, void *arg)
 				return;
 			}
 
-			session = g_slice_alloc0 (sizeof (*session));
+			session = g_malloc0 (sizeof (*session));
 			REF_INIT_RETAIN (session, fuzzy_session_destroy);
 			session->worker = worker;
 			session->fd = fd;
@@ -1860,7 +1864,7 @@ rspamd_fuzzy_storage_periodic_callback (void *ud)
 {
 	struct rspamd_fuzzy_storage_ctx *ctx = ud;
 
-	if (g_queue_get_length (ctx->updates_pending) > 0) {
+	if (ctx->updates_pending->len > 0) {
 		rspamd_fuzzy_process_updates_queue (ctx, local_db_name, FALSE);
 
 		return TRUE;
@@ -2108,7 +2112,7 @@ rspamd_fuzzy_storage_stat (struct rspamd_main *rspamd_main,
 	gchar tmppath[PATH_MAX];
 
 	memset (&rep, 0, sizeof (rep));
-	rep.type = RSPAMD_CONTROL_RELOAD;
+	rep.type = RSPAMD_CONTROL_FUZZY_STAT;
 
 	rspamd_snprintf (tmppath, sizeof (tmppath), "%s%c%s-XXXXXXXXXX",
 			rspamd_main->cfg->temp_dir, G_DIR_SEPARATOR, "fuzzy-stat");
@@ -2199,7 +2203,7 @@ fuzzy_storage_parse_mirror (rspamd_mempool_t *pool,
 		return FALSE;
 	}
 
-	up = g_slice_alloc0 (sizeof (*up));
+	up = g_malloc0 (sizeof (*up));
 	up->name = g_strdup (ucl_object_tostring (elt));
 
 	elt = ucl_object_lookup (obj, "key");
@@ -2244,7 +2248,7 @@ err:
 		rspamd_pubkey_unref (up->key);
 	}
 
-	g_slice_free1 (sizeof (*up), up);
+	g_free (up);
 
 	return FALSE;
 }
@@ -2523,6 +2527,15 @@ init_fuzzy (struct rspamd_config *cfg)
 			0,
 			"Allow master/slave updates from the following IP addresses");
 
+	rspamd_rcl_register_worker_option (cfg,
+			type,
+			"blocked",
+			rspamd_rcl_parse_struct_ucl,
+			ctx,
+			G_STRUCT_OFFSET (struct rspamd_fuzzy_storage_ctx, blocked_map),
+			0,
+			"Block requests from specific networks");
+
 	rspamd_rcl_register_worker_option (cfg,
 			type,
 			"master_key",
@@ -2613,7 +2626,7 @@ init_fuzzy (struct rspamd_config *cfg)
 static void
 rspamd_fuzzy_peer_io (gint fd, gshort what, gpointer d)
 {
-	struct fuzzy_peer_cmd cmd, *pcmd;
+	struct fuzzy_peer_cmd cmd;
 	struct rspamd_fuzzy_storage_ctx *ctx = d;
 	gssize r;
 
@@ -2629,9 +2642,7 @@ rspamd_fuzzy_peer_io (gint fd, gshort what, gpointer d)
 		}
 	}
 	else {
-		pcmd = g_slice_alloc (sizeof (*pcmd));
-		memcpy (pcmd, &cmd, sizeof (cmd));
-		g_queue_push_tail (ctx->updates_pending, pcmd);
+		g_array_append_val (ctx->updates_pending, cmd);
 	}
 }
 
@@ -2662,7 +2673,7 @@ fuzzy_peer_rep (struct rspamd_worker *worker,
 
 		if (ls->fd != -1) {
 			if (ls->type == RSPAMD_WORKER_SOCKET_UDP) {
-				accept_events = g_slice_alloc0 (sizeof (struct event) * 2);
+				accept_events = g_malloc0 (sizeof (struct event) * 2);
 				event_set (&accept_events[0], ls->fd, EV_READ | EV_PERSIST,
 						accept_fuzzy_socket, worker);
 				event_base_set (ctx->ev_base, &accept_events[0]);
@@ -2672,7 +2683,7 @@ fuzzy_peer_rep (struct rspamd_worker *worker,
 			}
 			else if (worker->index == 0) {
 				/* We allow TCP listeners only for a update worker */
-				accept_events = g_slice_alloc0 (sizeof (struct event) * 2);
+				accept_events = g_malloc0 (sizeof (struct event) * 2);
 
 				if (ctx->collection_mode) {
 					event_set (&accept_events[0], ls->fd, EV_READ | EV_PERSIST,
@@ -2714,8 +2725,7 @@ start_fuzzy (struct rspamd_worker *worker)
 
 	ctx->ev_base = rspamd_prepare_worker (worker,
 			"fuzzy",
-			NULL,
-			FALSE);
+			NULL);
 	ctx->peer_fd = -1;
 	ctx->worker = worker;
 	ctx->cfg = worker->srv->cfg;
@@ -2738,7 +2748,9 @@ start_fuzzy (struct rspamd_worker *worker)
 		if ((ctx->backend = rspamd_fuzzy_backend_create (ctx->ev_base,
 				worker->cf->options, cfg, &err)) == NULL) {
 			msg_err ("cannot open backend: %e", err);
-			g_error_free (err);
+			if (err) {
+				g_error_free (err);
+			}
 			exit (EXIT_SUCCESS);
 		}
 
@@ -2746,7 +2758,8 @@ start_fuzzy (struct rspamd_worker *worker)
 
 
 		if (worker->index == 0) {
-			ctx->updates_pending = g_queue_new ();
+			ctx->updates_pending = g_array_sized_new (FALSE, FALSE,
+					sizeof (struct fuzzy_peer_cmd), 1024);
 			rspamd_fuzzy_backend_start_update (ctx->backend, ctx->sync_timeout,
 					rspamd_fuzzy_storage_periodic_callback, ctx);
 		}
@@ -2770,7 +2783,8 @@ start_fuzzy (struct rspamd_worker *worker)
 		 * we collect fuzzy hashes in the updates queue and ignore all read commands
 		 */
 		if (worker->index == 0) {
-			ctx->updates_pending = g_queue_new ();
+			ctx->updates_pending = g_array_sized_new (FALSE, FALSE,
+					sizeof (struct fuzzy_peer_cmd), 1024);
 			double_to_tv (ctx->sync_timeout, &ctx->stat_tv);
 			event_set (&ctx->stat_ev, -1, EV_TIMEOUT|EV_PERSIST,
 					rspamd_fuzzy_collection_periodic, ctx);
@@ -2848,11 +2862,13 @@ start_fuzzy (struct rspamd_worker *worker)
 				"Allow fuzzy updates from specified addresses",
 				&ctx->update_ips, NULL);
 	}
+
 	if (ctx->masters_map != NULL) {
 		rspamd_config_radix_from_ucl (worker->srv->cfg, ctx->masters_map,
 				"Allow fuzzy master/slave updates from specified addresses",
 				&ctx->master_ips, NULL);
 	}
+
 	if (ctx->skip_map != NULL) {
 		struct rspamd_map *m;
 
@@ -2867,6 +2883,12 @@ start_fuzzy (struct rspamd_worker *worker)
 		}
 	}
 
+	if (ctx->blocked_map != NULL) {
+		rspamd_config_radix_from_ucl (worker->srv->cfg, ctx->blocked_map,
+				"Block fuzzy requests from the specific IPs",
+				&ctx->blocked_ips, NULL);
+	}
+
 	/* Maps events */
 	ctx->resolver = dns_resolver_init (worker->srv->logger,
 				ctx->ev_base,
@@ -2876,7 +2898,6 @@ start_fuzzy (struct rspamd_worker *worker)
 	/* Get peer pipe */
 	memset (&srv_cmd, 0, sizeof (srv_cmd));
 	srv_cmd.type = RSPAMD_SRV_SOCKETPAIR;
-	srv_cmd.id = ottery_rand_uint64 ();
 	srv_cmd.cmd.spair.af = SOCK_DGRAM;
 	srv_cmd.cmd.spair.pair_num = worker->index;
 	memset (srv_cmd.cmd.spair.pair_id, 0, sizeof (srv_cmd.cmd.spair.pair_id));
@@ -2888,7 +2909,7 @@ start_fuzzy (struct rspamd_worker *worker)
 	event_base_loop (ctx->ev_base, 0);
 	rspamd_worker_block_signals ();
 
-	if (worker->index == 0 && g_queue_get_length (ctx->updates_pending) > 0) {
+	if (worker->index == 0 && ctx->updates_pending->len > 0) {
 		if (!ctx->collection_mode) {
 			rspamd_fuzzy_process_updates_queue (ctx, local_db_name, FALSE);
 			event_base_loop (ctx->ev_base, 0);
@@ -2922,6 +2943,8 @@ start_fuzzy (struct rspamd_worker *worker)
 
 			close (fd);
 		}
+
+		g_array_free (ctx->updates_pending, TRUE);
 	}
 
 	rspamd_log_close (worker->srv->logger);
