@@ -54,14 +54,23 @@ enum rspamd_dkim_param_type {
 	DKIM_PARAM_BODYHASH,
 	DKIM_PARAM_BODYLENGTH,
 	DKIM_PARAM_IDX,
-	DKIM_PARAM_CV
+	DKIM_PARAM_CV,
+	DKIM_PARAM_IGNORE
 };
 
 /* Signature methods */
 enum rspamd_sign_type {
 	DKIM_SIGN_UNKNOWN = -2,
 	DKIM_SIGN_RSASHA1 = 0,
-	DKIM_SIGN_RSASHA256
+	DKIM_SIGN_RSASHA256,
+	DKIM_SIGN_RSASHA512,
+	DKIM_SIGN_ECDSASHA256,
+	DKIM_SIGN_ECDSASHA512
+};
+
+enum rspamd_dkim_key_type {
+	RSPAMD_DKIM_KEY_RSA = 0,
+	RSPAMD_DKIM_KEY_ECDSA
 };
 
 #define RSPAMD_DKIM_MAX_ARC_IDX 10
@@ -78,17 +87,31 @@ enum rspamd_sign_type {
         "dkim", ctx->pool->tag.uid, \
         G_STRFUNC, \
         __VA_ARGS__)
-#define msg_debug_dkim(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        "dkim", ctx->pool->tag.uid, \
+#define msg_debug_dkim(...)  rspamd_conditional_debug_fast (NULL, NULL, \
+        rspamd_dkim_log_id, "dkim", ctx->pool->tag.uid, \
         G_STRFUNC, \
         __VA_ARGS__)
 
+INIT_LOG_MODULE(dkim)
+
+#define RSPAMD_DKIM_FLAG_OVERSIGN (1u << 0)
+
+union rspamd_dkim_header_stat {
+	struct _st {
+		guint16 count;
+		guint16 flags;
+	} s;
+	guint32 n;
+};
+
 struct rspamd_dkim_common_ctx {
 	rspamd_mempool_t *pool;
+	guint64 sig_hash;
 	gsize len;
 	gint header_canon_type;
 	gint body_canon_type;
 	GPtrArray *hlist;
+	GHashTable *htable; /* header -> count mapping */
 	EVP_MD_CTX *headers_hash;
 	EVP_MD_CTX *body_hash;
 	enum rspamd_dkim_type type;
@@ -126,7 +149,11 @@ struct rspamd_dkim_key_s {
 	guint keylen;
 	gsize decoded_len;
 	guint ttl;
-	RSA *key_rsa;
+	union {
+		RSA *key_rsa;
+		EC_KEY *key_ecdsa;
+	} key;
+	enum rspamd_dkim_key_type type;
 	BIO *key_bio;
 	EVP_PKEY *key_evp;
 	ref_entry_t ref;
@@ -233,6 +260,7 @@ static const dkim_parse_param_f parser_funcs[] = {
 	[DKIM_PARAM_BODYLENGTH] = rspamd_dkim_parse_bodylength,
 	[DKIM_PARAM_IDX] = rspamd_dkim_parse_idx,
 	[DKIM_PARAM_CV] = rspamd_dkim_parse_cv,
+	[DKIM_PARAM_IGNORE] = rspamd_dkim_parse_ignore,
 };
 
 #define DKIM_ERROR dkim_error_quark ()
@@ -272,6 +300,20 @@ rspamd_dkim_parse_signalg (rspamd_dkim_context_t * ctx,
 			ctx->sig_alg = DKIM_SIGN_RSASHA256;
 			return TRUE;
 		}
+		else if (memcmp (param, "rsa-sha512", len) == 0) {
+			ctx->sig_alg = DKIM_SIGN_RSASHA512;
+			return TRUE;
+		}
+	}
+	else if (len == sizeof ("ecdsa256-sha256") - 1) {
+		if (memcmp (param, "ecdsa256-sha256", len) == 0) {
+			ctx->sig_alg = DKIM_SIGN_ECDSASHA256;
+			return TRUE;
+		}
+		else if (memcmp (param, "ecdsa256-sha512", len) == 0) {
+			ctx->sig_alg = DKIM_SIGN_ECDSASHA512;
+			return TRUE;
+		}
 	}
 
 	g_set_error (err,
@@ -387,15 +429,16 @@ static gboolean
 rspamd_dkim_parse_hdrlist_common (struct rspamd_dkim_common_ctx *ctx,
 	const gchar *param,
 	gsize len,
+	gboolean sign,
 	GError **err)
 {
 	const gchar *c, *p, *end = param + len;
 	gchar *h;
-	gboolean from_found = FALSE;
+	gboolean from_found = FALSE, oversign;
 	guint count = 0;
 	struct rspamd_dkim_header *new;
 	gpointer found;
-	GHashTable *htb;
+	union rspamd_dkim_header_stat u;
 
 	p = param;
 	while (p <= end) {
@@ -414,16 +457,22 @@ rspamd_dkim_parse_hdrlist_common (struct rspamd_dkim_common_ctx *ctx,
 
 	c = param;
 	p = param;
-	htb = g_hash_table_new (rspamd_strcase_hash, rspamd_strcase_equal);
+	ctx->htable = g_hash_table_new (rspamd_strcase_hash, rspamd_strcase_equal);
 
 	while (p <= end) {
 		if ((p == end || *p == ':') && p - c > 0) {
-
+			oversign = FALSE;
 			h = rspamd_mempool_alloc (ctx->pool, p - c + 1);
 			rspamd_strlcpy (h, c, p - c + 1);
 
 			g_strstrip (h);
 
+			if (sign && rspamd_lc_cmp (h, "(o)", 3) == 0) {
+				oversign = TRUE;
+				h += 3;
+				msg_debug_dkim ("oversign header: %s", h);
+			}
+
 			/* Check mandatory from */
 			if (!from_found && g_ascii_strcasecmp (h, "from") == 0) {
 				from_found = TRUE;
@@ -433,20 +482,33 @@ rspamd_dkim_parse_hdrlist_common (struct rspamd_dkim_common_ctx *ctx,
 					sizeof (struct rspamd_dkim_header));
 			new->name = h;
 			new->count = 0;
+			u.n = 0;
 
 			g_ptr_array_add (ctx->hlist, new);
+			found = g_hash_table_lookup (ctx->htable, h);
 
-			if ((found = g_hash_table_lookup (htb, h)) != NULL) {
-				count = GPOINTER_TO_UINT (found);
-				new->count = count;
-				count ++;
+			if (oversign) {
+				if (found) {
+					msg_err_dkim ("specified oversigned header more than once: %s",
+							h);
+				}
+
+				u.s.flags |= RSPAMD_DKIM_FLAG_OVERSIGN;
+				u.s.count = 0;
 			}
 			else {
-				/* Insert new header order to the list */
-				count = new->count + 1;
+				if (found != NULL) {
+					u.n = GPOINTER_TO_UINT (found);
+					new->count = u.s.count;
+					u.s.count ++;
+				}
+				else {
+					/* Insert new header order to the list */
+					u.s.count = new->count + 1;
+				}
 			}
 
-			g_hash_table_insert (htb, h, GUINT_TO_POINTER (count));
+			g_hash_table_insert (ctx->htable, h, GUINT_TO_POINTER (u.n));
 
 			c = p + 1;
 			p++;
@@ -456,8 +518,6 @@ rspamd_dkim_parse_hdrlist_common (struct rspamd_dkim_common_ctx *ctx,
 		}
 	}
 
-	g_hash_table_unref (htb);
-
 	if (!ctx->hlist) {
 		g_set_error (err,
 			DKIM_ERROR,
@@ -474,10 +534,13 @@ rspamd_dkim_parse_hdrlist_common (struct rspamd_dkim_common_ctx *ctx,
 				"invalid dkim header list, from header is missing");
 			return FALSE;
 		}
-		/* Reverse list */
+
 		rspamd_mempool_add_destructor (ctx->pool,
 			(rspamd_mempool_destruct_t)rspamd_dkim_hlist_free,
 			ctx->hlist);
+		rspamd_mempool_add_destructor (ctx->pool,
+				(rspamd_mempool_destruct_t)g_hash_table_unref,
+				ctx->htable);
 	}
 
 	return TRUE;
@@ -489,7 +552,7 @@ rspamd_dkim_parse_hdrlist (rspamd_dkim_context_t *ctx,
 	gsize len,
 	GError **err)
 {
-	return rspamd_dkim_parse_hdrlist_common (&ctx->common, param, len, err);
+	return rspamd_dkim_parse_hdrlist_common (&ctx->common, param, len, FALSE, err);
 }
 
 static gboolean
@@ -727,6 +790,8 @@ rspamd_create_dkim_context (const gchar *sig,
 	p = sig;
 	c = sig;
 	end = p + strlen (p);
+	ctx->common.sig_hash = rspamd_cryptobox_fast_hash (sig, end - sig,
+			rspamd_hash_seed ());
 
 	while (p <= end) {
 		switch (state) {
@@ -841,6 +906,9 @@ rspamd_create_dkim_context (const gchar *sig,
 				case 'z':
 					param = DKIM_PARAM_COPIEDHDRS;
 					break;
+				case 'r':
+					param = DKIM_PARAM_IGNORE;
+					break;
 				default:
 					g_set_error (err,
 						DKIM_ERROR,
@@ -1021,7 +1089,8 @@ rspamd_create_dkim_context (const gchar *sig,
 				return NULL;
 			}
 
-		} else if (ctx->sig_alg == DKIM_SIGN_RSASHA256) {
+		} else if (ctx->sig_alg == DKIM_SIGN_RSASHA256 ||
+				ctx->sig_alg == DKIM_SIGN_ECDSASHA256) {
 			if (ctx->bhlen !=
 					(guint) EVP_MD_size (EVP_sha256 ())) {
 				g_set_error (err,
@@ -1031,6 +1100,17 @@ rspamd_create_dkim_context (const gchar *sig,
 						ctx->bhlen);
 				return NULL;
 			}
+		} else if (ctx->sig_alg == DKIM_SIGN_RSASHA512 ||
+				ctx->sig_alg == DKIM_SIGN_ECDSASHA512) {
+			if (ctx->bhlen !=
+					(guint) EVP_MD_size (EVP_sha512 ())) {
+				g_set_error (err,
+						DKIM_ERROR,
+						DKIM_SIGERROR_BADSIG,
+						"signature has incorrect length: %zu",
+						ctx->bhlen);
+				return NULL;
+			}
 		}
 	}
 
@@ -1086,9 +1166,14 @@ rspamd_create_dkim_context (const gchar *sig,
 	if (ctx->sig_alg == DKIM_SIGN_RSASHA1) {
 		md_alg = EVP_sha1 ();
 	}
-	else if (ctx->sig_alg == DKIM_SIGN_RSASHA256) {
+	else if (ctx->sig_alg == DKIM_SIGN_RSASHA256 ||
+			ctx->sig_alg == DKIM_SIGN_ECDSASHA256) {
 		md_alg = EVP_sha256 ();
 	}
+	else if (ctx->sig_alg == DKIM_SIGN_RSASHA512 ||
+			ctx->sig_alg == DKIM_SIGN_ECDSASHA512) {
+		md_alg = EVP_sha512 ();
+	}
 	else {
 		g_set_error (err,
 			DKIM_ERROR,
@@ -1129,7 +1214,7 @@ struct rspamd_dkim_key_cbdata {
 
 static rspamd_dkim_key_t *
 rspamd_dkim_make_key (rspamd_dkim_context_t *ctx, const gchar *keydata,
-		guint keylen, GError **err)
+		guint keylen, enum rspamd_dkim_key_type type, GError **err)
 {
 	rspamd_dkim_key_t *key = NULL;
 
@@ -1138,11 +1223,12 @@ rspamd_dkim_make_key (rspamd_dkim_context_t *ctx, const gchar *keydata,
 		return NULL;
 	}
 
-	key = g_slice_alloc0 (sizeof (rspamd_dkim_key_t));
+	key = g_malloc0 (sizeof (rspamd_dkim_key_t));
 	REF_INIT_RETAIN (key, rspamd_dkim_key_free);
-	key->keydata = g_slice_alloc0 (keylen + 1);
+	key->keydata = g_malloc0 (keylen + 1);
 	key->decoded_len = keylen;
 	key->keylen = keylen;
+	key->type = type;
 
 	rspamd_cryptobox_base64_decode (keydata, keylen, key->keydata,
 			&key->decoded_len);
@@ -1171,16 +1257,31 @@ rspamd_dkim_make_key (rspamd_dkim_context_t *ctx, const gchar *keydata,
 		return NULL;
 	}
 
-	key->key_rsa = EVP_PKEY_get1_RSA (key->key_evp);
+	if (type == RSPAMD_DKIM_KEY_RSA) {
+		key->key.key_rsa = EVP_PKEY_get1_RSA (key->key_evp);
 
-	if (key->key_rsa == NULL) {
-		g_set_error (err,
-			DKIM_ERROR,
-			DKIM_SIGERROR_KEYFAIL,
-			"cannot extract rsa key from evp key");
-		REF_RELEASE (key);
+		if (key->key.key_rsa == NULL) {
+			g_set_error (err,
+					DKIM_ERROR,
+					DKIM_SIGERROR_KEYFAIL,
+					"cannot extract rsa key from evp key");
+			REF_RELEASE (key);
 
-		return NULL;
+			return NULL;
+		}
+	}
+	else {
+		key->key.key_ecdsa = EVP_PKEY_get1_EC_KEY (key->key_evp);
+
+		if (key->key.key_ecdsa == NULL) {
+			g_set_error (err,
+					DKIM_ERROR,
+					DKIM_SIGERROR_KEYFAIL,
+					"cannot extract ecdsa key from evp key");
+			REF_RELEASE (key);
+
+			return NULL;
+		}
 	}
 
 	return key;
@@ -1196,15 +1297,23 @@ rspamd_dkim_key_free (rspamd_dkim_key_t *key)
 	if (key->key_evp) {
 		EVP_PKEY_free (key->key_evp);
 	}
-	if (key->key_rsa) {
-		RSA_free (key->key_rsa);
+
+	if (key->type == RSPAMD_DKIM_KEY_RSA) {
+		if (key->key.key_rsa) {
+			RSA_free (key->key.key_rsa);
+		}
+	}
+	else {
+		if (key->key.key_ecdsa) {
+			EC_KEY_free (key->key.key_ecdsa);
+		}
 	}
 	if (key->key_bio) {
 		BIO_free (key->key_bio);
 	}
 
-	g_slice_free1 (key->keylen,				   key->keydata);
-	g_slice_free1 (sizeof (rspamd_dkim_key_t), key);
+	g_free (key->keydata);
+	g_free (key);
 }
 
 void
@@ -1230,64 +1339,118 @@ rspamd_dkim_sign_key_free (rspamd_dkim_sign_key_t *key)
 		}
 	}
 
-	g_slice_free1 (sizeof (rspamd_dkim_sign_key_t), key);
+	g_free (key);
 }
 
 static rspamd_dkim_key_t *
 rspamd_dkim_parse_key (rspamd_dkim_context_t *ctx, const gchar *txt,
 		gsize *keylen, GError **err)
 {
-	const gchar *c, *p, *end;
-	gint state = 0;
-	gsize len;
+	const gchar *c, *p, *end, *key = NULL, *alg = "rsa";
+	enum {
+		read_tag = 0,
+		read_eqsign,
+		read_p_tag,
+		read_k_tag,
+	} state = read_tag;
+	gchar tag = '\0';
+	gsize klen = 0, alglen = 0;
 
 	c = txt;
 	p = txt;
 	end = txt + strlen (txt);
 
-	while (p <= end) {
+	while (p < end) {
 		switch (state) {
-		case 0:
-			if (p != end && p[0] == 'p' && p[1] == '=') {
-				/* We got something like public key */
-				c = p + 2;
-				p = c;
-				state = 1;
+		case read_tag:
+			if (*p == '=') {
+				state = read_eqsign;
+			} else {
+				tag = *p;
 			}
-			else {
-				/* Ignore everything */
+			p++;
+			break;
+		case read_eqsign:
+			if (tag == 'p') {
+				state = read_p_tag;
+				c = p;
+			} else if (tag == 'k') {
+				state = read_k_tag;
+				c = p;
+			} else {
+				/* Unknown tag, ignore */
+				state = read_tag;
+				tag = '\0';
 				p++;
 			}
 			break;
-		case 1:
-			/* State when we got p= and looking for some public key */
-			if ((*p == ';' || p == end) && p > c) {
-				len = p - c;
-
-				if (keylen) {
-					*keylen = len;
-				}
-
-				return rspamd_dkim_make_key (ctx, c, len, err);
+		case read_p_tag:
+			if (*p == ';') {
+				klen = p - c;
+				key = c;
+				state = read_tag;
+				tag = '\0';
 			}
-			else {
-				p++;
+			p++;
+			break;
+		case read_k_tag:
+			if (*p == ';') {
+				alglen = p - c;
+				alg = c;
+				state = read_tag;
+				tag = '\0';
 			}
+			p++;
+			break;
+		default:
 			break;
 		}
 	}
 
-	if (p - c == 0) {
+	/* Leftover */
+	switch (state) {
+	case read_p_tag:
+		klen = p - c;
+		key = c;
+		break;
+	case read_k_tag:
+		alglen = p - c;
+		alg = c;
+		break;
+	default:
+		break;
+	}
+
+	if (klen == 0 || key == NULL) {
 		g_set_error (err,
-			DKIM_ERROR,
-			DKIM_SIGERROR_KEYREVOKED,
-			"key was revoked");
+				DKIM_ERROR,
+				DKIM_SIGERROR_KEYFAIL,
+				"key is missing");
+	}
+
+	if (alglen == 0 || alg == NULL) {
+		alg = "rsa"; /* Implicit */
+	}
+
+	if (alglen == 8 && rspamd_lc_cmp (alg, "ecdsa256", alglen) == 0) {
+		if (keylen) {
+			*keylen = klen;
+		}
+
+		return rspamd_dkim_make_key (ctx, c, klen,
+				RSPAMD_DKIM_KEY_ECDSA, err);
 	}
 	else {
-		g_set_error (err, DKIM_ERROR, DKIM_SIGERROR_KEYFAIL,
-			"key was not found");
+		/* We assume RSA default in all cases */
+		if (alglen != 3 || rspamd_lc_cmp (alg, "rsa", alglen) != 0) {
+			msg_info_dkim ("invalid key algorithm: %*s", (gint)alglen, alg);
+		}
+		return rspamd_dkim_make_key (ctx, c, klen,
+				RSPAMD_DKIM_KEY_RSA, err);
 	}
 
+	g_assert_not_reached ();
+
 	return NULL;
 }
 
@@ -1744,18 +1907,20 @@ rspamd_dkim_hash_update (EVP_MD_CTX *ck, const gchar *begin, gsize len)
 	end = begin + len;
 	p = begin;
 	c = p;
-	while (p != end) {
+
+	while (p < end) {
 		if (*p == '\r') {
-			EVP_DigestUpdate (ck, c,	 p - c);
+			EVP_DigestUpdate (ck, c, p - c);
 			EVP_DigestUpdate (ck, CRLF, sizeof (CRLF) - 1);
 			p++;
-			if (*p == '\n') {
+
+			if (p < end && *p == '\n') {
 				p++;
 			}
 			c = p;
 		}
 		else if (*p == '\n') {
-			EVP_DigestUpdate (ck, c,	 p - c);
+			EVP_DigestUpdate (ck, c, p - c);
 			EVP_DigestUpdate (ck, CRLF, sizeof (CRLF) - 1);
 			p++;
 			c = p;
@@ -1764,7 +1929,8 @@ rspamd_dkim_hash_update (EVP_MD_CTX *ck, const gchar *begin, gsize len)
 			p++;
 		}
 	}
-	if (p != c) {
+
+	if (p > c) {
 		EVP_DigestUpdate (ck, c, p - c);
 	}
 }
@@ -1966,7 +2132,10 @@ rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 				rh_num = ar->len - count - 1;
 			}
 			else {
-				/* Absence of header is just NULL signature update */
+				/*
+				 * If DKIM has less headers requested than there are in a
+				 * message, then it's fine, it allows adding extra headers
+				 */
 				return TRUE;
 			}
 
@@ -2001,9 +2170,10 @@ rspamd_dkim_canonize_header (struct rspamd_dkim_common_ctx *ctx,
 				}
 
 				PTR_ARRAY_FOREACH (ar, i, rh) {
-					if (rspamd_substring_search (rh->raw_value,
-							rh->raw_len, dkim_domain,
-							strlen (dkim_domain)) != -1) {
+					guint64 th = rspamd_cryptobox_fast_hash (rh->decoded,
+							strlen (rh->decoded), rspamd_hash_seed ());
+
+					if (th == ctx->sig_hash) {
 						rspamd_dkim_signature_update (ctx, rh->raw_value,
 								rh->raw_len);
 						break;
@@ -2246,17 +2416,32 @@ rspamd_dkim_check (rspamd_dkim_context_t *ctx,
 	if (ctx->sig_alg == DKIM_SIGN_RSASHA1) {
 		nid = NID_sha1;
 	}
-	else if (ctx->sig_alg == DKIM_SIGN_RSASHA256) {
+	else if (ctx->sig_alg == DKIM_SIGN_RSASHA256 ||
+			ctx->sig_alg == DKIM_SIGN_ECDSASHA256) {
 		nid = NID_sha256;
 	}
+	else if (ctx->sig_alg == DKIM_SIGN_RSASHA512 ||
+			ctx->sig_alg == DKIM_SIGN_ECDSASHA512) {
+		nid = NID_sha512;
+	}
 	else {
 		/* Not reached */
 		nid = NID_sha1;
 	}
 
-	if (RSA_verify (nid, raw_digest, dlen, ctx->b, ctx->blen, key->key_rsa) != 1) {
-		msg_debug_dkim ("rsa verify failed");
-		res = DKIM_REJECT;
+	if (key->type == RSPAMD_DKIM_KEY_RSA) {
+		if (RSA_verify (nid, raw_digest, dlen, ctx->b, ctx->blen,
+				key->key.key_rsa) != 1) {
+			msg_debug_dkim ("rsa verify failed");
+			res = DKIM_REJECT;
+		}
+	}
+	else {
+		if (ECDSA_verify (nid, raw_digest, dlen, ctx->b, ctx->blen,
+				key->key.key_ecdsa) != 1) {
+			msg_debug_dkim ("ecdsa verify failed");
+			res = DKIM_REJECT;
+		}
 	}
 
 	if (ctx->common.type == RSPAMD_DKIM_ARC_SEAL && res == DKIM_CONTINUE) {
@@ -2371,7 +2556,7 @@ rspamd_dkim_sign_key_load (const gchar *what, gsize len,
 		}
 	}
 
-	nkey = g_slice_alloc0 (sizeof (*nkey));
+	nkey = g_malloc0 (sizeof (*nkey));
 	nkey->type = type;
 	nkey->mtime = mtime;
 
@@ -2522,7 +2707,7 @@ rspamd_create_dkim_sign_context (struct rspamd_task *task,
 
 	if (type != RSPAMD_DKIM_ARC_SEAL) {
 		if (!rspamd_dkim_parse_hdrlist_common (&nctx->common, headers,
-				strlen (headers),
+				strlen (headers), TRUE,
 				err)) {
 			return NULL;
 		}
@@ -2575,6 +2760,8 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 	gchar *b64_data;
 	guchar *rsa_buf;
 	guint rsa_len;
+	guint headers_len = 0, cur_len = 0;
+	union rspamd_dkim_header_stat hstat;
 
 	g_assert (ctx != NULL);
 
@@ -2648,14 +2835,62 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 	for (i = 0; i < ctx->common.hlist->len; i++) {
 		dh = g_ptr_array_index (ctx->common.hlist, i);
 
-		if (g_hash_table_lookup (task->raw_headers, dh->name)) {
-			rspamd_dkim_canonize_header (&ctx->common, task, dh->name, dh->count,
-					NULL, NULL);
+		/* We allow oversigning if dh->count > number of headers with this name */
+		hstat.n = GPOINTER_TO_UINT (g_hash_table_lookup (ctx->common.htable, dh->name));
 
-			for (j = 0; j < dh->count + 1; j++) {
+		if (hstat.s.flags & RSPAMD_DKIM_FLAG_OVERSIGN) {
+			/* Do oversigning */
+			GPtrArray *ar;
+			guint count = 0;
+
+			ar = g_hash_table_lookup (task->raw_headers, dh->name);
+
+			if (ar) {
+				count = ar->len;
+			}
+
+			for (j = 0; j < count; j ++) {
+				/* Sign all existing headers */
+				rspamd_dkim_canonize_header (&ctx->common, task, dh->name, j,
+						NULL, NULL);
+			}
+
+			/* Now add one more entry to oversign */
+			cur_len = (strlen (dh->name) + 1) * (count + 1);
+			headers_len += cur_len;
+			if (headers_len > 70 && i > 0 && i < ctx->common.hlist->len - 1) {
+				rspamd_printf_gstring (hdr, "  ");
+				headers_len = cur_len;
+			}
+
+			for (j = 0; j < count + 1; j++) {
 				rspamd_printf_gstring (hdr, "%s:", dh->name);
 			}
 		}
+		else {
+			if (g_hash_table_lookup (task->raw_headers, dh->name)) {
+				if (hstat.s.count > 0) {
+
+					cur_len = (strlen (dh->name) + 1) * (hstat.s.count);
+					headers_len += cur_len;
+					if (headers_len > 70 && i > 0 && i < ctx->common.hlist->len - 1) {
+						rspamd_printf_gstring (hdr, "  ");
+						headers_len = cur_len;
+					}
+
+					for (j = 0; j < hstat.s.count; j++) {
+						rspamd_printf_gstring (hdr, "%s:", dh->name);
+					}
+				}
+
+
+				rspamd_dkim_canonize_header (&ctx->common, task,
+						dh->name, dh->count,
+						NULL, NULL);
+			}
+		}
+
+		g_hash_table_remove (ctx->common.htable, dh->name);
 	}
 
 	/* Replace the last ':' with ';' */
@@ -2680,6 +2915,7 @@ rspamd_dkim_sign (struct rspamd_task *task, const gchar *selector,
 
 	switch (ctx->common.type) {
 	case RSPAMD_DKIM_NORMAL:
+	default:
 		hname = RSPAMD_DKIM_SIGNHEADER;
 		break;
 	case RSPAMD_DKIM_ARC_SIG:
