@@ -22,12 +22,14 @@
 #include "html_colors.h"
 #include "url.h"
 #include <unicode/uversion.h>
+#include <unicode/ucnv.h>
 #if U_ICU_VERSION_MAJOR_NUM >= 46
 #include <unicode/uidna.h>
 #endif
 
 static sig_atomic_t tags_sorted = 0;
 static sig_atomic_t entities_sorted = 0;
+static const guint max_tags = 8192; /* Ignore tags if this maximum is reached */
 
 struct html_tag_def {
 	const gchar *name;
@@ -36,11 +38,13 @@ struct html_tag_def {
 	guint flags;
 };
 
-#define msg_debug_html(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        "html", pool->tag.uid, \
+#define msg_debug_html(...)  rspamd_conditional_debug_fast (NULL, NULL, \
+        rspamd_html_log_id, "html", pool->tag.uid, \
         G_STRFUNC, \
         __VA_ARGS__)
 
+INIT_LOG_MODULE(html)
+
 #define TAG_DEF(id, name, flags) {(name), (id), (sizeof(name) - 1), (flags)}
 
 static struct html_tag_def tag_defs[] = {
@@ -844,7 +848,7 @@ rspamd_html_url_is_phished (rspamd_mempool_t *pool,
 
 	if (end > url_text + 4 &&
 			rspamd_url_find (pool, url_text, end - url_text, &url_str, FALSE,
-					&url_pos) &&
+					&url_pos, NULL) &&
 			url_str != NULL) {
 		if (url_pos > 0) {
 			/*
@@ -997,29 +1001,40 @@ rspamd_html_process_tag (rspamd_mempool_t *pool, struct html_content *hc,
 				nnode);
 	}
 
+	if (hc->total_tags > max_tags) {
+		hc->flags |= RSPAMD_HTML_FLAG_TOO_MANY_TAGS;
+	}
+
+	if (tag->id == -1) {
+		/* Ignore unknown tags */
+		hc->total_tags ++;
+		return FALSE;
+	}
+
 	tag->parent = *cur_level;
 
 	if (!(tag->flags & CM_INLINE)) {
 		/* Block tag */
-		nnode = g_node_new (tag);
-
 		if (tag->flags & (FL_CLOSING|FL_CLOSED)) {
 			if (!*cur_level) {
 				msg_debug_html ("bad parent node");
-				g_node_destroy (nnode);
 				return FALSE;
 			}
 
-			g_node_append (*cur_level, nnode);
+			if (hc->total_tags < max_tags) {
+				nnode = g_node_new (tag);
+				g_node_append (*cur_level, nnode);
 
-			if (!rspamd_html_check_balance (nnode, cur_level)) {
-				msg_debug_html (
-						"mark part as unbalanced as it has not pairable closing tags");
-				hc->flags |= RSPAMD_HTML_FLAG_UNBALANCED;
-				*balanced = FALSE;
-			}
-			else {
-				*balanced = TRUE;
+				if (!rspamd_html_check_balance (nnode, cur_level)) {
+					msg_debug_html (
+							"mark part as unbalanced as it has not pairable closing tags");
+					hc->flags |= RSPAMD_HTML_FLAG_UNBALANCED;
+					*balanced = FALSE;
+				} else {
+					*balanced = TRUE;
+				}
+
+				hc->total_tags ++;
 			}
 		}
 		else {
@@ -1038,8 +1053,13 @@ rspamd_html_process_tag (rspamd_mempool_t *pool, struct html_content *hc,
 						hc->flags |= RSPAMD_HTML_FLAG_UNBALANCED;
 						*balanced = FALSE;
 						tag->parent = parent->parent;
-						g_node_append (parent->parent, nnode);
-						*cur_level = nnode;
+
+						if (hc->total_tags < max_tags) {
+							nnode = g_node_new (tag);
+							g_node_append (parent->parent, nnode);
+							*cur_level = nnode;
+							hc->total_tags ++;
+						}
 
 						return TRUE;
 					}
@@ -1048,10 +1068,15 @@ rspamd_html_process_tag (rspamd_mempool_t *pool, struct html_content *hc,
 				parent->content_length += tag->content_length;
 			}
 
-			g_node_append (*cur_level, nnode);
+			if (hc->total_tags < max_tags) {
+				nnode = g_node_new (tag);
+				g_node_append (*cur_level, nnode);
 
-			if ((tag->flags & FL_CLOSED) == 0) {
-				*cur_level = nnode;
+				if ((tag->flags & FL_CLOSED) == 0) {
+					*cur_level = nnode;
+				}
+
+				hc->total_tags ++;
 			}
 
 			if (tag->flags & (CM_HEAD|CM_UNKNOWN|FL_IGNORE)) {
@@ -1178,6 +1203,8 @@ rspamd_html_parse_tag_content (rspamd_mempool_t *pool,
 		if (!g_ascii_isalpha (*in) && !g_ascii_isspace (*in)) {
 			hc->flags |= RSPAMD_HTML_FLAG_BAD_ELEMENTS;
 			state = ignore_bad_tag;
+			tag->id = -1;
+			tag->flags |= FL_BROKEN;
 		}
 		else if (g_ascii_isalpha (*in)) {
 			state = parse_name;
@@ -1197,6 +1224,7 @@ rspamd_html_parse_tag_content (rspamd_mempool_t *pool,
 
 			if (tag->name.len == 0) {
 				hc->flags |= RSPAMD_HTML_FLAG_BAD_ELEMENTS;
+				tag->id = -1;
 				tag->flags |= FL_BROKEN;
 				state = ignore_bad_tag;
 			}
@@ -1206,8 +1234,7 @@ rspamd_html_parse_tag_content (rspamd_mempool_t *pool,
 
 				s = rspamd_mempool_alloc (pool, tag->name.len);
 				memcpy (s, tag->name.start, tag->name.len);
-				tag->name.len = rspamd_html_decode_entitles_inplace (
-						s,
+				tag->name.len = rspamd_html_decode_entitles_inplace (s,
 						tag->name.len);
 				tag->name.start = s;
 
@@ -1443,6 +1470,8 @@ rspamd_html_parse_tag_content (rspamd_mempool_t *pool,
 	*statep = state;
 }
 
+
+
 struct rspamd_url *
 rspamd_html_process_url (rspamd_mempool_t *pool, const gchar *start, guint len,
 		struct html_tag_component *comp)
@@ -1506,8 +1535,15 @@ rspamd_html_process_url (rspamd_mempool_t *pool, const gchar *start, guint len,
 	d = decoded;
 
 	if (no_prefix) {
-		memcpy (d, "http://", sizeof ("http://") - 1);
-		d += sizeof ("http://") - 1;
+		if (s[0] == '/' && (len > 2 && s[1] == '/')) {
+			/* //bla case */
+			memcpy (d, "http:", sizeof ("http:") - 1);
+			d += sizeof ("http:") - 1;
+		}
+		else {
+			memcpy (d, "http://", sizeof ("http://") - 1);
+			d += sizeof ("http://") - 1;
+		}
 	}
 
 	/* We also need to remove all internal newlines and encode unsafe characters */
@@ -1528,15 +1564,25 @@ rspamd_html_process_url (rspamd_mempool_t *pool, const gchar *start, guint len,
 	}
 
 	*d = '\0';
+	dlen = d - decoded;
 
 	url = rspamd_mempool_alloc0 (pool, sizeof (*url));
-	rc = rspamd_url_parse (url, decoded, d - decoded, pool);
+
+	if (rspamd_normalise_unicode_inplace (pool, decoded, &dlen)) {
+		url->flags |= RSPAMD_URL_FLAG_UNNORMALISED;
+	}
+
+	rc = rspamd_url_parse (url, decoded, dlen, pool);
 
 	if (rc == URI_ERRNO_OK) {
 		if (has_bad_chars) {
 			url->flags |= RSPAMD_URL_FLAG_OBSCURED;
 		}
 
+		if (no_prefix) {
+			url->flags |= RSPAMD_URL_FLAG_SCHEMALESS;
+		}
+
 		decoded = url->string;
 		decoded_len = url->urllen;
 
@@ -1588,16 +1634,22 @@ rspamd_html_process_url_tag (rspamd_mempool_t *pool, struct html_tag *tag)
 
 static void
 rspamd_process_html_url (rspamd_mempool_t *pool, struct rspamd_url *url,
-		GHashTable *target)
+		GHashTable *tbl_urls, GHashTable *tbl_emails)
 {
+	GHashTable *target_tbl;
 	struct rspamd_url *query_url, *existing;
 	gchar *url_str;
 	gint rc;
+	gboolean prefix_added;
+
+	if (url->flags & RSPAMD_URL_FLAG_UNNORMALISED) {
+		url->flags |= RSPAMD_URL_FLAG_OBSCURED;
+	}
 
 	if (url->querylen > 0) {
 
 		if (rspamd_url_find (pool, url->query, url->querylen, &url_str, TRUE,
-				NULL)) {
+				NULL, &prefix_added)) {
 			query_url = rspamd_mempool_alloc0 (pool,
 					sizeof (struct rspamd_url));
 
@@ -1607,13 +1659,36 @@ rspamd_process_html_url (rspamd_mempool_t *pool, struct rspamd_url *url,
 					pool);
 
 			if (rc == URI_ERRNO_OK &&
-					url->hostlen > 0) {
+					query_url->hostlen > 0) {
 				msg_debug_html ("found url %s in query of url"
 						" %*s", url_str, url->querylen, url->query);
 
-				if ((existing = g_hash_table_lookup (target,
+				if (query_url->protocol == PROTOCOL_MAILTO) {
+					target_tbl = tbl_emails;
+				}
+				else {
+					target_tbl = tbl_urls;
+				}
+
+				if (prefix_added) {
+					query_url->flags |= RSPAMD_URL_FLAG_SCHEMALESS;
+				}
+
+				if (query_url->flags
+						& (RSPAMD_URL_FLAG_UNNORMALISED|RSPAMD_URL_FLAG_OBSCURED|
+							RSPAMD_URL_FLAG_NUMERIC)) {
+					/* Set obscured flag if query url is bad */
+					url->flags |= RSPAMD_URL_FLAG_OBSCURED;
+				}
+
+				/* And vice-versa */
+				if (url->flags & RSPAMD_URL_FLAG_OBSCURED) {
+					query_url->flags |= RSPAMD_URL_FLAG_OBSCURED;
+				}
+
+				if ((existing = g_hash_table_lookup (target_tbl,
 						query_url)) == NULL) {
-					g_hash_table_insert (target,
+					g_hash_table_insert (target_tbl,
 							query_url,
 							query_url);
 				}
@@ -2076,6 +2151,11 @@ rspamd_html_check_displayed_url (rspamd_mempool_t *pool,
 	gboolean url_found = FALSE;
 	struct rspamd_process_exception *ex;
 
+	if (href_offset <= 0) {
+		/* No dispalyed url, just some text within <a> tag */
+		return;
+	}
+
 	rspamd_html_url_is_phished (pool, url,
 			dest->data + href_offset,
 			dest->len - href_offset,
@@ -2425,6 +2505,8 @@ rspamd_html_process_part_full (rspamd_mempool_t *pool, struct html_content *hc,
 			/* TODO: parse DOCTYPE here */
 			if (t == '>') {
 				state = tag_end;
+				/* We don't know a lot about sgml tags, ignore them */
+				cur_tag = NULL;
 				continue;
 			}
 			p ++;
@@ -2525,7 +2607,7 @@ rspamd_html_process_part_full (rspamd_mempool_t *pool, struct html_content *hc,
 								if (turl == NULL && url != NULL) {
 									rspamd_process_html_url (pool,
 											url,
-											target_tbl);
+											urls, emails);
 								}
 							}
 
