@@ -24,8 +24,28 @@
 #include "unix-std.h"
 #include "protocol_internal.h"
 #include "libserver/mempool_vars_internal.h"
+#include "task.h"
 #include <math.h>
 
+INIT_LOG_MODULE(protocol)
+
+#define msg_err_protocol(...) rspamd_default_log_function (G_LOG_LEVEL_CRITICAL, \
+        "protocol", task->task_pool->tag.uid, \
+        G_STRFUNC, \
+        __VA_ARGS__)
+#define msg_warn_protocol(...)   rspamd_default_log_function (G_LOG_LEVEL_WARNING, \
+        "protocol", task->task_pool->tag.uid, \
+        G_STRFUNC, \
+        __VA_ARGS__)
+#define msg_info_protocol(...)   rspamd_default_log_function (G_LOG_LEVEL_INFO, \
+        "protocol", task->task_pool->tag.uid, \
+        G_STRFUNC, \
+        __VA_ARGS__)
+#define msg_debug_protocol(...)  rspamd_conditional_debug_fast (NULL, NULL, \
+        rspamd_protocol_log_id, "protocol", task->task_pool->tag.uid, \
+        G_STRFUNC, \
+        __VA_ARGS__)
+
 static GQuark
 rspamd_protocol_quark (void)
 {
@@ -174,7 +194,7 @@ rspamd_protocol_handle_url (struct rspamd_task *task,
 		goto err;
 	}
 
-	if (u.field_set & (1 << UF_QUERY)) {
+	if (u.field_set & (1u << UF_QUERY)) {
 		/* In case if we have a query, we need to store it somewhere */
 		query_args = rspamd_http_message_parse_query (msg);
 
@@ -187,7 +207,7 @@ rspamd_protocol_handle_url (struct rspamd_task *task,
 			/* Steal strings */
 			g_hash_table_iter_steal (&it);
 			rspamd_task_add_request_header (task, key, value);
-			msg_debug_task ("added header \"%T\" -> \"%T\" from HTTP query",
+			msg_debug_protocol ("added header \"%T\" -> \"%T\" from HTTP query",
 					key, value);
 		}
 
@@ -197,8 +217,7 @@ rspamd_protocol_handle_url (struct rspamd_task *task,
 	return TRUE;
 
 err:
-	g_set_error (&task->err, rspamd_protocol_quark(), 400, "invalid command: %.*s",
-			(gint)pathlen, p);
+	g_set_error (&task->err, rspamd_protocol_quark(), 400, "invalid command");
 
 	return FALSE;
 }
@@ -220,8 +239,8 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 
 	HASH_ITER (hh, msg->headers, header, htmp) {
 		DL_FOREACH (header, h) {
-			hn = rspamd_fstring_new_init (h->name->begin, h->name->len);
-			hv = rspamd_fstring_new_init (h->value->begin, h->value->len);
+			hn = rspamd_fstring_new_init (h->name.begin, h->name.len);
+			hv = rspamd_fstring_new_init (h->value.begin, h->value.len);
 			hn_tok = rspamd_ftok_map (hn);
 			hv_tok = rspamd_ftok_map (hv);
 
@@ -230,23 +249,23 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 			case 'D':
 				IF_HEADER (DELIVER_TO_HEADER) {
 					task->deliver_to = rspamd_protocol_escape_braces (task, hv);
-					debug_task ("read deliver-to header, value: %s",
+					msg_debug_protocol ("read deliver-to header, value: %s",
 							task->deliver_to);
 				}
 				else {
-					debug_task ("wrong header: %V", hn);
+					msg_debug_protocol ("wrong header: %V", hn);
 				}
 				break;
 			case 'h':
 			case 'H':
 				IF_HEADER (HELO_HEADER) {
 					task->helo = rspamd_mempool_ftokdup (task->task_pool, hv_tok);
-					debug_task ("read helo header, value: %s", task->helo);
+					msg_debug_protocol ("read helo header, value: %s", task->helo);
 				}
 				IF_HEADER (HOSTNAME_HEADER) {
 					task->hostname = rspamd_mempool_ftokdup (task->task_pool,
 							hv_tok);
-					debug_task ("read hostname header, value: %s", task->hostname);
+					msg_debug_protocol ("read hostname header, value: %s", task->hostname);
 				}
 				break;
 			case 'f':
@@ -254,21 +273,23 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 				IF_HEADER (FROM_HEADER) {
 					task->from_envelope = rspamd_email_address_from_smtp (hv->str,
 							hv->len);
-					debug_task ("read from header, value: %V", hv);
+					msg_debug_protocol ("read from header, value: %V", hv);
 
 					if (!task->from_envelope) {
-						msg_err_task ("bad from header: '%V'", hv);
+						msg_err_protocol ("bad from header: '%V'", hv);
 						task->flags |= RSPAMD_TASK_FLAG_BROKEN_HEADERS;
 					}
 				}
-				else {
-					debug_task ("wrong header: %V", hn);
+				IF_HEADER (FILENAME_HEADER) {
+					task->msg.fpath = rspamd_mempool_ftokdup (task->task_pool,
+							hv_tok);
+					msg_debug_protocol ("read filename header, value: %s", task->msg.fpath);
 				}
 				break;
 			case 'j':
 			case 'J':
 				IF_HEADER (JSON_HEADER) {
-					debug_task ("read json header, value: %V", hv);
+					msg_debug_protocol ("read json header, value: %V", hv);
 					fl = rspamd_config_parse_flag (hv->str, hv->len);
 					if (fl) {
 						task->flags |= RSPAMD_TASK_FLAG_JSON;
@@ -278,7 +299,7 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 					}
 				}
 				else {
-					debug_task ("wrong header: %V", hn);
+					msg_debug_protocol ("wrong header: %V", hn);
 				}
 				break;
 			case 'q':
@@ -286,46 +307,67 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 				IF_HEADER (QUEUE_ID_HEADER) {
 					task->queue_id = rspamd_mempool_ftokdup (task->task_pool,
 							hv_tok);
-					debug_task ("read queue_id header, value: %s", task->queue_id);
+					msg_debug_protocol ("read queue_id header, value: %s", task->queue_id);
 				}
 				else {
-					debug_task ("wrong header: %V", hn);
+					msg_debug_protocol ("wrong header: %V", hn);
 				}
 				break;
 			case 'r':
 			case 'R':
 				IF_HEADER (RCPT_HEADER) {
-					addr = rspamd_email_address_from_smtp (hv->str, hv->len);
+					const gchar *p, *end;
+					gsize cur_len;
 
-					if (addr) {
-						if (task->rcpt_envelope == NULL) {
-							task->rcpt_envelope = g_ptr_array_sized_new (2);
+					p = hv->str;
+					end = p + hv->len;
+
+					while (p < end) {
+						cur_len = rspamd_memcspn (p, ",", end - p);
+
+						if (cur_len > 0) {
+							addr = rspamd_email_address_from_smtp (p, cur_len);
+
+							if (addr) {
+								if (task->rcpt_envelope == NULL) {
+									task->rcpt_envelope = g_ptr_array_sized_new (
+											2);
+								}
+
+								g_ptr_array_add (task->rcpt_envelope, addr);
+							} else {
+								msg_err_protocol ("bad rcpt header: '%T'",
+										&h->value);
+								task->flags |= RSPAMD_TASK_FLAG_BROKEN_HEADERS;
+							}
+
+							p += cur_len;
 						}
 
-						g_ptr_array_add (task->rcpt_envelope, addr);
+						while (p < end && *p == ',') {
+							p ++;
+						}
 					}
-					else {
-						msg_err_task ("bad rcpt header: '%T'", h->value);
-						task->flags |= RSPAMD_TASK_FLAG_BROKEN_HEADERS;
-					}
-					debug_task ("read rcpt header, value: %V", hv);
+
+					msg_debug_protocol ("read rcpt header, value: %V", hv);
 				}
 				else {
-					debug_task ("wrong header: %V", hn);
+					msg_debug_protocol ("wrong header: %V", hn);
 				}
 				break;
 			case 'i':
 			case 'I':
 				IF_HEADER (IP_ADDR_HEADER) {
 					if (!rspamd_parse_inet_address (&task->from_addr, hv->str, hv->len)) {
-						msg_err_task ("bad ip header: '%V'", hv);
-						return FALSE;
+						msg_err_protocol ("bad ip header: '%V'", hv);
+					}
+					else {
+						msg_debug_protocol ("read IP header, value: %V", hv);
+						has_ip = TRUE;
 					}
-					debug_task ("read IP header, value: %V", hv);
-					has_ip = TRUE;
 				}
 				else {
-					debug_task ("wrong header: %V", hn);
+					msg_debug_protocol ("wrong header: %V", hn);
 				}
 				break;
 			case 'p':
@@ -334,29 +376,29 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 					srch.begin = "all";
 					srch.len = 3;
 
-					debug_task ("read pass header, value: %V", hv);
+					msg_debug_protocol ("read pass header, value: %V", hv);
 
 					if (rspamd_ftok_casecmp (hv_tok, &srch) == 0) {
 						task->flags |= RSPAMD_TASK_FLAG_PASS_ALL;
-						debug_task ("pass all filters");
+						msg_debug_protocol ("pass all filters");
 					}
 				}
 				IF_HEADER (PROFILE_HEADER) {
-					debug_task ("read profile header, value: %V", hv);
+					msg_debug_protocol ("read profile header, value: %V", hv);
 					task->flags |= RSPAMD_TASK_FLAG_PROFILE;
 				}
 				break;
 			case 's':
 			case 'S':
 				IF_HEADER (SUBJECT_HEADER) {
-					debug_task ("read subject header, value: %V", hv);
+					msg_debug_protocol ("read subject header, value: %V", hv);
 					task->subject = rspamd_mempool_ftokdup (task->task_pool, hv_tok);
 				}
 				IF_HEADER (SETTINGS_ID_HEADER) {
 					guint64 h;
 					guint32 *hp;
 
-					debug_task ("read settings-id header, value: %V", hv);
+					msg_debug_protocol ("read settings-id header, value: %V", hv);
 					h = rspamd_cryptobox_fast_hash_specific (RSPAMD_CRYPTOBOX_XXHASH64,
 							hv_tok->begin, hv_tok->len, 0xdeadbabe);
 					hp = rspamd_mempool_alloc (task->task_pool, sizeof (*hp));
@@ -373,27 +415,27 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 					 * We must ignore User header in case of spamc, as SA has
 					 * different meaning of this header
 					 */
-					debug_task ("read user header, value: %V", hv);
+					msg_debug_protocol ("read user header, value: %V", hv);
 					if (!RSPAMD_TASK_IS_SPAMC (task)) {
 						task->user = rspamd_mempool_ftokdup (task->task_pool,
 								hv_tok);
 					}
 					else {
-						msg_info_task ("ignore user header: legacy SA protocol");
+						msg_info_protocol ("ignore user header: legacy SA protocol");
 					}
 				}
 				IF_HEADER (URLS_HEADER) {
 					srch.begin = "extended";
 					srch.len = 8;
 
-					debug_task ("read urls header, value: %V", hv);
+					msg_debug_protocol ("read urls header, value: %V", hv);
 					if (rspamd_ftok_casecmp (hv_tok, &srch) == 0) {
 						task->flags |= RSPAMD_TASK_FLAG_EXT_URLS;
-						debug_task ("extended urls information");
+						msg_debug_protocol ("extended urls information");
 					}
 				}
 				IF_HEADER (USER_AGENT_HEADER) {
-					debug_task ("read user-agent header, value: %V", hv);
+					msg_debug_protocol ("read user-agent header, value: %V", hv);
 
 					if (hv_tok->len == 6 &&
 							rspamd_lc_cmp (hv_tok->begin, "rspamc", 6) == 0) {
@@ -404,7 +446,7 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 			case 'l':
 			case 'L':
 				IF_HEADER (NO_LOG_HEADER) {
-					debug_task ("read log header, value: %V", hv);
+					msg_debug_protocol ("read log header, value: %V", hv);
 					srch.begin = "no";
 					srch.len = 2;
 
@@ -416,11 +458,11 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 			case 'm':
 			case 'M':
 				IF_HEADER (MLEN_HEADER) {
-					debug_task ("read message length header, value: %V", hv);
+					msg_debug_protocol ("read message length header, value: %V", hv);
 					if (!rspamd_strtoul (hv_tok->begin,
 							hv_tok->len,
 							&task->message_len)) {
-						msg_err_task ("Invalid message length header: %V", hv);
+						msg_err_protocol ("Invalid message length header: %V", hv);
 					}
 					else {
 						task->flags |= RSPAMD_TASK_FLAG_HAS_CONTROL;
@@ -432,7 +474,7 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 					rspamd_mempool_set_variable (task->task_pool,
 							RSPAMD_MEMPOOL_MTA_TAG,
 							mta_tag, NULL);
-					debug_task ("read MTA-Tag header, value: %s", mta_tag);
+					msg_debug_protocol ("read MTA-Tag header, value: %s", mta_tag);
 				}
 				IF_HEADER (MTA_NAME_HEADER) {
 					gchar *mta_name;
@@ -440,15 +482,22 @@ rspamd_protocol_handle_headers (struct rspamd_task *task,
 					rspamd_mempool_set_variable (task->task_pool,
 							RSPAMD_MEMPOOL_MTA_NAME,
 							mta_name, NULL);
-					debug_task ("read MTA-Name header, value: %s", mta_name);
+					msg_debug_protocol ("read MTA-Name header, value: %s", mta_name);
 				}
 				IF_HEADER (MILTER_HEADER) {
 					task->flags |= RSPAMD_TASK_FLAG_MILTER;
-					debug_task ("read Milter header, value: %V", hv);
+					msg_debug_protocol ("read Milter header, value: %V", hv);
+				}
+				break;
+			case 't':
+			case 'T':
+				IF_HEADER (TLS_CIPHER_HEADER) {
+					task->flags |= RSPAMD_TASK_FLAG_SSL;
+					msg_debug_protocol ("read TLS cipher header, value: %V", hv);
 				}
 				break;
 			default:
-				debug_task ("unknown header: %V", hn);
+				msg_debug_protocol ("generic header: %V", hn);
 				break;
 			}
 
@@ -567,7 +616,7 @@ rspamd_protocol_handle_control (struct rspamd_task *task,
 
 	if (!rspamd_rcl_parse (control_parser, task->cfg, task, task->task_pool,
 			control, &err)) {
-		msg_warn_task ("cannot parse control block: %e", err);
+		msg_warn_protocol ("cannot parse control block: %e", err);
 		g_error_free (err);
 
 		return FALSE;
@@ -583,11 +632,11 @@ rspamd_protocol_handle_request (struct rspamd_task *task,
 	gboolean ret = TRUE;
 
 	if (msg->method == HTTP_SYMBOLS) {
-		task->cmd = CMD_SYMBOLS;
+		task->cmd = CMD_CHECK_V2;
 		task->flags &= ~RSPAMD_TASK_FLAG_JSON;
 	}
 	else if (msg->method == HTTP_CHECK) {
-		task->cmd = CMD_CHECK;
+		task->cmd = CMD_CHECK_V2;
 		task->flags &= ~RSPAMD_TASK_FLAG_JSON;
 	}
 	else {
@@ -683,7 +732,7 @@ urls_protocol_cb (gpointer key, gpointer value, gpointer ud)
 			len = task->from_envelope->addr_len;
 		}
 
-		msg_info_task_encrypted ("<%s> %s: %*s; ip: %s; URL: %*s",
+		msg_notice_task_encrypted ("<%s> %s: %*s; ip: %s; URL: %*s",
 			task->message_id,
 			has_user ? "user" : "from",
 			len, user_field,
@@ -740,7 +789,7 @@ rspamd_emails_tree_ucl (GHashTable *input, struct rspamd_task *task)
 
 /* Write new subject */
 static const gchar *
-make_rewritten_subject (struct rspamd_metric *metric, struct rspamd_task *task)
+rspamd_protocol_rewrite_subject (struct rspamd_task *task)
 {
 	GString *subj_buf;
 	gchar *res;
@@ -750,7 +799,7 @@ make_rewritten_subject (struct rspamd_metric *metric, struct rspamd_task *task)
 	c = rspamd_mempool_get_variable (task->task_pool, "metric_subject");
 
 	if (c == NULL) {
-		c = metric->subject;
+		c = task->cfg->subject;
 	}
 
 	if (c == NULL) {
@@ -767,19 +816,36 @@ make_rewritten_subject (struct rspamd_metric *metric, struct rspamd_task *task)
 	subj_buf = g_string_sized_new (strlen (c) + slen);
 
 	while (*p) {
-		if (*p == '%' && *(p + 1) == 's') {
-			g_string_append_len (subj_buf, c, p - c);
+		if (*p == '%') {
+			switch (p[1]) {
+			case 's':
+				g_string_append_len (subj_buf, c, p - c);
 
-			if (s) {
-				g_string_append_len (subj_buf, s, slen);
+				if (s) {
+					g_string_append_len (subj_buf, s, slen);
+				}
+				c = p + 2;
+				p += 2;
+				break;
+			case 'd':
+				g_string_append_len (subj_buf, c, p - c);
+				rspamd_printf_gstring (subj_buf, "%.2f", task->result->score);
+				c = p + 2;
+				p += 2;
+				break;
+			case '%':
+				g_string_append_len (subj_buf, c, p - c);
+				g_string_append_c (subj_buf, '%');
+				c = p + 2;
+				p += 2;
+			default:
+				p ++; /* Just % something unknown */
+				break;
 			}
-
-			p += 2;
-			c = p;
-			continue;
 		}
-
-		p ++;
+		else {
+			p++;
+		}
 	}
 
 	if (p > c) {
@@ -797,29 +863,7 @@ make_rewritten_subject (struct rspamd_metric *metric, struct rspamd_task *task)
 }
 
 static ucl_object_t *
-rspamd_str_hash_ucl (GHashTable *ht)
-{
-	GHashTableIter it;
-	gpointer k, v;
-	ucl_object_t *top = NULL, *obj;
-
-	top = ucl_object_typed_new (UCL_ARRAY);
-
-	if (ht) {
-		g_hash_table_iter_init (&it, ht);
-
-		while (g_hash_table_iter_next (&it, &k, &v)) {
-			obj = ucl_object_fromstring ((const char *)v);
-			ucl_array_append (top, obj);
-		}
-	}
-
-	return top;
-}
-
-static ucl_object_t *
-rspamd_metric_symbol_ucl (struct rspamd_task *task, struct rspamd_metric *m,
-	struct rspamd_symbol_result *sym)
+rspamd_metric_symbol_ucl (struct rspamd_task *task, struct rspamd_symbol_result *sym)
 {
 	ucl_object_t *obj = NULL, *ar;
 	const gchar *description = NULL;
@@ -870,20 +914,13 @@ rspamd_metric_result_ucl (struct rspamd_task *task,
 {
 	GHashTableIter hiter;
 	struct rspamd_symbol_result *sym;
-	struct rspamd_metric *m;
 	gboolean is_spam;
-	enum rspamd_metric_action action = METRIC_ACTION_NOACTION;
+	enum rspamd_action_type action = METRIC_ACTION_NOACTION;
 	ucl_object_t *obj = NULL, *sobj;
 	gpointer h, v;
 	const gchar *subject;
 
-	m = mres->metric;
-
-	if (mres->action == METRIC_ACTION_MAX) {
-		mres->action = rspamd_check_action_metric (task, mres);
-	}
-
-	action = mres->action;
+	action = rspamd_check_action_metric (task, mres);
 	is_spam = (action < METRIC_ACTION_GREYLIST);
 
 	if (task->cmd != CMD_CHECK_V2) {
@@ -916,7 +953,7 @@ rspamd_metric_result_ucl (struct rspamd_task *task,
 			"action", 0, false);
 
 	if (action == METRIC_ACTION_REWRITE_SUBJECT) {
-		subject = make_rewritten_subject (m, task);
+		subject = rspamd_protocol_rewrite_subject (task);
 
 		if (subject) {
 			ucl_object_insert_key (obj, ucl_object_fromstring (subject),
@@ -933,7 +970,7 @@ rspamd_metric_result_ucl (struct rspamd_task *task,
 
 	while (g_hash_table_iter_next (&hiter, &h, &v)) {
 		sym = (struct rspamd_symbol_result *)v;
-		sobj = rspamd_metric_symbol_ucl (task, m, sym);
+		sobj = rspamd_metric_symbol_ucl (task, sym);
 		ucl_object_insert_key (obj, sobj, h, 0, false);
 	}
 
@@ -951,28 +988,35 @@ void
 rspamd_ucl_torspamc_output (const ucl_object_t *top,
 	rspamd_fstring_t **out)
 {
-	const ucl_object_t *metric, *score,
+	const ucl_object_t *symbols, *score,
 	*required_score, *is_spam, *elt, *cur;
 	ucl_object_iter_t iter = NULL;
 
-	metric = ucl_object_lookup (top, DEFAULT_METRIC);
-	if (metric != NULL) {
-		score = ucl_object_lookup (metric, "score");
-		required_score = ucl_object_lookup (metric, "required_score");
-		is_spam = ucl_object_lookup (metric, "is_spam");
-		rspamd_printf_fstring (out,
+	score = ucl_object_lookup (top, "score");
+	required_score = ucl_object_lookup (top, "required_score");
+	is_spam = ucl_object_lookup (top, "is_spam");
+	rspamd_printf_fstring (out,
 			"Metric: default; %s; %.2f / %.2f / 0.0\r\n",
 			ucl_object_toboolean (is_spam) ? "True" : "False",
 			ucl_object_todouble (score),
 			ucl_object_todouble (required_score));
-		elt = ucl_object_lookup (metric, "action");
-		if (elt != NULL) {
-			rspamd_printf_fstring (out, "Action: %s\r\n",
+	elt = ucl_object_lookup (top, "action");
+	if (elt != NULL) {
+		rspamd_printf_fstring (out, "Action: %s\r\n",
 				ucl_object_tostring (elt));
-		}
+	}
 
+	elt = ucl_object_lookup (top, "subject");
+	if (elt != NULL) {
+		rspamd_printf_fstring (out, "Subject: %s\r\n",
+				ucl_object_tostring (elt));
+	}
+
+	symbols = ucl_object_lookup (top, "symbols");
+
+	if (symbols != NULL) {
 		iter = NULL;
-		while ((elt = ucl_object_iterate (metric, &iter, true)) != NULL) {
+		while ((elt = ucl_object_iterate (symbols, &iter, true)) != NULL) {
 			if (elt->type == UCL_OBJECT) {
 				const ucl_object_t *sym_score;
 				sym_score = ucl_object_lookup (elt, "score");
@@ -981,12 +1025,6 @@ rspamd_ucl_torspamc_output (const ucl_object_t *top,
 					ucl_object_todouble (sym_score));
 			}
 		}
-
-		elt = ucl_object_lookup (metric, "subject");
-		if (elt != NULL) {
-			rspamd_printf_fstring (out, "Subject: %s\r\n",
-				ucl_object_tostring (elt));
-		}
 	}
 
 	elt = ucl_object_lookup (top, "messages");
@@ -1011,23 +1049,24 @@ void
 rspamd_ucl_tospamc_output (const ucl_object_t *top,
 	rspamd_fstring_t **out)
 {
-	const ucl_object_t *metric, *score,
+	const ucl_object_t *symbols, *score,
 		*required_score, *is_spam, *elt;
 	ucl_object_iter_t iter = NULL;
 	rspamd_fstring_t *f;
 
-	metric = ucl_object_lookup (top, DEFAULT_METRIC);
-	if (metric != NULL) {
-		score = ucl_object_lookup (metric, "score");
-		required_score = ucl_object_lookup (metric, "required_score");
-		is_spam = ucl_object_lookup (metric, "is_spam");
-		rspamd_printf_fstring (out,
+	score = ucl_object_lookup (top, "score");
+	required_score = ucl_object_lookup (top, "required_score");
+	is_spam = ucl_object_lookup (top, "is_spam");
+	rspamd_printf_fstring (out,
 			"Spam: %s ; %.2f / %.2f\r\n\r\n",
 			ucl_object_toboolean (is_spam) ? "True" : "False",
 			ucl_object_todouble (score),
 			ucl_object_todouble (required_score));
 
-		while ((elt = ucl_object_iterate (metric, &iter, true)) != NULL) {
+	symbols = ucl_object_lookup (top, "symbols");
+
+	if (symbols != NULL) {
+		while ((elt = ucl_object_iterate (symbols, &iter, true)) != NULL) {
 			if (elt->type == UCL_OBJECT) {
 				rspamd_printf_fstring (out, "%s,",
 					ucl_object_key (elt));
@@ -1069,19 +1108,6 @@ rspamd_protocol_output_profiling (struct rspamd_task *task,
 	ucl_object_insert_key (top, prof, "profile", 0, false);
 }
 
-struct rspamd_saved_protocol_reply {
-	ucl_object_t *obj;
-	enum rspamd_protocol_flags flags;
-};
-
-static void
-rspamd_protocol_cached_dtor (gpointer p)
-{
-	struct rspamd_saved_protocol_reply *cached = p;
-
-	ucl_object_unref (cached->obj);
-}
-
 ucl_object_t *
 rspamd_protocol_write_ucl (struct rspamd_task *task,
 		enum rspamd_protocol_flags flags)
@@ -1089,35 +1115,12 @@ rspamd_protocol_write_ucl (struct rspamd_task *task,
 	ucl_object_t *top = NULL;
 	GString *dkim_sig;
 	const ucl_object_t *milter_reply;
-	struct rspamd_saved_protocol_reply *cached;
-	static const gchar *varname = RSPAMD_MEMPOOL_CACHED_REPLY;
 
-	/* Check for cached reply */
-	cached = rspamd_mempool_get_variable (task->task_pool, varname);
+	rspamd_task_set_finish_time (task);
+	top = ucl_object_typed_new (UCL_OBJECT);
 
-	if (cached) {
-		top = cached->obj;
-		/*
-		 * Update flags: we don't need to set more flags than we need,
-		 * so just xor previous flags and current flags and then or them
-		 * to the cached one
-		 */
-		flags ^= cached->flags;
-		cached->flags |= flags;
-
-	}
-	else {
-		top = ucl_object_typed_new (UCL_OBJECT);
-		cached = rspamd_mempool_alloc (task->task_pool, sizeof (*cached));
-		cached->obj = top;
-		cached->flags = flags;
-		rspamd_mempool_set_variable (task->task_pool, varname,
-				cached, rspamd_protocol_cached_dtor);
-
-		/* We also set scan time here */
-		task->time_real_finish = rspamd_get_ticks ();
-		task->time_virtual_finish = rspamd_get_virtual_ticks ();
-	}
+	rspamd_mempool_add_destructor (task->task_pool,
+			(rspamd_mempool_destruct_t)ucl_object_unref, top);
 
 	if (flags & RSPAMD_PROTOCOL_METRICS) {
 		rspamd_metric_result_ucl (task, task->result, top);
@@ -1184,11 +1187,11 @@ rspamd_protocol_write_ucl (struct rspamd_task *task,
 
 			if (task->flags & RSPAMD_TASK_FLAG_MILTER) {
 				folded_header = rspamd_header_value_fold ("DKIM-Signature",
-						dkim_sig->str, 80, RSPAMD_TASK_NEWLINES_LF);
+						dkim_sig->str, 80, RSPAMD_TASK_NEWLINES_LF, NULL);
 			}
 			else {
 				folded_header = rspamd_header_value_fold ("DKIM-Signature",
-						dkim_sig->str, 80, task->nlines_type);
+						dkim_sig->str, 80, task->nlines_type, NULL);
 			}
 			/*
 			 * According to milter docs, we need to be extra careful
@@ -1202,8 +1205,8 @@ rspamd_protocol_write_ucl (struct rspamd_task *task,
 			 * space (ASCII 0x20) or tab (ASCII 0x09, or \t in C).
 			 * The line feed should NOT be preceded by a carriage return (ASCII 0x0d);
 			 * the MTA will add this automatically.
-			 * It is the filter writer's responsibility to ensure that no s
-			 * tandards are violated.
+			 * It is the filter writer's responsibility to ensure that no
+			 * standards are violated.
 			 */
 			ucl_object_insert_key (top,
 					ucl_object_fromstring_common (folded_header->str,
@@ -1268,10 +1271,10 @@ rspamd_protocol_http_reply (struct rspamd_http_message *msg,
 
 	rspamd_task_write_log (task);
 
-	if (task->cfg->log_re_cache) {
+	if (task->cfg->log_flags & RSPAMD_LOG_FLAG_RE_CACHE) {
 		restat = rspamd_re_cache_get_stat (task->re_rt);
 		g_assert (restat != NULL);
-		msg_info_task (
+		msg_notice_task (
 				"regexp statistics: %ud pcre regexps scanned, %ud regexps matched,"
 				" %ud regexps total, %ud regexps cached,"
 				" %HL bytes scanned using pcre, %HL bytes scanned total",
@@ -1319,7 +1322,7 @@ rspamd_protocol_http_reply (struct rspamd_http_message *msg,
 			r = ZSTD_compressStream (zstream, &zout, &zin);
 
 			if (ZSTD_isError (r)) {
-				msg_err_task ("cannot compress: %s", ZSTD_getErrorName (r));
+				msg_err_protocol ("cannot compress: %s", ZSTD_getErrorName (r));
 				rspamd_fstring_free (compressed_reply);
 				rspamd_http_message_set_body_from_fstring_steal (msg, reply);
 
@@ -1327,17 +1330,18 @@ rspamd_protocol_http_reply (struct rspamd_http_message *msg,
 			}
 		}
 
+		ZSTD_flushStream (zstream, &zout);
 		r = ZSTD_endStream (zstream, &zout);
 
 		if (ZSTD_isError (r)) {
-			msg_err_task ("cannot finalize compress: %s", ZSTD_getErrorName (r));
+			msg_err_protocol ("cannot finalize compress: %s", ZSTD_getErrorName (r));
 			rspamd_fstring_free (compressed_reply);
 			rspamd_http_message_set_body_from_fstring_steal (msg, reply);
 
 			goto end;
 		}
 
-		msg_info_task ("writing compressed results: %z bytes before "
+		msg_info_protocol ("writing compressed results: %z bytes before "
 				"%z bytes after", zin.pos, zout.pos);
 		compressed_reply->len = zout.pos;
 		rspamd_fstring_free (reply);
@@ -1364,12 +1368,7 @@ end:
 
 		if (metric_res != NULL) {
 
-			if (metric_res->action != METRIC_ACTION_MAX) {
-				action = metric_res->action;
-			}
-			else {
-				action = rspamd_check_action_metric (task, metric_res);
-			}
+			action = rspamd_check_action_metric (task, metric_res);
 
 			if (action == METRIC_ACTION_SOFT_REJECT &&
 					(task->flags & RSPAMD_TASK_FLAG_GREYLISTED)) {
@@ -1451,9 +1450,9 @@ rspamd_protocol_write_log_pipe (struct rspamd_task *task)
 					 * -5: key (module name)
 					 * -6: global
 					 */
-					msg_debug_task ("calling for %s", lua_tostring (L, -3));
+					msg_debug_protocol ("calling for %s", lua_tostring (L, -3));
 					if (lua_pcall (L, 1, 1, 0) != 0) {
-						msg_info_task ("call to log callback %s failed: %s",
+						msg_info_protocol ("call to log callback %s failed: %s",
 								lua_tostring (L, -2), lua_tostring (L, -1));
 						lua_pop (L, 1);
 						/* stack:
@@ -1514,7 +1513,7 @@ rspamd_protocol_write_log_pipe (struct rspamd_task *task)
 							lua_pop (L, 1); /* Table result of pcall */
 						}
 						else {
-							msg_info_task ("call to log callback %s returned "
+							msg_info_protocol ("call to log callback %s returned "
 									"wrong type: %s",
 									lua_tostring (L, -2),
 									lua_typename (L, lua_type (L, -1)));
@@ -1554,7 +1553,7 @@ rspamd_protocol_write_log_pipe (struct rspamd_task *task)
 					sz = sizeof (*ls) +
 							sizeof (struct rspamd_protocol_log_symbol_result) *
 							(n + nextra);
-					ls = g_slice_alloc (sz);
+					ls = g_malloc0 (sz);
 
 					/* Handle settings id */
 					sid = rspamd_mempool_get_variable (task->task_pool,
@@ -1597,20 +1596,20 @@ rspamd_protocol_write_log_pipe (struct rspamd_task *task)
 				}
 				else {
 					sz = sizeof (*ls);
-					ls = g_slice_alloc0 (sz);
+					ls = g_malloc0 (sz);
 					ls->nresults = 0;
 				}
 
 				/* We don't really care about return value here */
 				if (write (lp->fd, ls, sz) == -1) {
-					msg_info_task ("cannot write to log pipe: %s",
+					msg_info_protocol ("cannot write to log pipe: %s",
 							strerror (errno));
 				}
 
-				g_slice_free1 (sz, ls);
+				g_free (ls);
 				break;
 			default:
-				msg_err_task ("unknown log format %d", lp->type);
+				msg_err_protocol ("unknown log format %d", lp->type);
 				break;
 			}
 		}
@@ -1629,7 +1628,7 @@ rspamd_protocol_write_reply (struct rspamd_task *task)
 	msg = rspamd_http_new_message (HTTP_RESPONSE);
 
 	if (rspamd_http_connection_is_encrypted (task->http_conn)) {
-		msg_info_task ("<%s> writing encrypted reply", task->message_id);
+		msg_info_protocol ("<%s> writing encrypted reply", task->message_id);
 	}
 
 	if (!RSPAMD_TASK_IS_JSON (task)) {
@@ -1642,7 +1641,7 @@ rspamd_protocol_write_reply (struct rspamd_task *task)
 
 	msg->date = time (NULL);
 
-	debug_task ("writing reply to client");
+	msg_debug_protocol ("writing reply to client");
 	if (task->err != NULL) {
 		ucl_object_t *top = NULL;
 
@@ -1679,7 +1678,7 @@ rspamd_protocol_write_reply (struct rspamd_task *task)
 			ctype = "text/plain";
 			break;
 		case CMD_OTHER:
-			msg_err_task ("BROKEN");
+			msg_err_protocol ("BROKEN");
 			break;
 		}
 	}
