@@ -52,15 +52,18 @@
         "re_cache", cache->hash, \
         G_STRFUNC, \
         __VA_ARGS__)
-#define msg_debug_re_cache(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        "re_cache", cache->hash, \
+
+#define msg_debug_re_task(...)  rspamd_conditional_debug_fast (NULL, NULL, \
+        rspamd_re_cache_log_id, "re_cache", task->task_pool->tag.uid, \
         G_STRFUNC, \
         __VA_ARGS__)
-#define msg_debug_re_task(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        "re_cache", task->task_pool->tag.uid, \
+#define msg_debug_re_cache(...)  rspamd_conditional_debug_fast (NULL, NULL, \
+        rspamd_re_cache_log_id, "re_cache", cache->hash, \
         G_STRFUNC, \
         __VA_ARGS__)
 
+INIT_LOG_MODULE(re_cache)
+
 #ifdef WITH_HYPERSCAN
 #define RSPAMD_HS_MAGIC_LEN (sizeof (rspamd_hs_magic))
 static const guchar rspamd_hs_magic[] = {'r', 's', 'h', 's', 'r', 'e', '1', '1'},
@@ -156,7 +159,7 @@ rspamd_re_cache_destroy (struct rspamd_re_cache *cache)
 		g_hash_table_unref (re_class->re);
 
 		if (re_class->type_data) {
-			g_slice_free1 (re_class->type_len, re_class->type_data);
+			g_free (re_class->type_data);
 		}
 
 #ifdef WITH_HYPERSCAN
@@ -170,12 +173,12 @@ rspamd_re_cache_destroy (struct rspamd_re_cache *cache)
 			g_free (re_class->hs_ids);
 		}
 #endif
-		g_slice_free1 (sizeof (*re_class), re_class);
+		g_free (re_class);
 	}
 
 	g_hash_table_unref (cache->re_classes);
 	g_ptr_array_free (cache->re, TRUE);
-	g_slice_free1 (sizeof (*cache), cache);
+	g_free (cache);
 }
 
 static void
@@ -184,7 +187,7 @@ rspamd_re_cache_elt_dtor (gpointer e)
 	struct rspamd_re_cache_elt *elt = e;
 
 	rspamd_regexp_unref (elt->re);
-	g_slice_free1 (sizeof (*elt), elt);
+	g_free (elt);
 }
 
 struct rspamd_re_cache *
@@ -192,7 +195,7 @@ rspamd_re_cache_new (void)
 {
 	struct rspamd_re_cache *cache;
 
-	cache = g_slice_alloc0 (sizeof (*cache));
+	cache = g_malloc0 (sizeof (*cache));
 	cache->re_classes = g_hash_table_new (g_int64_hash, g_int64_equal);
 	cache->nre = 0;
 	cache->re = g_ptr_array_new_full (256, rspamd_re_cache_elt_dtor);
@@ -232,7 +235,7 @@ rspamd_re_cache_add (struct rspamd_re_cache *cache, rspamd_regexp_t *re,
 	re_class = g_hash_table_lookup (cache->re_classes, &class_id);
 
 	if (re_class == NULL) {
-		re_class = g_slice_alloc0 (sizeof (*re_class));
+		re_class = g_malloc0 (sizeof (*re_class));
 		re_class->id = class_id;
 		re_class->type_len = datalen;
 		re_class->type = type;
@@ -240,7 +243,7 @@ rspamd_re_cache_add (struct rspamd_re_cache *cache, rspamd_regexp_t *re,
 				rspamd_regexp_equal, NULL, (GDestroyNotify)rspamd_regexp_unref);
 
 		if (datalen > 0) {
-			re_class->type_data = g_slice_alloc (datalen);
+			re_class->type_data = g_malloc0 (datalen);
 			memcpy (re_class->type_data, type_data, datalen);
 		}
 
@@ -252,7 +255,7 @@ rspamd_re_cache_add (struct rspamd_re_cache *cache, rspamd_regexp_t *re,
 		/*
 		 * We set re id based on the global position in the cache
 		 */
-		elt = g_slice_alloc0 (sizeof (*elt));
+		elt = g_malloc0 (sizeof (*elt));
 		/* One ref for re_class */
 		nre = rspamd_regexp_ref (re);
 		rspamd_regexp_set_cache_id (re, cache->nre++);
@@ -343,7 +346,7 @@ rspamd_re_cache_init (struct rspamd_re_cache *cache, struct rspamd_config *cfg)
 		rspamd_regexp_set_cache_id (re, i);
 
 		if (re_class->st == NULL) {
-			re_class->st = g_slice_alloc (sizeof (*re_class->st));
+			re_class->st = g_malloc (sizeof (*re_class->st));
 			rspamd_cryptobox_hash_init (re_class->st, NULL, 0);
 		}
 
@@ -405,7 +408,7 @@ rspamd_re_cache_init (struct rspamd_re_cache *cache, struct rspamd_config *cfg)
 			rspamd_cryptobox_hash_final (re_class->st, hash_out);
 			rspamd_snprintf (re_class->hash, sizeof (re_class->hash), "%*xs",
 					(gint) rspamd_cryptobox_HASHBYTES, hash_out);
-			g_slice_free1 (sizeof (*re_class->st), re_class->st);
+			g_free (re_class->st);
 			re_class->st = NULL;
 		}
 	}
@@ -443,7 +446,7 @@ rspamd_re_cache_init (struct rspamd_re_cache *cache, struct rspamd_config *cfg)
 
 	hs_set_allocator (g_malloc, g_free);
 
-	msg_info_re_cache ("loaded hyperscan engine witch cpu tune '%s' and features '%V'",
+	msg_info_re_cache ("loaded hyperscan engine with cpu tune '%s' and features '%V'",
 			platform, features);
 
 	rspamd_fstring_free (features);
@@ -456,11 +459,11 @@ rspamd_re_cache_runtime_new (struct rspamd_re_cache *cache)
 	struct rspamd_re_runtime *rt;
 	g_assert (cache != NULL);
 
-	rt = g_slice_alloc0 (sizeof (*rt));
+	rt = g_malloc0 (sizeof (*rt));
 	rt->cache = cache;
 	REF_RETAIN (cache);
-	rt->checked = g_slice_alloc0 (NBYTES (cache->nre));
-	rt->results = g_slice_alloc0 (cache->nre);
+	rt->checked = g_malloc0 (NBYTES (cache->nre));
+	rt->results = g_malloc0 (cache->nre);
 	rt->stat.regexp_total = cache->nre;
 #ifdef WITH_HYPERSCAN
 	rt->has_hs = cache->hyperscan_loaded;
@@ -488,7 +491,7 @@ rspamd_re_cache_process_pcre (struct rspamd_re_runtime *rt,
 	guint max_hits = rspamd_regexp_get_maxhits (re);
 	guint64 id = rspamd_regexp_get_cache_id (re);
 	gdouble t1, t2, pr;
-	const gdouble slow_time = 0.1;
+	const gdouble slow_time = 1e8;
 
 	if (in == NULL) {
 		return rt->results[id];
@@ -508,7 +511,7 @@ rspamd_re_cache_process_pcre (struct rspamd_re_runtime *rt,
 		pr = rspamd_random_double_fast ();
 
 		if (pr > 0.9) {
-			t1 = rspamd_get_ticks ();
+			t1 = rspamd_get_ticks (TRUE);
 		}
 
 		while (rspamd_regexp_search (re,
@@ -527,6 +530,7 @@ rspamd_re_cache_process_pcre (struct rspamd_re_runtime *rt,
 			}
 		}
 
+		rt->results[id] += r;
 		rt->stat.regexp_checked++;
 		rt->stat.bytes_scanned_pcre += len;
 		rt->stat.bytes_scanned += len;
@@ -536,10 +540,10 @@ rspamd_re_cache_process_pcre (struct rspamd_re_runtime *rt,
 		}
 
 		if (pr > 0.9) {
-			t2 = rspamd_get_ticks ();
+			t2 = rspamd_get_ticks (TRUE);
 
 			if (t2 - t1 > slow_time) {
-				msg_info_task ("regexp '%16s' took %.2f seconds to execute",
+				msg_info_task ("regexp '%16s' took %.0f ticks to execute",
 						rspamd_regexp_get_pattern (re), t2 - t1);
 			}
 		}
@@ -593,13 +597,12 @@ rspamd_re_cache_hyperscan_cb (unsigned int id,
 			processed = 0;
 
 			for (i = 0; i < cbdata->count; i ++) {
-				ret = rspamd_re_cache_process_pcre (rt,
+				rspamd_re_cache_process_pcre (rt,
 						pcre_elt->re,
 						cbdata->task,
 						cbdata->ins[i],
 						cbdata->lens[i],
 						FALSE);
-				rt->results[id] = ret;
 				setbit (rt->checked, id);
 
 				processed += cbdata->lens[i];
@@ -665,7 +668,6 @@ rspamd_re_cache_process_regexp_data (struct rspamd_re_runtime *rt,
 					in[i],
 					lens[i],
 					is_raw);
-			rt->results[re_id] += ret;
 		}
 
 		setbit (rt->checked, re_id);
@@ -810,9 +812,10 @@ rspamd_re_cache_exec_re (struct rspamd_task *task,
 
 			ret = rspamd_re_cache_process_regexp_data (rt, re,
 					task, scvec, lenvec, headerlist->len, raw);
-			msg_debug_re_task ("checking header %s regexp: %s -> %d",
+			msg_debug_re_task ("checking header %s regexp: %s=%*s -> %d",
 					re_class->type_data,
-					rspamd_regexp_get_pattern (re), ret);
+					rspamd_regexp_get_pattern (re),
+					(int)lenvec[0], scvec[0], ret);
 			g_free (scvec);
 			g_free (lenvec);
 		}
@@ -877,26 +880,34 @@ rspamd_re_cache_exec_re (struct rspamd_task *task,
 			for (i = 0; i < task->text_parts->len; i++) {
 				part = g_ptr_array_index (task->text_parts, i);
 
-				/* Skip empty parts */
-				if (IS_PART_EMPTY (part)) {
-					lenvec[i] = 0;
-					scvec[i] = (guchar *) "";
-					continue;
-				}
-
-				/* Check raw flags */
-				if (!IS_PART_UTF (part)) {
-					raw = TRUE;
-				}
 				/* Select data for regexp */
 				if (re_class->type == RSPAMD_RE_RAWMIME) {
-					in = part->raw.begin;
-					len = part->raw.len;
+					if (part->raw.len == 0) {
+						len = 0;
+						in = "";
+					}
+					else {
+						in = part->raw.begin;
+						len = part->raw.len;
+					}
+
 					raw = TRUE;
 				}
 				else {
-					in = part->content->data;
-					len = part->content->len;
+					/* Skip empty parts */
+					if (IS_PART_EMPTY (part)) {
+						len = 0;
+						in = "";
+					}
+					else {
+						/* Check raw flags */
+						if (!IS_PART_UTF (part)) {
+							raw = TRUE;
+						}
+
+						in = part->content->data;
+						len = part->content->len;
+					}
 				}
 
 				scvec[i] = (guchar *) in;
@@ -1136,10 +1147,10 @@ rspamd_re_cache_runtime_destroy (struct rspamd_re_runtime *rt)
 {
 	g_assert (rt != NULL);
 
-	g_slice_free1 (NBYTES (rt->cache->nre), rt->checked);
-	g_slice_free1 (rt->cache->nre, rt->results);
+	g_free (rt->checked);
+	g_free (rt->results);
 	REF_RELEASE (rt->cache);
-	g_slice_free1 (sizeof (*rt), rt);
+	g_free (rt);
 }
 
 void
@@ -1371,6 +1382,7 @@ rspamd_re_cache_compile_hyperscan (struct rspamd_re_cache *cache,
 	rspamd_regexp_t *re;
 	hs_compile_error_t *hs_errors;
 	guint *hs_flags = NULL;
+	const hs_expr_ext_t **hs_exts = NULL;
 	const gchar **hs_pats = NULL;
 	gchar *hs_serialized;
 	gsize serialized_len, total = 0;
@@ -1390,7 +1402,7 @@ rspamd_re_cache_compile_hyperscan (struct rspamd_re_cache *cache,
 			/* Read number of regexps */
 			g_assert (fd != -1);
 			lseek (fd, RSPAMD_HS_MAGIC_LEN + sizeof (cache->plt), SEEK_SET);
-			read (fd, &n, sizeof (n));
+			g_assert (read (fd, &n, sizeof (n)) == sizeof (n));
 			close (fd);
 
 			if (re_class->type_len > 0) {
@@ -1432,6 +1444,7 @@ rspamd_re_cache_compile_hyperscan (struct rspamd_re_cache *cache,
 		hs_flags = g_malloc0 (sizeof (*hs_flags) * n);
 		hs_ids = g_malloc (sizeof (*hs_ids) * n);
 		hs_pats = g_malloc (sizeof (*hs_pats) * n);
+		hs_exts = g_malloc0 (sizeof (*hs_exts) * n);
 		i = 0;
 
 		while (g_hash_table_iter_next (&cit, &k, &v)) {
@@ -1449,6 +1462,7 @@ rspamd_re_cache_compile_hyperscan (struct rspamd_re_cache *cache,
 			}
 
 			hs_flags[i] = 0;
+			hs_exts[i] = NULL;
 #ifndef WITH_PCRE2
 			if (pcre_flags & PCRE_FLAG(UTF8)) {
 				hs_flags[i] |= HS_FLAG_UTF8;
@@ -1503,9 +1517,10 @@ rspamd_re_cache_compile_hyperscan (struct rspamd_re_cache *cache,
 
 		if (n > 0) {
 			/* Create the hs tree */
-			if (hs_compile_multi (hs_pats,
+			if (hs_compile_ext_multi (hs_pats,
 					hs_flags,
 					hs_ids,
+					hs_exts,
 					n,
 					cache->vectorized_hyperscan ? HS_MODE_VECTORED : HS_MODE_BLOCK,
 					&cache->plt,
@@ -1518,6 +1533,7 @@ rspamd_re_cache_compile_hyperscan (struct rspamd_re_cache *cache,
 				g_free (hs_flags);
 				g_free (hs_ids);
 				g_free (hs_pats);
+				g_free (hs_exts);
 				close (fd);
 				unlink (path);
 				hs_free_compile_error (hs_errors);
@@ -1526,6 +1542,7 @@ rspamd_re_cache_compile_hyperscan (struct rspamd_re_cache *cache,
 			}
 
 			g_free (hs_pats);
+			g_free (hs_exts);
 
 			if (hs_serialize_database (test_db, &hs_serialized,
 					&serialized_len) != HS_SUCCESS) {
