@@ -37,11 +37,13 @@
         cache->static_pool->tag.tagname, cache->cfg->checksum, \
         G_STRFUNC, \
         __VA_ARGS__)
-#define msg_debug_cache(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        cache->static_pool->tag.tagname, cache->cfg->checksum, \
+#define msg_debug_cache(...)  rspamd_conditional_debug_fast (NULL, NULL, \
+        rspamd_symcache_log_id, "symcache", cache->cfg->checksum, \
         G_STRFUNC, \
         __VA_ARGS__)
 
+INIT_LOG_MODULE(symcache)
+
 static const guchar rspamd_symbols_cache_magic[8] = {'r', 's', 'c', 2, 0, 0, 0, 0 };
 
 static gint rspamd_symbols_cache_find_symbol_parent (struct symbols_cache *cache,
@@ -67,15 +69,16 @@ struct symbols_cache {
 	GPtrArray *prefilters;
 	GPtrArray *postfilters;
 	GPtrArray *composites;
+	GPtrArray *idempotent;
 	GList *delayed_deps;
 	GList *delayed_conditions;
 	rspamd_mempool_t *static_pool;
 	guint64 cksum;
 	gdouble total_weight;
 	guint used_items;
+	guint stats_symbols_count;
 	guint64 total_hits;
 	struct rspamd_config *cfg;
-	rspamd_mempool_mutex_t *mtx;
 	gdouble reload_time;
 	gint peak_cb;
 };
@@ -153,6 +156,8 @@ enum rspamd_cache_savepoint_stage {
 	RSPAMD_CACHE_PASS_WAIT_FILTERS,
 	RSPAMD_CACHE_PASS_POSTFILTERS,
 	RSPAMD_CACHE_PASS_WAIT_POSTFILTERS,
+	RSPAMD_CACHE_PASS_IDEMPOTENT,
+	RSPAMD_CACHE_PASS_WAIT_IDEMPOTENT,
 	RSPAMD_CACHE_PASS_DONE,
 };
 
@@ -212,7 +217,7 @@ rspamd_symbols_cache_order_dtor (gpointer p)
 	struct symbols_cache_order *ord = p;
 
 	g_ptr_array_free (ord->d, TRUE);
-	g_slice_free1 (sizeof (*ord), ord);
+	g_free (ord);
 }
 
 static void
@@ -228,7 +233,7 @@ rspamd_symbols_cache_order_new (gsize nelts)
 {
 	struct symbols_cache_order *ord;
 
-	ord = g_slice_alloc (sizeof (*ord));
+	ord = g_malloc0 (sizeof (*ord));
 	ord->d = g_ptr_array_sized_new (nelts);
 	REF_INIT_RETAIN (ord, rspamd_symbols_cache_order_dtor);
 
@@ -360,7 +365,9 @@ rspamd_symbols_cache_resort (struct symbols_cache *cache)
 		it = g_ptr_array_index (cache->items_by_id, i);
 		total_hits += it->st->total_hits;
 
-		if (!(it->type & (SYMBOL_TYPE_PREFILTER|SYMBOL_TYPE_POSTFILTER|SYMBOL_TYPE_COMPOSITE))) {
+		if (!(it->type & (SYMBOL_TYPE_PREFILTER|
+				SYMBOL_TYPE_POSTFILTER|
+				SYMBOL_TYPE_COMPOSITE))) {
 			g_ptr_array_add (ord->d, it);
 		}
 	}
@@ -406,6 +413,8 @@ rspamd_symbols_cache_post_init (struct symbols_cache *cache)
 					"%s is missing", ddep->from, ddep->to, ddep->from);
 		}
 		else {
+			msg_debug_cache ("delayed between %s(%d) -> %s", ddep->from,
+					it->id, ddep->to);
 			rspamd_symbols_cache_add_dependency (cache, it->id, ddep->to);
 		}
 
@@ -450,15 +459,24 @@ rspamd_symbols_cache_post_init (struct symbols_cache *cache)
 					dit = g_ptr_array_index (cache->items_by_id, dit->parent);
 				}
 
-				rdep = rspamd_mempool_alloc (cache->static_pool, sizeof (*rdep));
-				rdep->sym = dep->sym;
-				rdep->item = it;
-				rdep->id = i;
-				g_ptr_array_add (dit->rdeps, rdep);
-				dep->item = dit;
-				dep->id = dit->id;
+				if (dit->id == i) {
+					msg_err_cache ("cannot add dependency on self: %s -> %s "
+							"(resolved to %s)",
+							it->symbol, dep->sym, dit->symbol);
+				}
+				else {
+					rdep = rspamd_mempool_alloc (cache->static_pool,
+							sizeof (*rdep));
+					rdep->sym = dep->sym;
+					rdep->item = it;
+					rdep->id = i;
+					g_ptr_array_add (dit->rdeps, rdep);
+					dep->item = dit;
+					dep->id = dit->id;
 
-				msg_debug_cache ("add dependency from %d on %d", it->id, dit->id);
+					msg_debug_cache ("add dependency from %d on %d", it->id,
+							dit->id);
+				}
 			}
 			else {
 				msg_err_cache ("cannot find dependency on symbol %s", dep->sym);
@@ -468,6 +486,7 @@ rspamd_symbols_cache_post_init (struct symbols_cache *cache)
 
 	g_ptr_array_sort_with_data (cache->prefilters, prefilters_cmp, cache);
 	g_ptr_array_sort_with_data (cache->postfilters, postfilters_cmp, cache);
+	g_ptr_array_sort_with_data (cache->idempotent, postfilters_cmp, cache);
 }
 
 static gboolean
@@ -607,7 +626,8 @@ rspamd_symbols_cache_load_items (struct symbols_cache *cache, const gchar *name)
 				}
 			}
 
-			if ((item->type & SYMBOL_TYPE_VIRTUAL) && item->parent != -1) {
+			if ((item->type & SYMBOL_TYPE_VIRTUAL) &&
+					!(item->type & SYMBOL_TYPE_SQUEEZED) && item->parent != -1) {
 				g_assert (item->parent < (gint)cache->items_by_id->len);
 				parent = g_ptr_array_index (cache->items_by_id, item->parent);
 
@@ -633,6 +653,8 @@ rspamd_symbols_cache_load_items (struct symbols_cache *cache, const gchar *name)
 	return TRUE;
 }
 
+#define ROUND_DOUBLE(x) (floor((x) * 100.0) / 100.0)
+
 static gboolean
 rspamd_symbols_cache_save_items (struct symbols_cache *cache, const gchar *name)
 {
@@ -675,17 +697,21 @@ rspamd_symbols_cache_save_items (struct symbols_cache *cache, const gchar *name)
 	while (g_hash_table_iter_next (&it, &k, &v)) {
 		item = v;
 		elt = ucl_object_typed_new (UCL_OBJECT);
-		ucl_object_insert_key (elt, ucl_object_fromdouble (item->st->weight),
+		ucl_object_insert_key (elt,
+				ucl_object_fromdouble (ROUND_DOUBLE (item->st->weight)),
 				"weight", 0, false);
-		ucl_object_insert_key (elt, ucl_object_fromdouble (item->st->time_counter.mean),
+		ucl_object_insert_key (elt,
+				ucl_object_fromdouble (ROUND_DOUBLE (item->st->time_counter.mean)),
 				"time", 0, false);
-		ucl_object_insert_key (elt, ucl_object_fromdouble (item->st->total_hits),
+		ucl_object_insert_key (elt, ucl_object_fromint (item->st->total_hits),
 				"count", 0, false);
 
 		freq = ucl_object_typed_new (UCL_OBJECT);
-		ucl_object_insert_key (freq, ucl_object_fromdouble (item->st->frequency_counter.mean),
+		ucl_object_insert_key (freq,
+				ucl_object_fromdouble (ROUND_DOUBLE (item->st->frequency_counter.mean)),
 				"avg", 0, false);
-		ucl_object_insert_key (freq, ucl_object_fromdouble (item->st->frequency_counter.stddev),
+		ucl_object_insert_key (freq,
+				ucl_object_fromdouble (ROUND_DOUBLE (item->st->frequency_counter.stddev)),
 				"stddev", 0, false);
 		ucl_object_insert_key (elt, freq, "frequency", 0, false);
 
@@ -702,6 +728,8 @@ rspamd_symbols_cache_save_items (struct symbols_cache *cache, const gchar *name)
 	return ret;
 }
 
+#undef ROUND_DOUBLE
+
 gint
 rspamd_symbols_cache_add_symbol (struct symbols_cache *cache,
 	const gchar *name,
@@ -730,6 +758,12 @@ rspamd_symbols_cache_add_symbol (struct symbols_cache *cache,
 		}
 	}
 
+	if (type & (SYMBOL_TYPE_CLASSIFIER|SYMBOL_TYPE_CALLBACK|
+			SYMBOL_TYPE_PREFILTER|SYMBOL_TYPE_POSTFILTER|
+			SYMBOL_TYPE_IDEMPOTENT)) {
+		type |= SYMBOL_TYPE_NOSTAT;
+	}
+
 	item = rspamd_mempool_alloc0 (cache->static_pool,
 			sizeof (struct cache_item));
 	item->st = rspamd_mempool_alloc0_shared (cache->static_pool,
@@ -757,18 +791,26 @@ rspamd_symbols_cache_add_symbol (struct symbols_cache *cache,
 	item->parent = parent;
 	cache->used_items ++;
 
+	if (!(item->type &
+			(SYMBOL_TYPE_IDEMPOTENT|SYMBOL_TYPE_NOSTAT|SYMBOL_TYPE_CLASSIFIER))) {
+		if (name != NULL) {
+			cache->cksum = t1ha (name, strlen (name),
+					cache->cksum);
+		} else {
+			cache->cksum = t1ha (&item->id, sizeof (item->id),
+					cache->cksum);
+		}
+
+		cache->stats_symbols_count ++;
+	}
+
 	if (name != NULL) {
 		item->symbol = rspamd_mempool_strdup (cache->static_pool, name);
 		msg_debug_cache ("used items: %d, added symbol: %s, %d",
 				cache->used_items, name, item->id);
-		cache->cksum = t1ha (item->symbol, strlen (item->symbol),
-				cache->cksum);
-	}
-	else {
+	} else {
 		msg_debug_cache ("used items: %d, added unnamed symbol: %d",
 				cache->used_items, item->id);
-		cache->cksum = t1ha (&item->id, sizeof (item->id),
-				cache->cksum);
 	}
 
 	g_ptr_array_add (cache->items_by_id, item);
@@ -786,6 +828,9 @@ rspamd_symbols_cache_add_symbol (struct symbols_cache *cache,
 	if (item->type & SYMBOL_TYPE_PREFILTER) {
 		g_ptr_array_add (cache->prefilters, item);
 	}
+	else if (item->type & SYMBOL_TYPE_IDEMPOTENT) {
+		g_ptr_array_add (cache->idempotent, item);
+	}
 	else if (item->type & SYMBOL_TYPE_POSTFILTER) {
 		g_ptr_array_add (cache->postfilters, item);
 	}
@@ -856,7 +901,7 @@ rspamd_symbols_cache_add_condition_delayed (struct symbols_cache *cache,
 		return rspamd_symbols_cache_add_condition (cache, id, L, cbref);
 	}
 
-	ncond = g_slice_alloc (sizeof (*ncond));
+	ncond = g_malloc0 (sizeof (*ncond));
 	ncond->sym = g_strdup (sym);
 	ncond->cbref = cbref;
 	ncond->L = L;
@@ -899,7 +944,7 @@ rspamd_symbols_cache_destroy (struct symbols_cache *cache)
 				ddep = cur->data;
 				g_free (ddep->from);
 				g_free (ddep->to);
-				g_slice_free1 (sizeof (*ddep), ddep);
+				g_free (ddep);
 				cur = g_list_next (cur);
 			}
 
@@ -912,7 +957,7 @@ rspamd_symbols_cache_destroy (struct symbols_cache *cache)
 			while (cur) {
 				dcond = cur->data;
 				g_free (dcond->sym);
-				g_slice_free1 (sizeof (*dcond), dcond);
+				g_free (dcond);
 				cur = g_list_next (cur);
 			}
 
@@ -924,6 +969,7 @@ rspamd_symbols_cache_destroy (struct symbols_cache *cache)
 		g_ptr_array_free (cache->items_by_id, TRUE);
 		g_ptr_array_free (cache->prefilters, TRUE);
 		g_ptr_array_free (cache->postfilters, TRUE);
+		g_ptr_array_free (cache->idempotent, TRUE);
 		g_ptr_array_free (cache->composites, TRUE);
 		REF_RELEASE (cache->items_by_order);
 
@@ -931,7 +977,7 @@ rspamd_symbols_cache_destroy (struct symbols_cache *cache)
 			luaL_unref (cache->cfg->lua_state, LUA_REGISTRYINDEX, cache->peak_cb);
 		}
 
-		g_slice_free1 (sizeof (*cache), cache);
+		g_free (cache);
 	}
 }
 
@@ -940,7 +986,7 @@ rspamd_symbols_cache_new (struct rspamd_config *cfg)
 {
 	struct symbols_cache *cache;
 
-	cache = g_slice_alloc0 (sizeof (struct symbols_cache));
+	cache = g_malloc0 (sizeof (struct symbols_cache));
 	cache->static_pool =
 			rspamd_mempool_new (rspamd_mempool_suggest_size (), "symcache");
 	cache->items_by_symbol = g_hash_table_new (rspamd_str_hash,
@@ -948,8 +994,8 @@ rspamd_symbols_cache_new (struct rspamd_config *cfg)
 	cache->items_by_id = g_ptr_array_new ();
 	cache->prefilters = g_ptr_array_new ();
 	cache->postfilters = g_ptr_array_new ();
+	cache->idempotent = g_ptr_array_new ();
 	cache->composites = g_ptr_array_new ();
-	cache->mtx = rspamd_mempool_get_mutex (cache->static_pool);
 	cache->reload_time = cfg->cache_reload_time;
 	cache->total_hits = 1;
 	cache->total_weight = 1.0;
@@ -987,57 +1033,47 @@ static void
 rspamd_symbols_cache_validate_cb (gpointer k, gpointer v, gpointer ud)
 {
 	struct cache_item *item = v, *parent;
+	struct rspamd_config *cfg;
 	struct symbols_cache *cache = (struct symbols_cache *)ud;
-	GList *cur;
-	struct rspamd_metric *m;
 	struct rspamd_symbol *s;
 	gboolean skipped, ghost;
 	gint p1, p2;
 
 	ghost = item->st->weight == 0 ? TRUE : FALSE;
+	cfg = cache->cfg;
 
 	/* Check whether this item is skipped */
 	skipped = !ghost;
-	g_assert (cache->cfg != NULL);
+	g_assert (cfg != NULL);
 
 	if ((item->type &
 			(SYMBOL_TYPE_NORMAL|SYMBOL_TYPE_VIRTUAL|SYMBOL_TYPE_COMPOSITE|SYMBOL_TYPE_CLASSIFIER))
-			&& g_hash_table_lookup (cache->cfg->metrics_symbols, item->symbol) == NULL) {
-		cur = g_list_first (cache->cfg->metrics_list);
-		while (cur) {
-			m = cur->data;
+			&& g_hash_table_lookup (cfg->symbols, item->symbol) == NULL) {
 
-			if (m->accept_unknown_symbols) {
-				GList *mlist;
+		if (cfg->unknown_weight != 0) {
 
-				skipped = FALSE;
-				item->st->weight = m->unknown_weight;
-				s = rspamd_mempool_alloc0 (cache->static_pool,
-						sizeof (*s));
-				s->name = item->symbol;
-				s->weight_ptr = &item->st->weight;
-				g_hash_table_insert (m->symbols, item->symbol, s);
-				mlist = g_hash_table_lookup (cache->cfg->metrics_symbols,
-						item->symbol);
-				mlist = g_list_append (mlist, m);
-				g_hash_table_insert (cache->cfg->metrics_symbols,
-						item->symbol, mlist);
+			skipped = FALSE;
+			item->st->weight = cfg->unknown_weight;
+			s = rspamd_mempool_alloc0 (cache->static_pool,
+					sizeof (*s));
+			s->name = item->symbol;
+			s->weight_ptr = &item->st->weight;
+			g_hash_table_insert (cfg->symbols, item->symbol, s);
 
-				msg_info_cache ("adding unknown symbol %s to metric %s", item->symbol,
-						m->name);
-				ghost = FALSE;
-			}
-
-			cur = g_list_next (cur);
+			msg_info_cache ("adding unknown symbol %s", item->symbol);
+			ghost = FALSE;
+		}
+		else {
+			skipped = TRUE;
 		}
 	}
 	else {
 		skipped = FALSE;
 	}
 
-	if (skipped) {
+	if (!ghost && skipped) {
 		item->type |= SYMBOL_TYPE_SKIPPED;
-		msg_warn_cache ("symbol %s is not registered in any metric, so skip its check",
+		msg_warn_cache ("symbol %s has no score registered, skip its check",
 				item->symbol);
 	}
 
@@ -1050,7 +1086,8 @@ rspamd_symbols_cache_validate_cb (gpointer k, gpointer v, gpointer ud)
 		item->priority ++;
 	}
 
-	if ((item->type & SYMBOL_TYPE_VIRTUAL) && item->parent != -1) {
+	if ((item->type & SYMBOL_TYPE_VIRTUAL) &&
+			!(item->type & SYMBOL_TYPE_SQUEEZED) && item->parent != -1) {
 		g_assert (item->parent < (gint)cache->items_by_id->len);
 		parent = g_ptr_array_index (cache->items_by_id, item->parent);
 
@@ -1094,10 +1131,8 @@ rspamd_symbols_cache_validate (struct symbols_cache *cache,
 {
 	struct cache_item *item;
 	GHashTableIter it;
-	GList *cur;
 	gpointer k, v;
 	struct rspamd_symbol *sym_def;
-	struct rspamd_metric *metric;
 	gboolean ignore_symbol = FALSE, ret = TRUE;
 
 	if (cache == NULL) {
@@ -1106,32 +1141,23 @@ rspamd_symbols_cache_validate (struct symbols_cache *cache,
 	}
 
 	/* Now adjust symbol weights according to default metric */
-	if (cfg->default_metric != NULL) {
-		g_hash_table_foreach (cfg->default_metric->symbols,
+	g_hash_table_foreach (cfg->symbols,
 			rspamd_symbols_cache_metric_validate_cb,
 			cache);
-	}
 
 	g_hash_table_foreach (cache->items_by_symbol,
 			rspamd_symbols_cache_validate_cb,
 			cache);
 	/* Now check each metric item and find corresponding symbol in a cache */
-	g_hash_table_iter_init (&it, cfg->metrics_symbols);
+	g_hash_table_iter_init (&it, cfg->symbols);
 
 	while (g_hash_table_iter_next (&it, &k, &v)) {
 		ignore_symbol = FALSE;
-		cur = v;
+		sym_def = v;
 
-		while (cur) {
-			metric = cur->data;
-			sym_def = g_hash_table_lookup (metric->symbols, k);
-
-			if (sym_def && (sym_def->flags & RSPAMD_SYMBOL_FLAG_IGNORE)) {
-				ignore_symbol = TRUE;
-				break;
-			}
-
-			cur = g_list_next (cur);
+		if (sym_def && (sym_def->flags & RSPAMD_SYMBOL_FLAG_IGNORE)) {
+			ignore_symbol = TRUE;
+			break;
 		}
 
 		if (!ignore_symbol) {
@@ -1244,7 +1270,7 @@ rspamd_symbols_cache_check_symbol (struct rspamd_task *task,
 	struct rspamd_task **ptask;
 	lua_State *L;
 	gboolean check = TRUE;
-	const gdouble slow_diff_limit = 1e5;
+	const gdouble slow_diff_limit = 1e7;
 
 	if (item->func) {
 
@@ -1283,10 +1309,10 @@ rspamd_symbols_cache_check_symbol (struct rspamd_task *task,
 					rspamd_symbols_cache_watcher_cb,
 					item);
 			msg_debug_task ("execute %s, %d", item->symbol, item->id);
-			t1 = rspamd_get_ticks ();
+			t1 = rspamd_get_ticks (TRUE);
 			item->func (task, item->user_data);
-			t2 = rspamd_get_ticks ();
-			diff = (t2 - t1) * 1e6;
+			t2 = rspamd_get_ticks (TRUE);
+			diff = (t2 - t1);
 
 			if (G_UNLIKELY (RSPAMD_TASK_IS_PROFILING (task))) {
 				rspamd_task_profile_set (task, item->symbol, diff);
@@ -1296,12 +1322,12 @@ rspamd_symbols_cache_check_symbol (struct rspamd_task *task,
 				*total_diff += diff;
 			}
 
-			if (diff > slow_diff_limit) {
-				msg_info_task ("slow rule: %s: %d ms", item->symbol,
-						(gint)(diff / 1000.));
+			if (diff > slow_diff_limit && !(item->type & SYMBOL_TYPE_SQUEEZED)) {
+				msg_info_task ("slow rule: %s: %.0f ticks", item->symbol,
+						diff);
 			}
 
-			if (rspamd_worker_is_normal (task->worker)) {
+			if (rspamd_worker_is_scanner (task->worker)) {
 				rspamd_set_counter (item->cd, diff);
 			}
 
@@ -1453,7 +1479,8 @@ rspamd_symbols_cache_make_checkpoint (struct rspamd_task *task,
 	guint nitems;
 
 	nitems = cache->items_by_id->len - cache->postfilters->len -
-			cache->prefilters->len - cache->composites->len;
+			cache->prefilters->len - cache->composites->len -
+			cache->idempotent->len;
 
 	if (nitems != cache->items_by_order->d->len) {
 		/*
@@ -1491,7 +1518,6 @@ rspamd_symbols_cache_process_settings (struct rspamd_task *task,
 		struct symbols_cache *cache)
 {
 	const ucl_object_t *wl, *cur, *disabled, *enabled;
-	struct rspamd_metric *def;
 	struct rspamd_symbols_group *gr;
 	GHashTableIter gr_it;
 	ucl_object_iter_t it = NULL;
@@ -1520,15 +1546,14 @@ rspamd_symbols_cache_process_settings (struct rspamd_task *task,
 
 	/* Enable groups of symbols */
 	enabled = ucl_object_lookup (task->settings, "groups_enabled");
-	def = g_hash_table_lookup (task->cfg->metrics, DEFAULT_METRIC);
 
-	if (def && enabled) {
+	if (enabled) {
 		it = NULL;
 		rspamd_symbols_cache_disable_all_symbols (task, cache);
 
 		while ((cur = ucl_iterate_object (enabled, &it, true)) != NULL) {
 			if (ucl_object_type (cur) == UCL_STRING) {
-				gr = g_hash_table_lookup (def->groups,
+				gr = g_hash_table_lookup (task->cfg->groups,
 						ucl_object_tostring (cur));
 
 				if (gr) {
@@ -1555,14 +1580,13 @@ rspamd_symbols_cache_process_settings (struct rspamd_task *task,
 
 	/* Disable groups of symbols */
 	disabled = ucl_object_lookup (task->settings, "groups_disabled");
-	def = g_hash_table_lookup (task->cfg->metrics, DEFAULT_METRIC);
 
-	if (def && disabled) {
+	if (disabled) {
 		it = NULL;
 
 		while ((cur = ucl_iterate_object (disabled, &it, true)) != NULL) {
 			if (ucl_object_type (cur) == UCL_STRING) {
-				gr = g_hash_table_lookup (def->groups,
+				gr = g_hash_table_lookup (task->cfg->groups,
 						ucl_object_tostring (cur));
 
 				if (gr) {
@@ -1586,10 +1610,10 @@ rspamd_symbols_cache_process_symbols (struct rspamd_task * task,
 	struct cache_item *item = NULL;
 	struct cache_savepoint *checkpoint;
 	gint i;
-	gdouble total_microseconds = 0;
+	gdouble total_ticks = 0;
 	gboolean all_done;
 	gint saved_priority;
-	const gdouble max_microseconds = 3e5;
+	const gdouble max_ticks = 3e8;
 	guint start_events_pending;
 
 	g_assert (cache != NULL);
@@ -1607,6 +1631,11 @@ rspamd_symbols_cache_process_symbols (struct rspamd_task * task,
 		checkpoint->pass = RSPAMD_CACHE_PASS_POSTFILTERS;
 	}
 
+	if (stage == RSPAMD_TASK_STAGE_IDEMPOTENT && checkpoint->pass <
+			RSPAMD_CACHE_PASS_IDEMPOTENT) {
+		checkpoint->pass = RSPAMD_CACHE_PASS_IDEMPOTENT;
+	}
+
 	msg_debug_task ("symbols processing stage at pass: %d", checkpoint->pass);
 	start_events_pending = rspamd_session_events_pending (task->s);
 
@@ -1638,7 +1667,7 @@ rspamd_symbols_cache_process_symbols (struct rspamd_task * task,
 				}
 
 				rspamd_symbols_cache_check_symbol (task, cache, item,
-						checkpoint, &total_microseconds);
+						checkpoint, &total_ticks);
 			}
 		}
 
@@ -1715,10 +1744,10 @@ rspamd_symbols_cache_process_symbols (struct rspamd_task * task,
 				}
 
 				rspamd_symbols_cache_check_symbol (task, cache, item,
-						checkpoint, &total_microseconds);
+						checkpoint, &total_ticks);
 			}
 
-			if (total_microseconds > max_microseconds) {
+			if (total_ticks > max_ticks) {
 				/* Maybe we should stop and check pending events? */
 				if (rspamd_session_events_pending (task->s) > start_events_pending) {
 					/* Add some timeout event to avoid too long waiting */
@@ -1738,8 +1767,8 @@ rspamd_symbols_cache_process_symbols (struct rspamd_task * task,
 							(rspamd_mempool_destruct_t)event_del, ev);
 #endif
 					msg_info_task ("trying to check async events after spending "
-							"%d microseconds processing symbols",
-							(gint)total_microseconds);
+							"%.0f ticks processing symbols",
+							total_ticks);
 
 					return TRUE;
 				}
@@ -1761,16 +1790,16 @@ rspamd_symbols_cache_process_symbols (struct rspamd_task * task,
 				}
 
 				rspamd_symbols_cache_check_symbol (task, cache, item,
-						checkpoint, &total_microseconds);
+						checkpoint, &total_ticks);
 			}
 
-			if (total_microseconds > max_microseconds) {
+			if (total_ticks > max_ticks) {
 				/* Maybe we should stop and check pending events? */
 				if (rspamd_session_events_pending (task->s) >
 						start_events_pending) {
 					msg_debug_task ("trying to check async events after spending "
-							"%d microseconds processing symbols",
-							(gint)total_microseconds);
+							"%.0f microseconds processing symbols",
+							total_ticks);
 					return TRUE;
 				}
 			}
@@ -1811,8 +1840,9 @@ rspamd_symbols_cache_process_symbols (struct rspamd_task * task,
 						return TRUE;
 					}
 				}
+
 				rspamd_symbols_cache_check_symbol (task, cache, item,
-						checkpoint, &total_microseconds);
+						checkpoint, &total_ticks);
 			}
 		}
 		checkpoint->pass = RSPAMD_CACHE_PASS_WAIT_POSTFILTERS;
@@ -1830,6 +1860,64 @@ rspamd_symbols_cache_process_symbols (struct rspamd_task * task,
 			}
 		}
 
+		if (all_done) {
+			checkpoint->pass = RSPAMD_CACHE_PASS_IDEMPOTENT;
+		}
+
+		if (checkpoint->waitq->len == 0 ||
+				stage == RSPAMD_TASK_STAGE_IDEMPOTENT) {
+			checkpoint->pass = RSPAMD_CACHE_PASS_IDEMPOTENT;
+		}
+
+		if (stage == RSPAMD_TASK_STAGE_IDEMPOTENT) {
+			return rspamd_symbols_cache_process_symbols (task, cache, stage);
+		}
+
+		break;
+
+	case RSPAMD_CACHE_PASS_IDEMPOTENT:
+		/* Check for postfilters */
+		saved_priority = G_MININT;
+
+		for (i = 0; i < (gint)cache->idempotent->len; i ++) {
+			item = g_ptr_array_index (cache->idempotent, i);
+
+			if (!isset (checkpoint->processed_bits, item->id * 2) &&
+					!isset (checkpoint->processed_bits, item->id * 2 + 1)) {
+				/* Check priorities */
+				if (saved_priority == G_MININT) {
+					saved_priority = item->priority;
+				}
+				else {
+					if (item->priority > saved_priority &&
+							rspamd_session_events_pending (task->s) > start_events_pending) {
+						/*
+						 * Delay further checks as we have higher
+						 * priority filters to be processed
+						 */
+						checkpoint->pass = RSPAMD_CACHE_PASS_IDEMPOTENT;
+						return TRUE;
+					}
+				}
+				rspamd_symbols_cache_check_symbol (task, cache, item,
+						checkpoint, &total_ticks);
+			}
+		}
+		checkpoint->pass = RSPAMD_CACHE_PASS_WAIT_IDEMPOTENT;
+		break;
+
+	case RSPAMD_CACHE_PASS_WAIT_IDEMPOTENT:
+		all_done = TRUE;
+
+		for (i = 0; i < (gint)cache->idempotent->len; i ++) {
+			item = g_ptr_array_index (cache->idempotent, i);
+
+			if (!isset (checkpoint->processed_bits, item->id * 2 + 1)) {
+				all_done = FALSE;
+				break;
+			}
+		}
+
 		if (all_done) {
 			checkpoint->pass = RSPAMD_CACHE_PASS_DONE;
 
@@ -1850,6 +1938,8 @@ struct counters_cbdata {
 	struct symbols_cache *cache;
 };
 
+#define ROUND_DOUBLE(x) (floor((x) * 100.0) / 100.0)
+
 static void
 rspamd_symbols_cache_counters_cb (gpointer v, gpointer ud)
 {
@@ -1864,27 +1954,36 @@ rspamd_symbols_cache_counters_cb (gpointer v, gpointer ud)
 		ucl_object_insert_key (obj, ucl_object_fromstring (item->symbol),
 				"symbol", 0, false);
 
-		if ((item->type & SYMBOL_TYPE_VIRTUAL) && item->parent != -1) {
+		if ((item->type & SYMBOL_TYPE_VIRTUAL) &&
+				!(item->type & SYMBOL_TYPE_SQUEEZED) && item->parent != -1) {
 			g_assert (item->parent < (gint)cbd->cache->items_by_id->len);
 			parent = g_ptr_array_index (cbd->cache->items_by_id,
 					item->parent);
-			ucl_object_insert_key (obj, ucl_object_fromdouble (item->st->weight),
+			ucl_object_insert_key (obj,
+					ucl_object_fromdouble (ROUND_DOUBLE (item->st->weight)),
 					"weight", 0, false);
-			ucl_object_insert_key (obj, ucl_object_fromdouble (parent->st->avg_frequency),
+			ucl_object_insert_key (obj,
+					ucl_object_fromdouble (ROUND_DOUBLE (parent->st->avg_frequency)),
 					"frequency", 0, false);
-			ucl_object_insert_key (obj, ucl_object_fromint (parent->st->total_hits),
+			ucl_object_insert_key (obj,
+					ucl_object_fromint (parent->st->total_hits),
 					"hits", 0, false);
-			ucl_object_insert_key (obj, ucl_object_fromdouble (parent->st->avg_time),
+			ucl_object_insert_key (obj,
+					ucl_object_fromdouble (ROUND_DOUBLE (parent->st->avg_time)),
 					"time", 0, false);
 		}
 		else {
-			ucl_object_insert_key (obj, ucl_object_fromdouble (item->st->weight),
+			ucl_object_insert_key (obj,
+					ucl_object_fromdouble (ROUND_DOUBLE (item->st->weight)),
 					"weight", 0, false);
-			ucl_object_insert_key (obj, ucl_object_fromdouble (item->st->avg_frequency),
+			ucl_object_insert_key (obj,
+					ucl_object_fromdouble (ROUND_DOUBLE (item->st->avg_frequency)),
 					"frequency", 0, false);
-			ucl_object_insert_key (obj, ucl_object_fromint (item->st->total_hits),
+			ucl_object_insert_key (obj,
+					ucl_object_fromint (item->st->total_hits),
 					"hits", 0, false);
-			ucl_object_insert_key (obj, ucl_object_fromdouble (item->st->avg_time),
+			ucl_object_insert_key (obj,
+					ucl_object_fromdouble (ROUND_DOUBLE (item->st->avg_time)),
 					"time", 0, false);
 		}
 
@@ -1892,6 +1991,8 @@ rspamd_symbols_cache_counters_cb (gpointer v, gpointer ud)
 	}
 }
 
+#undef ROUND_DOUBLE
+
 ucl_object_t *
 rspamd_symbols_cache_counters (struct symbols_cache * cache)
 {
@@ -1949,7 +2050,7 @@ rspamd_symbols_cache_resort_cb (gint fd, short what, gpointer ud)
 	cache = cbdata->cache;
 	/* Plan new event */
 	tm = rspamd_time_jitter (cache->reload_time, 0);
-	cur_ticks = rspamd_get_ticks ();
+	cur_ticks = rspamd_get_ticks (TRUE);
 	msg_debug_cache ("resort symbols cache, next reload in %.2f seconds", tm);
 	g_assert (cache != NULL);
 	evtimer_set (&cbdata->resort_ev, rspamd_symbols_cache_resort_cb, cbdata);
@@ -1957,9 +2058,7 @@ rspamd_symbols_cache_resort_cb (gint fd, short what, gpointer ud)
 	double_to_tv (tm, &tv);
 	event_add (&cbdata->resort_ev, &tv);
 
-	if (rspamd_worker_is_normal (cbdata->w)) {
-		rspamd_mempool_lock_mutex (cache->mtx);
-
+	if (rspamd_worker_is_primary_controller (cbdata->w))
 		/* Gather stats from shared execution times */
 		for (i = 0; i < cache->items_by_id->len; i ++) {
 			item = g_ptr_array_index (cache->items_by_id, i);
@@ -2020,20 +2119,17 @@ rspamd_symbols_cache_resort_cb (gint fd, short what, gpointer ud)
 			}
 		}
 
-		/* Sync virtual symbols */
-		for (i = 0; i < cache->items_by_id->len; i ++) {
-			item = g_ptr_array_index (cache->items_by_id, i);
+	/* Sync virtual symbols */
+	for (i = 0; i < cache->items_by_id->len; i ++) {
+		item = g_ptr_array_index (cache->items_by_id, i);
 
-			if (item->parent != -1) {
-				parent = g_ptr_array_index (cache->items_by_id, item->parent);
+		if (item->parent != -1) {
+			parent = g_ptr_array_index (cache->items_by_id, item->parent);
 
-				if (parent) {
-					item->st->avg_time = parent->st->avg_time;
-				}
+			if (parent) {
+				item->st->avg_time = parent->st->avg_time;
 			}
 		}
-
-		rspamd_mempool_unlock_mutex (cache->mtx);
 	}
 
 	cbdata->last_resort = cur_ticks;
@@ -2049,19 +2145,20 @@ rspamd_symbols_cache_start_refresh (struct symbols_cache * cache,
 	struct rspamd_cache_refresh_cbdata *cbdata;
 
 	cbdata = rspamd_mempool_alloc0 (cache->static_pool, sizeof (*cbdata));
-	cbdata->last_resort = rspamd_get_ticks ();
+	cbdata->last_resort = rspamd_get_ticks (TRUE);
 	cbdata->ev_base = ev_base;
 	cbdata->w = w;
 	cbdata->cache = cache;
 	tm = rspamd_time_jitter (cache->reload_time, 0);
 	msg_debug_cache ("next reload in %.2f seconds", tm);
 	g_assert (cache != NULL);
-	evtimer_set (&cbdata->resort_ev, rspamd_symbols_cache_resort_cb, cbdata);
+	evtimer_set (&cbdata->resort_ev, rspamd_symbols_cache_resort_cb,
+			cbdata);
 	event_base_set (ev_base, &cbdata->resort_ev);
 	double_to_tv (tm, &tv);
 	event_add (&cbdata->resort_ev, &tv);
 	rspamd_mempool_add_destructor (cache->static_pool,
-			(rspamd_mempool_destruct_t)event_del,
+			(rspamd_mempool_destruct_t) event_del,
 			&cbdata->resort_ev);
 }
 
@@ -2107,7 +2204,7 @@ rspamd_symbols_cache_add_delayed_dependency (struct symbols_cache *cache,
 	g_assert (from != NULL);
 	g_assert (to != NULL);
 
-	ddep = g_slice_alloc (sizeof (*ddep));
+	ddep = g_malloc0 (sizeof (*ddep));
 	ddep->from = g_strdup (from);
 	ddep->to = g_strdup (to);
 
@@ -2208,11 +2305,11 @@ rspamd_symbols_cache_symbol_by_id (struct symbols_cache *cache,
 }
 
 guint
-rspamd_symbols_cache_symbols_count (struct symbols_cache *cache)
+rspamd_symbols_cache_stats_symbols_count (struct symbols_cache *cache)
 {
 	g_assert (cache != NULL);
 
-	return cache->items_by_id->len;
+	return cache->stats_symbols_count;
 }
 
 static void
@@ -2252,14 +2349,19 @@ rspamd_symbols_cache_disable_symbol_checkpoint (struct rspamd_task *task,
 
 	id = rspamd_symbols_cache_find_symbol_parent (cache, symbol);
 
-	if (id > 0) {
+	if (id >= 0) {
 		/* Set executed and finished flags */
 		item = g_ptr_array_index (cache->items_by_id, id);
 
-		setbit (checkpoint->processed_bits, item->id * 2);
-		setbit (checkpoint->processed_bits, item->id * 2 + 1);
+		if (!(item->type & SYMBOL_TYPE_SQUEEZED)) {
+			setbit (checkpoint->processed_bits, item->id * 2);
+			setbit (checkpoint->processed_bits, item->id * 2 + 1);
 
-		msg_debug_task ("disable execution of %s", symbol);
+			msg_debug_task ("disable execution of %s", symbol);
+		}
+		else {
+			msg_debug_task ("skip squeezed symbol %s", symbol);
+		}
 	}
 	else {
 		msg_info_task ("cannot disable %s: not found", symbol);
@@ -2284,14 +2386,14 @@ rspamd_symbols_cache_enable_symbol_checkpoint (struct rspamd_task *task,
 
 	id = rspamd_symbols_cache_find_symbol_parent (cache, symbol);
 
-	if (id > 0) {
+	if (id >= 0) {
 		/* Set executed and finished flags */
 		item = g_ptr_array_index (cache->items_by_id, id);
 
 		clrbit (checkpoint->processed_bits, item->id * 2);
 		clrbit (checkpoint->processed_bits, item->id * 2 + 1);
 
-		msg_debug_task ("enable execution of %s", symbol);
+		msg_debug_task ("enable execution of %s (%d)", symbol, id);
 	}
 	else {
 		msg_info_task ("cannot enable %s: not found", symbol);
@@ -2462,3 +2564,16 @@ rspamd_symbols_cache_is_symbol_enabled (struct rspamd_task *task,
 
 	return ret;
 }
+
+void
+rspamd_symbols_cache_foreach (struct symbols_cache *cache,
+		void (*func)(gint , const gchar *, gint , gpointer ),
+		gpointer ud)
+{
+	guint i;
+	struct cache_item *item;
+
+	PTR_ARRAY_FOREACH (cache->items_by_id, i, item) {
+		func (item->id, item->symbol, item->type, ud);
+	}
+}
