@@ -18,6 +18,7 @@
 #include "stat_internal.h"
 #include "upstream.h"
 #include "lua/lua_common.h"
+#include "libserver/mempool_vars_internal.h"
 
 #ifdef WITH_HIREDIS
 #include "hiredis.h"
@@ -46,6 +47,7 @@ struct redis_stat_ctx {
 	gboolean enable_users;
 	gboolean store_tokens;
 	gboolean new_schema;
+	gboolean enable_signatures;
 	guint expiry;
 	gint cbref_user;
 };
@@ -70,6 +72,7 @@ struct redis_stat_runtime {
 	guint64 learned;
 	gint id;
 	gboolean has_event;
+	GError *err;
 };
 
 /* Used to get statistics from redis */
@@ -598,6 +601,85 @@ rspamd_redis_tokens_to_query (struct rspamd_task *task,
 	return out;
 }
 
+static void
+rspamd_redis_store_stat_signature (struct rspamd_task *task,
+		struct redis_stat_runtime *rt,
+		GPtrArray *tokens,
+		const gchar *prefix)
+{
+	gchar *sig, keybuf[512], nbuf[64];
+	rspamd_token_t *tok;
+	guint i, blen, klen;
+	rspamd_fstring_t *out;
+
+	out = rspamd_fstring_sized_new (1024);
+	sig = rspamd_mempool_get_variable (task->task_pool,
+			RSPAMD_MEMPOOL_STAT_SIGNATURE);
+
+	if (sig == NULL) {
+		msg_err_task ("cannot get bayes signature");
+		return;
+	}
+
+	klen = rspamd_snprintf (keybuf, sizeof (keybuf), "%s_%s_%s",
+			prefix, sig, rt->stcf->is_spam ? "S" : "H");
+
+	out->len = 0;
+
+	/* Cleanup key */
+	rspamd_printf_fstring (&out, ""
+					"*2\r\n"
+					"$3\r\n"
+					"DEL\r\n"
+					"$%d\r\n"
+					"%s\r\n",
+			klen, keybuf);
+	redisAsyncFormattedCommand (rt->redis, NULL, NULL,
+			out->str, out->len);
+	out->len = 0;
+
+	rspamd_printf_fstring (&out, ""
+					"*%d\r\n"
+					"$5\r\n"
+					"LPUSH\r\n"
+					"$%d\r\n"
+					"%s\r\n",
+			tokens->len + 2,
+			klen, keybuf);
+
+	PTR_ARRAY_FOREACH (tokens, i, tok) {
+		blen = rspamd_snprintf (nbuf, sizeof (nbuf), "%uL", tok->data);
+		rspamd_printf_fstring (&out, ""
+				"$%d\r\n"
+				"%s\r\n", blen, nbuf);
+	}
+
+	redisAsyncFormattedCommand (rt->redis, NULL, NULL,
+			out->str, out->len);
+	out->len = 0;
+
+	if (rt->ctx->expiry > 0) {
+		out->len = 0;
+		blen = rspamd_snprintf (nbuf, sizeof (nbuf), "%d",
+				rt->ctx->expiry);
+
+		rspamd_printf_fstring (&out, ""
+						"*3\r\n"
+						"$6\r\n"
+						"EXPIRE\r\n"
+						"$%d\r\n"
+						"%s\r\n"
+						"$%d\r\n"
+						"%s\r\n",
+				klen, keybuf,
+				blen, nbuf);
+		redisAsyncFormattedCommand (rt->redis, NULL, NULL,
+				out->str, out->len);
+	}
+
+	rspamd_fstring_free (out);
+}
+
 static void
 rspamd_redis_async_cbdata_cleanup (struct rspamd_redis_stat_cbdata *cbdata)
 {
@@ -636,7 +718,7 @@ rspamd_redis_async_cbdata_cleanup (struct rspamd_redis_stat_cbdata *cbdata)
 			ucl_object_unref (cbdata->cur);
 		}
 
-		g_slice_free1 (sizeof (*cbdata), cbdata);
+		g_free (cbdata);
 	}
 }
 
@@ -663,7 +745,7 @@ rspamd_redis_stat_learns (redisAsyncContext *c, gpointer r, gpointer priv)
 			rspamd_strtoul (reply->str, reply->len, &num);
 		}
 
-		obj = (ucl_object_t *)ucl_object_lookup (cbdata->cur, "revision");
+		obj = (ucl_object_t *) ucl_object_lookup (cbdata->cur, "revision");
 		if (obj) {
 			obj->value.iv += num;
 		}
@@ -761,15 +843,35 @@ rspamd_redis_stat_keys (redisAsyncContext *c, gpointer r, gpointer priv)
 					k = (gchar *)g_ptr_array_index (cbdata->cur_keys, i);
 
 					if (k) {
-						redisAsyncCommand (cbdata->redis, rspamd_redis_stat_key,
-								cbdata,
-								"HLEN %s",
-								k);
-						redisAsyncCommand (cbdata->redis, rspamd_redis_stat_learns,
-								cbdata,
-								"HGET %s learns",
-								k);
-						cbdata->inflight += 2;
+						const gchar *learned_key = "learns";
+
+						if (cbdata->elt->ctx->new_schema) {
+							if (cbdata->elt->ctx->stcf->is_spam) {
+								learned_key = "learns_spam";
+							}
+							else {
+								learned_key = "learns_ham";
+							}
+							redisAsyncCommand (cbdata->redis,
+									rspamd_redis_stat_learns,
+									cbdata,
+									"HGET %s %s",
+									k, learned_key);
+							cbdata->inflight += 1;
+						}
+						else {
+							redisAsyncCommand (cbdata->redis,
+									rspamd_redis_stat_key,
+									cbdata,
+									"HLEN %s",
+									k);
+							redisAsyncCommand (cbdata->redis,
+									rspamd_redis_stat_learns,
+									cbdata,
+									"HGET %s %s",
+									k, learned_key);
+							cbdata->inflight += 2;
+						}
 					}
 				}
 			}
@@ -832,7 +934,7 @@ rspamd_redis_async_stat_cb (struct rspamd_stat_async_elt *elt, gpointer d)
 	/* Disable further events unless needed */
 	elt->enabled = FALSE;
 
-	cbdata = g_slice_alloc0 (sizeof (*cbdata));
+	cbdata = g_malloc0 (sizeof (*cbdata));
 	cbdata->selected = rspamd_upstream_get (ctx->read_servers,
 					RSPAMD_UPSTREAM_ROUND_ROBIN,
 					NULL,
@@ -937,6 +1039,12 @@ rspamd_redis_timeout (gint fd, short what, gpointer d)
 		/* This calls for all callbacks pending */
 		redisAsyncFree (redis);
 	}
+
+	if (!rt->err) {
+		g_set_error (&rt->err, rspamd_redis_stat_quark (), ETIMEDOUT,
+				"error getting reply from redis server %s: timeout",
+				rspamd_upstream_name (rt->selected));
+	}
 }
 
 /* Called when we have connected to the redis server and got stats */
@@ -984,6 +1092,12 @@ rspamd_redis_connected (redisAsyncContext *c, gpointer r, gpointer priv)
 		msg_err_task ("error getting reply from redis server %s: %s",
 				rspamd_upstream_name (rt->selected), c->errstr);
 		rspamd_upstream_fail (rt->selected);
+
+		if (!rt->err) {
+			g_set_error (&rt->err, rspamd_redis_stat_quark (), c->err,
+					"error getting reply from redis server %s: %s",
+					rspamd_upstream_name (rt->selected), c->errstr);
+		}
 	}
 
 }
@@ -1066,6 +1180,12 @@ rspamd_redis_processed (redisAsyncContext *c, gpointer r, gpointer priv)
 		if (rt->redis) {
 			rspamd_upstream_fail (rt->selected);
 		}
+
+		if (!rt->err) {
+			g_set_error (&rt->err, rspamd_redis_stat_quark (), c->err,
+					"cannot get values: error getting reply from redis server %s: %s",
+					rspamd_upstream_name (rt->selected), c->errstr);
+		}
 	}
 
 	if (rt->has_event) {
@@ -1092,6 +1212,12 @@ rspamd_redis_learned (redisAsyncContext *c, gpointer r, gpointer priv)
 		if (rt->redis) {
 			rspamd_upstream_fail (rt->selected);
 		}
+
+		if (!rt->err) {
+			g_set_error (&rt->err, rspamd_redis_stat_quark (), c->err,
+					"cannot get learned: error getting reply from redis server %s: %s",
+					rspamd_upstream_name (rt->selected), c->errstr);
+		}
 	}
 
 	if (rt->has_event) {
@@ -1105,8 +1231,7 @@ rspamd_redis_try_ucl (struct redis_stat_ctx *backend,
 		struct rspamd_config *cfg,
 		const gchar *symbol)
 {
-	const ucl_object_t *elt, *relt, *users_enabled;
-	const gchar *lua_script;
+	const ucl_object_t *elt, *relt;
 
 	elt = ucl_object_lookup_any (obj, "read_servers", "servers", NULL);
 
@@ -1147,6 +1272,38 @@ rspamd_redis_try_ucl (struct redis_stat_ctx *backend,
 		}
 	}
 
+	elt = ucl_object_lookup_any (obj, "db", "database", "dbname", NULL);
+	if (elt) {
+		if (ucl_object_type (elt) == UCL_STRING) {
+			backend->dbname = ucl_object_tostring (elt);
+		}
+		else if (ucl_object_type (elt) == UCL_INT) {
+			backend->dbname = ucl_object_tostring_forced (elt);
+		}
+	}
+	else {
+		backend->dbname = NULL;
+	}
+
+	elt = ucl_object_lookup (obj, "password");
+	if (elt) {
+		backend->password = ucl_object_tostring (elt);
+	}
+	else {
+		backend->password = NULL;
+	}
+
+	return TRUE;
+}
+
+static void
+rspamd_redis_parse_classifier_opts (struct redis_stat_ctx *backend,
+		const ucl_object_t *obj,
+		struct rspamd_config *cfg)
+{
+	const gchar *lua_script;
+	const ucl_object_t *elt, *users_enabled;
+
 	users_enabled = ucl_object_lookup_any (obj, "per_user",
 			"users_enabled", NULL);
 
@@ -1204,14 +1361,6 @@ rspamd_redis_try_ucl (struct redis_stat_ctx *backend,
 		backend->timeout = REDIS_DEFAULT_TIMEOUT;
 	}
 
-	elt = ucl_object_lookup (obj, "password");
-	if (elt) {
-		backend->password = ucl_object_tostring (elt);
-	}
-	else {
-		backend->password = NULL;
-	}
-
 	elt = ucl_object_lookup (obj, "store_tokens");
 	if (elt) {
 		backend->store_tokens = ucl_object_toboolean (elt);
@@ -1226,6 +1375,18 @@ rspamd_redis_try_ucl (struct redis_stat_ctx *backend,
 	}
 	else {
 		backend->new_schema = FALSE;
+
+		msg_warn_config ("you are using old bayes schema for redis statistics, "
+				"please consider converting it to a new one "
+				"by using 'rspamadm configwizard statistics'");
+	}
+
+	elt = ucl_object_lookup (obj, "signatures");
+	if (elt) {
+		backend->enable_signatures = ucl_object_toboolean (elt);
+	}
+	else {
+		backend->enable_signatures = FALSE;
 	}
 
 	elt = ucl_object_lookup_any (obj, "expiry", "expire", NULL);
@@ -1235,16 +1396,6 @@ rspamd_redis_try_ucl (struct redis_stat_ctx *backend,
 	else {
 		backend->expiry = 0;
 	}
-
-	elt = ucl_object_lookup_any (obj, "db", "database", "dbname", NULL);
-	if (elt) {
-		backend->dbname = ucl_object_tostring (elt);
-	}
-	else {
-		backend->dbname = NULL;
-	}
-
-	return TRUE;
 }
 
 gpointer
@@ -1257,7 +1408,7 @@ rspamd_redis_init (struct rspamd_stat_ctx *ctx,
 	const ucl_object_t *obj;
 	gboolean ret = FALSE;
 
-	backend = g_slice_alloc0 (sizeof (*backend));
+	backend = g_malloc0 (sizeof (*backend));
 
 	/* First search in backend configuration */
 	obj = ucl_object_lookup (st->classifier->cfg->opts, "backend");
@@ -1298,14 +1449,15 @@ rspamd_redis_init (struct rspamd_stat_ctx *ctx,
 
 	if (!ret) {
 		msg_err_config ("cannot init redis backend for %s", stf->symbol);
-		g_slice_free1 (sizeof (*backend), backend);
+		g_free (backend);
 		return NULL;
 	}
 
+	rspamd_redis_parse_classifier_opts (backend, st->classifier->cfg->opts, cfg);
 	stf->clcf->flags |= RSPAMD_FLAG_CLASSIFIER_INCREMENTING_BACKEND;
 	backend->stcf = stf;
 
-	st_elt = g_slice_alloc0 (sizeof (*st_elt));
+	st_elt = g_malloc0 (sizeof (*st_elt));
 	st_elt->ev_base = ctx->ev_base;
 	st_elt->ctx = backend;
 	backend->stat_elt = rspamd_stat_ctx_register_async (
@@ -1355,6 +1507,8 @@ rspamd_redis_runtime (struct rspamd_task *task,
 	}
 
 	rt = rspamd_mempool_alloc0 (task->task_pool, sizeof (*rt));
+	rspamd_mempool_add_destructor (task->task_pool,
+			rspamd_gerror_free_maybe, &rt->err);
 	rspamd_redis_expand_object (ctx->redis_object, ctx, task,
 			&rt->redis_object_expanded);
 	rt->selected = up;
@@ -1397,7 +1551,7 @@ rspamd_redis_close (gpointer p)
 		rspamd_upstreams_destroy (ctx->write_servers);
 	}
 
-	g_slice_free1 (sizeof (*ctx), ctx);
+	g_free (ctx);
 }
 
 gboolean
@@ -1409,6 +1563,8 @@ rspamd_redis_process_tokens (struct rspamd_task *task,
 	rspamd_fstring_t *query;
 	struct timeval tv;
 	gint ret;
+	const gchar *learned_key = "learns";
+
 
 	if (tokens == NULL || tokens->len == 0 || rt->redis == NULL) {
 		return FALSE;
@@ -1416,8 +1572,17 @@ rspamd_redis_process_tokens (struct rspamd_task *task,
 
 	rt->id = id;
 
+	if (rt->ctx->new_schema) {
+		if (rt->ctx->stcf->is_spam) {
+			learned_key = "learns_spam";
+		}
+		else {
+			learned_key = "learns_ham";
+		}
+	}
+
 	if (redisAsyncCommand (rt->redis, rspamd_redis_connected, rt, "HGET %s %s",
-			rt->redis_object_expanded, "learns") == REDIS_OK) {
+			rt->redis_object_expanded, learned_key) == REDIS_OK) {
 
 		rspamd_session_add_event (task->s, rspamd_redis_fin, rt,
 				rspamd_redis_stat_quark ());
@@ -1453,7 +1618,7 @@ rspamd_redis_process_tokens (struct rspamd_task *task,
 	return FALSE;
 }
 
-void
+gboolean
 rspamd_redis_finalize_process (struct rspamd_task *task, gpointer runtime,
 		gpointer ctx)
 {
@@ -1469,6 +1634,12 @@ rspamd_redis_finalize_process (struct rspamd_task *task, gpointer runtime,
 		rt->redis = NULL;
 		redisAsyncFree (redis);
 	}
+
+	if (rt->err) {
+		return FALSE;
+	}
+
+	return TRUE;
 }
 
 gboolean
@@ -1484,6 +1655,7 @@ rspamd_redis_learn_tokens (struct rspamd_task *task, GPtrArray *tokens,
 	rspamd_token_t *tok;
 	gint ret;
 	goffset off;
+	const gchar *learned_key = "learns";
 
 	up = rspamd_upstream_get (rt->ctx->write_servers,
 			RSPAMD_UPSTREAM_MASTER_SLAVE,
@@ -1497,6 +1669,15 @@ rspamd_redis_learn_tokens (struct rspamd_task *task, GPtrArray *tokens,
 
 	rt->selected = up;
 
+	if (rt->ctx->new_schema) {
+		if (rt->ctx->stcf->is_spam) {
+			learned_key = "learns_spam";
+		}
+		else {
+			learned_key = "learns_ham";
+		}
+	}
+
 	addr = rspamd_upstream_addr (up);
 	g_assert (addr != NULL);
 
@@ -1548,12 +1729,14 @@ rspamd_redis_learn_tokens (struct rspamd_task *task, GPtrArray *tokens,
 				"HINCRBY\r\n"
 				"$%d\r\n"
 				"%s\r\n"
-				"$6\r\n"
-				"learns\r\n"
+				"$%d\r\n"
+				"%s\r\n" /* Learned key */
 				"$1\r\n"
 				"1\r\n",
 				(gint)strlen (rt->redis_object_expanded),
-				rt->redis_object_expanded);
+				rt->redis_object_expanded,
+				(gint)strlen (learned_key),
+				learned_key);
 	}
 	else {
 		rspamd_printf_fstring (&query, ""
@@ -1562,12 +1745,14 @@ rspamd_redis_learn_tokens (struct rspamd_task *task, GPtrArray *tokens,
 				"HINCRBY\r\n"
 				"$%d\r\n"
 				"%s\r\n"
-				"$6\r\n"
-				"learns\r\n"
+				"$%d\r\n"
+				"%s\r\n" /* Learned key */
 				"$2\r\n"
 				"-1\r\n",
 				(gint)strlen (rt->redis_object_expanded),
-				rt->redis_object_expanded);
+				rt->redis_object_expanded,
+				(gint)strlen (learned_key),
+				learned_key);
 	}
 
 	ret = redisAsyncFormattedCommand (rt->redis, NULL, NULL,
@@ -1584,10 +1769,17 @@ rspamd_redis_learn_tokens (struct rspamd_task *task, GPtrArray *tokens,
 	ret = rspamd_printf_fstring (&query, "*1\r\n$4\r\nEXEC\r\n");
 	ret = redisAsyncFormattedCommand (rt->redis, rspamd_redis_learned, rt,
 			query->str + off, ret);
-
 	rspamd_mempool_add_destructor (task->task_pool,
 			(rspamd_mempool_destruct_t)rspamd_fstring_free, query);
+
 	if (ret == REDIS_OK) {
+
+		/* Add signature if needed */
+		if (rt->ctx->enable_signatures) {
+			rspamd_redis_store_stat_signature (task, rt, tokens,
+					"RSIG");
+		}
+
 		rspamd_session_add_event (task->s, rspamd_redis_fin_learn, rt,
 				rspamd_redis_stat_quark ());
 		rt->has_event = TRUE;
@@ -1611,9 +1803,9 @@ rspamd_redis_learn_tokens (struct rspamd_task *task, GPtrArray *tokens,
 }
 
 
-void
+gboolean
 rspamd_redis_finalize_learn (struct rspamd_task *task, gpointer runtime,
-		gpointer ctx)
+		gpointer ctx, GError **err)
 {
 	struct redis_stat_runtime *rt = REDIS_RUNTIME (runtime);
 	redisAsyncContext *redis;
@@ -1627,6 +1819,15 @@ rspamd_redis_finalize_learn (struct rspamd_task *task, gpointer runtime,
 		rt->redis = NULL;
 		redisAsyncFree (redis);
 	}
+
+	if (rt->err) {
+		g_propagate_error (err, rt->err);
+		rt->err = NULL;
+
+		return FALSE;
+	}
+
+	return TRUE;
 }
 
 gulong
