@@ -16,6 +16,7 @@
 #include "config.h"
 #include "addr.h"
 #include "util.h"
+#include "map_helpers.h"
 #include "logger.h"
 #include "cryptobox.h"
 #include "unix-std.h"
@@ -28,7 +29,7 @@
 #include <grp.h>
 #endif
 
-static radix_compressed_t *local_addrs;
+static struct rspamd_radix_map_helper *local_addrs;
 
 enum {
 	RSPAMD_IPV6_UNDEFINED = 0,
@@ -109,16 +110,17 @@ rspamd_inet_addr_create (gint af)
 {
 	rspamd_inet_addr_t *addr;
 
-	addr = g_slice_alloc0 (sizeof (rspamd_inet_addr_t));
-
-	if (af == AF_UNIX) {
-		addr->u.un = g_slice_alloc0 (sizeof (*addr->u.un));
-		addr->slen = sizeof (addr->u.un->addr);
-	}
+	addr = g_malloc0 (sizeof (rspamd_inet_addr_t));
 
 	addr->af = af;
 
-	rspamd_ip_validate_af (addr);
+	if (af == AF_UNIX) {
+		addr->u.un = g_malloc0 (sizeof (*addr->u.un));
+		addr->slen = sizeof (addr->u.un->addr);
+	}
+	else {
+		rspamd_ip_validate_af (addr);
+	}
 
 	return addr;
 }
@@ -129,10 +131,10 @@ rspamd_inet_address_free (rspamd_inet_addr_t *addr)
 	if (addr) {
 		if (addr->af == AF_UNIX) {
 			if (addr->u.un) {
-				g_slice_free1 (sizeof (*addr->u.un), addr->u.un);
+				g_free (addr->u.un);
 			}
 		}
-		g_slice_free1 (sizeof (rspamd_inet_addr_t), addr);
+		g_free (addr);
 	}
 }
 
@@ -344,8 +346,9 @@ rspamd_parse_unix_path (rspamd_inet_addr_t **target, const char *src)
 	struct passwd pw, *ppw;
 	struct group gr, *pgr;
 	rspamd_inet_addr_t *addr;
+	bool has_group = false;
 
-	tokens = g_strsplit_set (src, " ", -1);
+	tokens = g_strsplit_set (src, " ,", -1);
 
 	addr = rspamd_inet_addr_create (AF_UNIX);
 
@@ -369,7 +372,7 @@ rspamd_parse_unix_path (rspamd_inet_addr_t **target, const char *src)
 	pwlen = 8192;
 #endif
 
-	pwbuf = g_alloca (pwlen);
+	pwbuf = g_malloc0 (pwlen);
 
 	while (*cur_tok) {
 		if (g_ascii_strncasecmp (*cur_tok, "mode=", sizeof ("mode=") - 1) == 0) {
@@ -395,7 +398,10 @@ rspamd_parse_unix_path (rspamd_inet_addr_t **target, const char *src)
 				goto err;
 			}
 			addr->u.un->owner = pw.pw_uid;
-			addr->u.un->group = pw.pw_gid;
+
+			if (!has_group) {
+				addr->u.un->group = pw.pw_gid;
+			}
 		}
 		else if (g_ascii_strncasecmp (*cur_tok, "group=",
 				sizeof ("group=") - 1) == 0) {
@@ -408,11 +414,15 @@ rspamd_parse_unix_path (rspamd_inet_addr_t **target, const char *src)
 				}
 				goto err;
 			}
+
+			has_group = true;
 			addr->u.un->group = gr.gr_gid;
 		}
 		cur_tok ++;
 	}
 
+	g_free (pwbuf);
+
 	if (target) {
 		rspamd_ip_validate_af (addr);
 		*target = addr;
@@ -425,6 +435,7 @@ rspamd_parse_unix_path (rspamd_inet_addr_t **target, const char *src)
 
 err:
 
+	g_free (pwbuf);
 	rspamd_inet_address_free (addr);
 	return FALSE;
 }
@@ -707,8 +718,7 @@ rspamd_parse_inet_address (rspamd_inet_addr_t **target,
 
 		rspamd_strlcpy (ipbuf, src + 1, iplen + 1);
 
-		if (ipv6_status == RSPAMD_IPV6_SUPPORTED &&
-				rspamd_parse_inet_address_ip6 (ipbuf, iplen,
+		if (rspamd_parse_inet_address_ip6 (ipbuf, iplen,
 						&su.s6.sin6_addr)) {
 			addr = rspamd_inet_address_v6_maybe_map (&su.s6);
 			ret = TRUE;
@@ -726,7 +736,6 @@ rspamd_parse_inet_address (rspamd_inet_addr_t **target,
 			/* This is either port number and ipv4 addr or ipv6 addr */
 			/* Search for another semicolon */
 			if (memchr (end + 1, ':', srclen - (end - src + 1)) &&
-					ipv6_status == RSPAMD_IPV6_SUPPORTED &&
 					rspamd_parse_inet_address_ip6 (src, srclen, &su.s6.sin6_addr)) {
 				addr = rspamd_inet_address_v6_maybe_map (&su.s6);
 				ret = TRUE;
@@ -760,8 +769,7 @@ rspamd_parse_inet_address (rspamd_inet_addr_t **target,
 						sizeof (struct in_addr));
 				ret = TRUE;
 			}
-			else if (ipv6_status == RSPAMD_IPV6_SUPPORTED &&
-					rspamd_parse_inet_address_ip6 (src, srclen, &su.s6.sin6_addr)) {
+			else if (rspamd_parse_inet_address_ip6 (src, srclen, &su.s6.sin6_addr)) {
 				addr = rspamd_inet_address_v6_maybe_map (&su.s6);
 				ret = TRUE;
 			}
@@ -1096,7 +1104,7 @@ rspamd_inet_address_recvfrom (gint fd, void *buf, gsize len, gint fl,
 		addr->slen = slen;
 
 		if (addr->af == AF_UNIX) {
-			addr->u.un = g_slice_alloc (sizeof (*addr->u.un));
+			addr->u.un = g_malloc (sizeof (*addr->u.un));
 			memcpy (&addr->u.un->addr, &su.su, sizeof (struct sockaddr_un));
 		}
 		else {
@@ -1726,7 +1734,7 @@ rspamd_inet_address_is_local (const rspamd_inet_addr_t *addr,
 		}
 
 		if (check_laddrs && local_addrs) {
-			if (radix_find_compressed_addr (local_addrs, addr) != RADIX_NO_VALUE) {
+			if (rspamd_match_radix_map_addr (local_addrs, addr) != NULL) {
 				return TRUE;
 			}
 		}
@@ -1735,7 +1743,7 @@ rspamd_inet_address_is_local (const rspamd_inet_addr_t *addr,
 	return FALSE;
 }
 
-radix_compressed_t **
+struct rspamd_radix_map_helper **
 rspamd_inet_library_init (void)
 {
 	return &local_addrs;
@@ -1745,7 +1753,7 @@ void
 rspamd_inet_library_destroy (void)
 {
 	if (local_addrs != NULL) {
-		radix_destroy_compressed (local_addrs);
+		rspamd_map_helper_destroy_radix (local_addrs);
 	}
 }
 
