@@ -24,15 +24,6 @@
 #include "rspamd.h"
 #include "contrib/zstd/zstd.h"
 
-#ifdef WITH_HYPERSCAN
-#include "hs.h"
-#endif
-#ifndef WITH_PCRE2
-#include <pcre.h>
-#else
-#include <pcre2.h>
-#endif
-
 #undef MAP_DEBUG_REFS
 #ifdef MAP_DEBUG_REFS
 #define MAP_RETAIN(x, t) do { \
@@ -49,7 +40,6 @@
 #define MAP_RELEASE(x, t) REF_RELEASE(x)
 #endif
 
-static const gchar *hash_fill = "1";
 static void free_http_cbdata_common (struct http_callback_data *cbd, gboolean plan_new);
 static void free_http_cbdata_dtor (gpointer p);
 static void free_http_cbdata (struct http_callback_data *cbd);
@@ -66,6 +56,12 @@ struct rspamd_http_map_cached_cbdata {
 	time_t last_checked;
 };
 
+guint rspamd_map_log_id = (guint)-1;
+RSPAMD_CONSTRUCTOR(rspamd_map_log_init)
+{
+	rspamd_map_log_id = rspamd_logger_add_debug_module("map");
+}
+
 /**
  * Write HTTP request
  */
@@ -99,11 +95,17 @@ write_http_request (struct http_callback_data *cbd)
 			msg->url = rspamd_fstring_append (msg->url,
 					cbd->data->path, strlen (cbd->data->path));
 
-			if (cbd->check &&
-					cbd->data->last_modified != 0 && cbd->stage == map_load_file) {
-				rspamd_http_date_format (datebuf, sizeof (datebuf),
-						cbd->data->last_modified);
-				rspamd_http_message_add_header (msg, "If-Modified-Since", datebuf);
+			if (cbd->check && cbd->stage == map_load_file) {
+				if (cbd->data->last_modified != 0) {
+					rspamd_http_date_format (datebuf, sizeof (datebuf),
+							cbd->data->last_modified);
+					rspamd_http_message_add_header (msg, "If-Modified-Since",
+							datebuf);
+				}
+				if (cbd->data->etag) {
+					rspamd_http_message_add_header_len (msg, "If-None-Match",
+							cbd->data->etag->str, cbd->data->etag->len);
+				}
 			}
 		}
 		else if (cbd->stage == map_load_pubkey) {
@@ -277,7 +279,7 @@ free_http_cbdata_common (struct http_callback_data *cbd, gboolean plan_new)
 
 	MAP_RELEASE (cbd->bk, "rspamd_map_backend");
 	MAP_RELEASE (periodic, "periodic");
-	g_slice_free1 (sizeof (struct http_callback_data), cbd);
+	g_free (cbd);
 }
 
 static void
@@ -344,9 +346,9 @@ rspamd_map_cache_cb (gint fd, short what, gpointer ud)
 		/* We have another update, so this cache element is obviously expired */
 		/* Important: we do not set cache availability to zero here */
 		MAP_RELEASE (cache_cbd->shm, "rspamd_http_map_cached_cbdata");
-		msg_debug_map ("cached data is now expired (gen missmatch) for %s", map->name);
+		msg_debug_map ("cached data is now expired (gen mismatch) for %s", map->name);
 		event_del (&cache_cbd->timeout);
-		g_slice_free1 (sizeof (*cache_cbd), cache_cbd);
+		g_free (cache_cbd);
 	}
 	else if (cache_cbd->data->last_checked > cache_cbd->last_checked) {
 		/*
@@ -363,10 +365,65 @@ rspamd_map_cache_cb (gint fd, short what, gpointer ud)
 		MAP_RELEASE (cache_cbd->shm, "rspamd_http_map_cached_cbdata");
 		msg_debug_map ("cached data is now expired for %s", map->name);
 		event_del (&cache_cbd->timeout);
-		g_slice_free1 (sizeof (*cache_cbd), cache_cbd);
+		g_free (cache_cbd);
 	}
 }
 
+static gboolean
+rspamd_http_check_pubkey (struct http_callback_data *cbd,
+		struct rspamd_http_message *msg)
+{
+	const rspamd_ftok_t *pubkey_hdr;
+
+	pubkey_hdr = rspamd_http_message_find_header (msg, "Pubkey");
+
+	if (pubkey_hdr) {
+		cbd->pk = rspamd_pubkey_from_base32 (pubkey_hdr->begin,
+				pubkey_hdr->len,
+				RSPAMD_KEYPAIR_SIGN, RSPAMD_CRYPTOBOX_MODE_25519);
+
+		return cbd->pk != NULL;
+	}
+
+	return FALSE;
+}
+
+static gboolean
+rspamd_http_check_signature (struct rspamd_map *map,
+		struct http_callback_data *cbd,
+		struct rspamd_http_message *msg)
+{
+	const rspamd_ftok_t *sig_hdr;
+	guchar *in;
+	size_t dlen;
+
+	sig_hdr = rspamd_http_message_find_header (msg, "Signature");
+
+	if (sig_hdr && cbd->pk) {
+		in = rspamd_shmem_xmap (cbd->shmem_data->shm_name, PROT_READ, &dlen);
+
+		if (in == NULL) {
+			msg_err_map ("cannot read tempfile %s: %s",
+					cbd->shmem_data->shm_name,
+					strerror (errno));
+			return FALSE;
+		}
+
+		if (!rspamd_map_check_sig_pk_mem (sig_hdr->begin, sig_hdr->len,
+				map, in,
+				cbd->data_len, cbd->pk)) {
+			munmap (in, dlen);
+			return FALSE;
+		}
+
+		munmap (in, dlen);
+
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
 static int
 http_map_finish (struct rspamd_http_connection *conn,
 		struct rspamd_http_message *msg)
@@ -376,7 +433,8 @@ http_map_finish (struct rspamd_http_connection *conn,
 	struct rspamd_map_backend *bk;
 	struct rspamd_http_map_cached_cbdata *cache_cbd;
 	struct timeval tv;
-	const rspamd_ftok_t *expires_hdr;
+	const rspamd_ftok_t *expires_hdr, *etag_hdr;
+	char next_check_date[128];
 	guchar *aux_data, *in = NULL;
 	gsize inlen = 0, dlen = 0;
 
@@ -408,17 +466,30 @@ http_map_finish (struct rspamd_http_connection *conn,
 			/* Maybe we need to check signature ? */
 			if (bk->is_signed) {
 
-				if (bk->trusted_pubkey) {
-					/* No need to load key */
-					cbd->stage = map_load_signature;
-					cbd->pk = rspamd_pubkey_ref (bk->trusted_pubkey);
-				}
-				else {
-					cbd->stage = map_load_pubkey;
-				}
-
 				cbd->shmem_data = rspamd_http_message_shmem_ref (msg);
 				cbd->data_len = msg->body_buf.len;
+
+				if (bk->trusted_pubkey) {
+					/* No need to load key */
+					cbd->pk = rspamd_pubkey_ref (bk->trusted_pubkey);
+					cbd->stage = map_load_signature;
+				}
+				else {
+					if (!rspamd_http_check_pubkey (cbd, msg)) {
+						cbd->stage = map_load_pubkey;
+					}
+					else {
+						cbd->stage = map_load_signature;
+					}
+				}
+
+				if (cbd->stage == map_load_signature) {
+					/* Try HTTP header */
+					if (rspamd_http_check_signature (map, cbd, msg)) {
+						goto read_data;
+					}
+				}
+
 				rspamd_http_connection_reset (cbd->conn);
 				write_http_request (cbd);
 				MAP_RELEASE (cbd, "http_callback_data");
@@ -557,6 +628,25 @@ read_data:
 			double_to_tv (map->poll_timeout * 2, &tv);
 		}
 
+		/* Check for etag */
+		etag_hdr = rspamd_http_message_find_header (msg, "ETag");
+
+		if (etag_hdr) {
+			if (cbd->data->etag) {
+				/* Remove old etag */
+				rspamd_fstring_free (cbd->data->etag);
+				cbd->data->etag = rspamd_fstring_new_init (etag_hdr->begin,
+						etag_hdr->len);
+			}
+		}
+		else {
+			if (cbd->data->etag) {
+				/* Remove and clear old etag */
+				rspamd_fstring_free (cbd->data->etag);
+				cbd->data->etag = NULL;
+			}
+		}
+
 		MAP_RETAIN (cbd->shmem_data, "shmem_data");
 		cbd->data->gen ++;
 		/*
@@ -568,7 +658,7 @@ read_data:
 				sizeof (map->cache->shmem_name));
 		map->cache->len = cbd->data_len;
 		map->cache->last_modified = cbd->data->last_modified;
-		cache_cbd = g_slice_alloc0 (sizeof (*cache_cbd));
+		cache_cbd = g_malloc0 (sizeof (*cache_cbd));
 		cache_cbd->shm = cbd->shmem_data;
 		cache_cbd->map = map;
 		cache_cbd->data = cbd->data;
@@ -581,6 +671,15 @@ read_data:
 		event_base_set (cbd->ev_base, &cache_cbd->timeout);
 		event_add (&cache_cbd->timeout, &tv);
 
+		if (map->next_check) {
+			rspamd_http_date_format (next_check_date, sizeof (next_check_date),
+					map->next_check);
+		}
+		else {
+			rspamd_http_date_format (next_check_date, sizeof (next_check_date),
+					time (NULL) + map->poll_timeout);
+		}
+
 
 		if (cbd->bk->is_compressed) {
 			ZSTD_DStream *zstream;
@@ -630,18 +729,18 @@ read_data:
 
 			ZSTD_freeDStream (zstream);
 			msg_info_map ("%s(%s): read map data %z bytes compressed, "
-					"%z uncompressed",
+					"%z uncompressed, next check at %s",
 					cbd->bk->uri,
 					rspamd_inet_address_to_string_pretty (cbd->addr),
-					dlen, zout.pos);
+					dlen, zout.pos, next_check_date);
 			map->read_callback (out, zout.pos, &cbd->periodic->cbdata, TRUE);
 			g_free (out);
 		}
 		else {
-			msg_info_map ("%s(%s): read map data %z bytes",
+			msg_info_map ("%s(%s): read map data %z bytes, next check at %s",
 					cbd->bk->uri,
 					rspamd_inet_address_to_string_pretty (cbd->addr),
-					dlen);
+					dlen, next_check_date);
 			map->read_callback (in, cbd->data_len, &cbd->periodic->cbdata, TRUE);
 		}
 
@@ -652,9 +751,6 @@ read_data:
 		rspamd_map_periodic_callback (-1, EV_TIMEOUT, cbd->periodic);
 	}
 	else if (msg->code == 304 && (cbd->check && cbd->stage == map_load_file)) {
-		msg_debug_map ("data is not modified for server %s",
-				cbd->data->host);
-
 		cbd->data->last_checked = msg->date;
 
 		if (msg->last_modified) {
@@ -681,12 +777,35 @@ read_data:
 			}
 		}
 
+		etag_hdr = rspamd_http_message_find_header (msg, "ETag");
+
+		if (etag_hdr) {
+			if (cbd->data->etag) {
+				/* Remove old etag */
+				rspamd_fstring_free (cbd->data->etag);
+				cbd->data->etag = rspamd_fstring_new_init (etag_hdr->begin,
+						etag_hdr->len);
+			}
+		}
+
+		if (map->next_check) {
+			rspamd_http_date_format (next_check_date, sizeof (next_check_date),
+					map->next_check);
+		}
+		else {
+			rspamd_http_date_format (next_check_date, sizeof (next_check_date),
+					time (NULL) + map->poll_timeout);
+		}
+		msg_info_map ("data is not modified for server %s, next check at %s",
+				cbd->data->host, next_check_date);
+
 		cbd->periodic->cur_backend ++;
 		rspamd_map_periodic_callback (-1, EV_TIMEOUT, cbd->periodic);
 	}
 	else {
 		msg_info_map ("cannot load map %s from %s: HTTP error %d",
 				bk->uri, cbd->data->host, msg->code);
+		goto err;
 	}
 
 	MAP_RELEASE (cbd, "http_callback_data");
@@ -718,9 +837,17 @@ read_map_file (struct rspamd_map *map, struct file_map_data *data,
 
 	if (access (data->filename, R_OK) == -1) {
 		/* File does not exist, skipping */
-		msg_err_map ("%s: map file is unavailable for reading",
-				data->filename);
-		return TRUE;
+		if (errno != ENOENT) {
+			msg_err_map ("%s: map file is unavailable for reading: %s",
+					data->filename, strerror (errno));
+
+			return FALSE;
+		}
+		else {
+			msg_info_map ("%s: map file is not found",
+					data->filename);
+			return TRUE;
+		}
 	}
 
 	bytes = rspamd_file_xmap (data->filename, PROT_READ, &len, TRUE);
@@ -918,7 +1045,7 @@ rspamd_map_periodic_dtor (struct map_periodic_cbdata *periodic)
 		msg_debug_map ("unlocked map");
 	}
 
-	g_slice_free1 (sizeof (*periodic), periodic);
+	g_free (periodic);
 }
 
 static void
@@ -930,6 +1057,11 @@ rspamd_map_schedule_periodic (struct rspamd_map *map,
 	gdouble timeout;
 	struct map_periodic_cbdata *cbd;
 
+	if (map->scheduled_check) {
+		/* Do not schedule check if some check is already scheduled */
+		return;
+	}
+
 	if (map->next_check != 0) {
 		timeout = map->next_check - rspamd_get_calendar_ticks ();
 
@@ -969,12 +1101,13 @@ rspamd_map_schedule_periodic (struct rspamd_map *map,
 		jittered_sec = rspamd_time_jitter (timeout, 0);
 	}
 
-	cbd = g_slice_alloc0 (sizeof (*cbd));
+	cbd = g_malloc0 (sizeof (*cbd));
 	cbd->cbdata.state = 0;
 	cbd->cbdata.prev_data = *map->user_data;
 	cbd->cbdata.cur_data = NULL;
 	cbd->cbdata.map = map;
 	cbd->map = map;
+	map->scheduled_check = TRUE;
 	REF_INIT_RETAIN (cbd, rspamd_map_periodic_dtor);
 
 	evtimer_set (&cbd->ev, rspamd_map_periodic_callback, cbd);
@@ -984,7 +1117,6 @@ rspamd_map_schedule_periodic (struct rspamd_map *map,
 	msg_debug_map ("schedule new periodic event %p in %.2f seconds",
 			cbd, jittered_sec);
 	double_to_tv (jittered_sec, &map->tv);
-
 	evtimer_add (&cbd->ev, &map->tv);
 }
 
@@ -1043,6 +1175,8 @@ rspamd_map_dns_callback (struct rdns_reply *reply, void *arg)
 		else {
 			/* We could not resolve host, so cowardly fail here */
 			msg_err_map ("cannot resolve %s", cbd->data->host);
+			cbd->periodic->errored = 1;
+			rspamd_map_periodic_callback (-1, EV_TIMEOUT, cbd->periodic);
 		}
 	}
 
@@ -1197,7 +1331,7 @@ rspamd_map_common_http_callback (struct rspamd_map *map,
 	}
 
 check:
-	cbd = g_slice_alloc0 (sizeof (struct http_callback_data));
+	cbd = g_malloc0 (sizeof (struct http_callback_data));
 
 	cbd->ev_base = map->ev_base;
 	cbd->map = map;
@@ -1406,6 +1540,7 @@ rspamd_map_periodic_callback (gint fd, short what, void *ud)
 	struct rspamd_map *map;
 
 	map = cbd->map;
+	map->scheduled_check = FALSE;
 
 	if (!cbd->locked) {
 		if (!g_atomic_int_compare_and_exchange (cbd->map->locked, 0, 1)) {
@@ -1500,6 +1635,15 @@ rspamd_map_watch (struct rspamd_config *cfg, struct event_base *ev_base,
 			map->active_http = active_http;
 		}
 
+		if (!map->active_http) {
+			/* Check cached version more frequently as it is cheap */
+
+			if (map->poll_timeout >= cfg->map_timeout &&
+					cfg->map_file_watch_multiplier < 1.0) {
+				map->poll_timeout = map->poll_timeout * cfg->map_file_watch_multiplier;
+			}
+		}
+
 		rspamd_map_schedule_periodic (map, FALSE, TRUE, FALSE);
 
 		cur = g_list_next (cur);
@@ -1671,7 +1815,7 @@ rspamd_map_backend_dtor (struct rspamd_map_backend *bk)
 	case MAP_PROTO_FILE:
 		if (bk->data.fd) {
 			g_free (bk->data.fd->filename);
-			g_slice_free1 (sizeof (*bk->data.fd), bk->data.fd);
+			g_free (bk->data.fd);
 		}
 		break;
 	case MAP_PROTO_STATIC:
@@ -1686,7 +1830,12 @@ rspamd_map_backend_dtor (struct rspamd_map_backend *bk)
 		if (bk->data.hd) {
 			g_free (bk->data.hd->host);
 			g_free (bk->data.hd->path);
-			g_slice_free1 (sizeof (*bk->data.hd), bk->data.hd);
+
+			if (bk->data.hd->etag) {
+				rspamd_fstring_free (bk->data.hd->etag);
+			}
+
+			g_free (bk->data.hd);
 		}
 		break;
 	}
@@ -1695,7 +1844,7 @@ rspamd_map_backend_dtor (struct rspamd_map_backend *bk)
 		rspamd_pubkey_unref (bk->trusted_pubkey);
 	}
 
-	g_slice_free1 (sizeof (*bk), bk);
+	g_free (bk);
 }
 
 static struct rspamd_map_backend *
@@ -1709,7 +1858,7 @@ rspamd_map_parse_backend (struct rspamd_config *cfg, const gchar *map_line)
 	const gchar *end, *p;
 	rspamd_ftok_t tok;
 
-	bk = g_slice_alloc0 (sizeof (*bk));
+	bk = g_malloc0 (sizeof (*bk));
 	REF_INIT_RETAIN (bk, rspamd_map_backend_dtor);
 
 	if (!rspamd_map_check_proto (cfg, map_line, bk)) {
@@ -1730,7 +1879,7 @@ rspamd_map_parse_backend (struct rspamd_config *cfg, const gchar *map_line)
 
 	/* Now check for each proto separately */
 	if (bk->protocol == MAP_PROTO_FILE) {
-		fdata = g_slice_alloc0 (sizeof (struct file_map_data));
+		fdata = g_malloc0 (sizeof (struct file_map_data));
 		fdata->st.st_mtime = -1;
 
 		if (access (bk->uri, R_OK) == -1) {
@@ -1748,7 +1897,7 @@ rspamd_map_parse_backend (struct rspamd_config *cfg, const gchar *map_line)
 		bk->data.fd = fdata;
 	}
 	else if (bk->protocol == MAP_PROTO_HTTP || bk->protocol == MAP_PROTO_HTTPS) {
-		hdata = g_slice_alloc0 (sizeof (struct http_map_data));
+		hdata = g_malloc0 (sizeof (struct http_map_data));
 
 		memset (&up, 0, sizeof (up));
 		if (http_parser_parse_url (bk->uri, strlen (bk->uri), FALSE,
@@ -1788,7 +1937,7 @@ rspamd_map_parse_backend (struct rspamd_config *cfg, const gchar *map_line)
 
 		bk->data.hd = hdata;
 	}else if (bk->protocol == MAP_PROTO_STATIC) {
-		sdata = g_slice_alloc0 (sizeof (*sdata));
+		sdata = g_malloc0 (sizeof (*sdata));
 		bk->data.sd = sdata;
 	}
 
@@ -1801,7 +1950,7 @@ err:
 	MAP_RELEASE (bk, "rspamd_map_backend");
 
 	if (hdata) {
-		g_slice_free1 (sizeof (*hdata), hdata);
+		g_free (hdata);
 	}
 
 	return NULL;
@@ -1828,6 +1977,35 @@ rspamd_map_calculate_hash (struct rspamd_map *map)
 	g_free (cksum_encoded);
 }
 
+static gboolean
+rspamd_map_add_static_string (struct rspamd_config *cfg,
+		const ucl_object_t *elt,
+		GString *target)
+{
+	gsize sz;
+	const gchar *dline;
+
+	if (ucl_object_type (elt) != UCL_STRING) {
+		msg_err_config ("map has static backend but `data` is "
+						"not string like: %s",
+				ucl_object_type_to_string (elt->type));
+		return FALSE;
+	}
+
+	/* Otherwise, we copy data to the backend */
+	dline = ucl_object_tolstring (elt, &sz);
+
+	if (sz == 0) {
+		msg_err_config ("map has static backend but empty no data");
+		return FALSE;
+	}
+
+	g_string_append_len (target, dline, sz);
+	g_string_append_c (target, '\n');
+
+	return TRUE;
+}
+
 struct rspamd_map *
 rspamd_map_add (struct rspamd_config *cfg,
 	const gchar *map_line,
@@ -1890,8 +2068,6 @@ rspamd_map_add_from_ucl (struct rspamd_config *cfg,
 	const ucl_object_t *cur, *elt;
 	struct rspamd_map *map;
 	struct rspamd_map_backend *bk;
-	gsize sz;
-	const gchar *dline;
 	guint i;
 
 	g_assert (obj != NULL);
@@ -1929,7 +2105,8 @@ rspamd_map_add_from_ucl (struct rspamd_config *cfg,
 
 				if (bk != NULL) {
 					if (bk->protocol == MAP_PROTO_FILE) {
-						map->poll_timeout = (map->poll_timeout * cfg->map_file_watch_multiplier);
+						map->poll_timeout = (map->poll_timeout *
+								cfg->map_file_watch_multiplier);
 					}
 					g_ptr_array_add (map->backends, bk);
 
@@ -2033,25 +2210,49 @@ rspamd_map_add_from_ucl (struct rspamd_config *cfg,
 
 		PTR_ARRAY_FOREACH (map->backends, i, bk) {
 			if (bk->protocol == MAP_PROTO_STATIC) {
+				GString *map_data;
 				/* We need data field in ucl */
 				elt = ucl_object_lookup (obj, "data");
 
-				if (elt == NULL || ucl_object_type (elt) != UCL_STRING) {
+				if (elt == NULL) {
 					msg_err_config ("map has static backend but no `data` field");
 					goto err;
 				}
 
-				/* Otherwise, we copy data to the backend */
-				dline = ucl_object_tolstring (elt, &sz);
 
-				if (sz == 0) {
-					msg_err_config ("map has static backend but empty `data` field");
-					goto err;
+				if (ucl_object_type (elt) == UCL_STRING) {
+					map_data = g_string_sized_new (32);
+
+					if (rspamd_map_add_static_string (cfg, elt, map_data)) {
+						bk->data.sd->data = map_data->str;
+						bk->data.sd->len = map_data->len;
+						g_string_free (map_data, FALSE);
+					}
+					else {
+						g_string_free (map_data, TRUE);
+						msg_err_config ("map has static backend with invalid `data` field");
+						goto err;
+					}
 				}
+				else if (ucl_object_type (elt) == UCL_ARRAY) {
+					map_data = g_string_sized_new (32);
+					it = ucl_object_iterate_new (elt);
 
-				bk->data.sd->data = g_malloc (sz);
-				bk->data.sd->len = sz;
-				memcpy (bk->data.sd->data, dline, sz);
+					while ((cur = ucl_object_iterate_safe (it, true))) {
+						if (!rspamd_map_add_static_string (cfg, cur, map_data)) {
+							g_string_free (map_data, TRUE);
+							msg_err_config ("map has static backend with invalid "
+											"`data` field");
+							ucl_object_iterate_free (it);
+							goto err;
+						}
+					}
+
+					ucl_object_iterate_free (it);
+					bk->data.sd->data = map_data->str;
+					bk->data.sd->len = map_data->len;
+					g_string_free (map_data, FALSE);
+				}
 			}
 		}
 	}
@@ -2062,7 +2263,7 @@ rspamd_map_add_from_ucl (struct rspamd_config *cfg,
 	}
 
 	rspamd_map_calculate_hash (map);
-	msg_info_map ("added map from ucl");
+	msg_debug_map ("added map from ucl");
 
 	cfg->maps = g_list_prepend (cfg->maps, map);
 
@@ -2073,907 +2274,21 @@ err:
 	return NULL;
 }
 
-/**
- * FSM for parsing lists
- */
-
-#define MAP_STORE_KEY do { \
-	while (g_ascii_isspace (*c) && p > c) { c ++; } \
-	key = g_malloc (p - c + 1); \
-	rspamd_strlcpy (key, c, p - c + 1); \
-	key = g_strchomp (key); \
-} while (0)
-
-#define MAP_STORE_VALUE do { \
-	while (g_ascii_isspace (*c) && p > c) { c ++; } \
-	value = g_malloc (p - c + 1); \
-	rspamd_strlcpy (value, c, p - c + 1); \
-	value = g_strchomp (value); \
-} while (0)
-
-gchar *
-rspamd_parse_kv_list (
-	gchar * chunk,
-	gint len,
-	struct map_cb_data *data,
-	insert_func func,
-	const gchar *default_value,
-	gboolean final)
+rspamd_map_traverse_function
+rspamd_map_get_traverse_function (struct rspamd_map *map)
 {
-	enum {
-		map_skip_spaces_before_key = 0,
-		map_read_key,
-		map_read_key_quoted,
-		map_read_key_slashed,
-		map_skip_spaces_after_key,
-		map_backslash_quoted,
-		map_backslash_slashed,
-		map_read_key_after_slash,
-		map_read_value,
-		map_read_comment_start,
-		map_skip_comment,
-		map_read_eol,
-	};
-
-	gchar *c, *p, *key = NULL, *value = NULL, *end;
-	struct rspamd_map *map = data->map;
-
-	p = chunk;
-	c = p;
-	end = p + len;
-
-	while (p < end) {
-		switch (data->state) {
-		case map_skip_spaces_before_key:
-			if (g_ascii_isspace (*p)) {
-				p ++;
-			}
-			else {
-				if (*p == '"') {
-					p++;
-					c = p;
-					data->state = map_read_key_quoted;
-				}
-				else if (*p == '/') {
-					/* Note that c is on '/' here as '/' is a part of key */
-					c = p;
-					p++;
-					data->state = map_read_key_slashed;
-				}
-				else {
-					c = p;
-					data->state = map_read_key;
-				}
-			}
-			break;
-		case map_read_key:
-			/* read key */
-			/* Check here comments, eol and end of buffer */
-			if (*p == '#') {
-				if (p - c > 0) {
-					/* Store a single key */
-					MAP_STORE_KEY;
-					func (data->cur_data, key, default_value);
-					msg_debug_map ("insert key only pair: %s -> %s",
-							key, default_value);
-					g_free (key);
-				}
-
-				key = NULL;
-				data->state = map_read_comment_start;
-			}
-			else if (*p == '\r' || *p == '\n') {
-				if (p - c > 0) {
-					/* Store a single key */
-					MAP_STORE_KEY;
-					func (data->cur_data, key, default_value);
-					msg_debug_map ("insert key only pair: %s -> %s",
-							key, default_value);
-					g_free (key);
-				}
-
-				data->state = map_read_eol;
-				key = NULL;
-			}
-			else if (g_ascii_isspace (*p)) {
-				if (p - c > 0) {
-					MAP_STORE_KEY;
-					data->state = map_skip_spaces_after_key;
-				}
-				else {
-					/* Should not happen */
-					g_assert_not_reached ();
-				}
-			}
-			else {
-				p++;
-			}
-			break;
-		case map_read_key_quoted:
-			if (*p == '\\') {
-				data->state = map_backslash_quoted;
-				p ++;
-			}
-			else if (*p == '"') {
-				/* Allow empty keys in this case */
-				if (p - c >= 0) {
-					MAP_STORE_KEY;
-					data->state = map_skip_spaces_after_key;
-				}
-				else {
-					g_assert_not_reached ();
-				}
-				p ++;
-			}
-			else {
-				p ++;
-			}
-			break;
-		case map_read_key_slashed:
-			if (*p == '\\') {
-				data->state = map_backslash_slashed;
-				p ++;
-			}
-			else if (*p == '/') {
-				/* Allow empty keys in this case */
-				if (p - c >= 0) {
-					data->state = map_read_key_after_slash;
-				}
-				else {
-					g_assert_not_reached ();
-				}
-			}
-			else {
-				p ++;
-			}
-			break;
-		case map_read_key_after_slash:
-			/*
-			 * This state is equal to reading of key but '/' is not
-			 * treated specially
-			 */
-			if (*p == '#') {
-				if (p - c > 0) {
-					/* Store a single key */
-					MAP_STORE_KEY;
-					func (data->cur_data, key, default_value);
-					msg_debug_map ("insert key only pair: %s -> %s",
-							key, default_value);
-					g_free (key);
-					key = NULL;
-				}
-
-				data->state = map_read_comment_start;
-			}
-			else if (*p == '\r' || *p == '\n') {
-				if (p - c > 0) {
-					/* Store a single key */
-					MAP_STORE_KEY;
-					func (data->cur_data, key, default_value);
-
-					msg_debug_map ("insert key only pair: %s -> %s",
-							key, default_value);
-					g_free (key);
-					key = NULL;
-				}
-
-				data->state = map_read_eol;
-				key = NULL;
-			}
-			else if (g_ascii_isspace (*p)) {
-				if (p - c > 0) {
-					MAP_STORE_KEY;
-					data->state = map_skip_spaces_after_key;
-				}
-				else {
-					/* Should not happen */
-					g_assert_not_reached ();
-				}
-			}
-			else {
-				p ++;
-			}
-			break;
-		case map_backslash_quoted:
-			p ++;
-			data->state = map_read_key_quoted;
-			break;
-		case map_backslash_slashed:
-			p ++;
-			data->state = map_read_key_slashed;
-			break;
-		case map_skip_spaces_after_key:
-			if (*p == ' ' || *p == '\t') {
-				p ++;
-			}
-			else {
-				c = p;
-				data->state = map_read_value;
-			}
-			break;
-		case map_read_value:
-			g_assert (key != NULL);
-			if (*p == '#') {
-				if (p - c > 0) {
-					/* Store a single key */
-					MAP_STORE_VALUE;
-					func (data->cur_data, key, value);
-					msg_debug_map ("insert key value pair: %s -> %s",
-							key, value);
-					g_free (key);
-					g_free (value);
-					key = NULL;
-					value = NULL;
-				}
-				else {
-					func (data->cur_data, key, default_value);
-					msg_debug_map ("insert key only pair: %s -> %s",
-							key, default_value);
-					g_free (key);
-					key = NULL;
-				}
-
-				data->state = map_read_comment_start;
-			}
-			else if (*p == '\r' || *p == '\n') {
-				if (p - c > 0) {
-					/* Store a single key */
-					MAP_STORE_VALUE;
-					func (data->cur_data, key, value);
-					msg_debug_map ("insert key value pair: %s -> %s",
-							key, value);
-					g_free (key);
-					g_free (value);
-					key = NULL;
-					value = NULL;
-				}
-				else {
-					func (data->cur_data, key, default_value);
-					msg_debug_map ("insert key only pair: %s -> %s",
-							key, default_value);
-					g_free (key);
-					key = NULL;
-				}
-
-				data->state = map_read_eol;
-				key = NULL;
-			}
-			else {
-				p ++;
-			}
-			break;
-		case map_read_comment_start:
-			if (*p == '#') {
-				data->state = map_skip_comment;
-				p ++;
-				key = NULL;
-				value = NULL;
-			}
-			else {
-				g_assert_not_reached ();
-			}
-			break;
-		case map_skip_comment:
-			if (*p == '\r' || *p == '\n') {
-				data->state = map_read_eol;
-			}
-			else {
-				p ++;
-			}
-			break;
-		case map_read_eol:
-			/* Skip \r\n and whitespaces */
-			if (*p == '\r' || *p == '\n') {
-				p++;
-			}
-			else {
-				data->state = map_skip_spaces_before_key;
-			}
-			break;
-		default:
-			g_assert_not_reached ();
-			break;
-		}
+	if (map) {
+		return map->traverse_function;
 	}
 
-	if (final) {
-		/* Examine the state */
-		switch (data->state) {
-		case map_read_key:
-			if (p - c > 0) {
-				/* Store a single key */
-				MAP_STORE_KEY;
-				func (data->cur_data, key, default_value);
-				msg_debug_map ("insert key only pair: %s -> %s",
-						key, default_value);
-				g_free (key);
-				key = NULL;
-			}
-			break;
-		case map_read_value:
-			g_assert (key != NULL);
-			if (p - c > 0) {
-				/* Store a single key */
-				MAP_STORE_VALUE;
-				func (data->cur_data, key, value);
-				msg_debug_map ("insert key value pair: %s -> %s",
-						key, value);
-				g_free (key);
-				g_free (value);
-				key = NULL;
-				value = NULL;
-			}
-			else {
-				func (data->cur_data, key, default_value);
-				msg_debug_map ("insert key only pair: %s -> %s",
-						key, default_value);
-				g_free (key);
-				key = NULL;
-			}
-			break;
-		}
-	}
-
-	return c;
-}
-
-/**
- * Radix tree helper function
- */
-static void
-radix_tree_insert_helper (gpointer st, gconstpointer key, gconstpointer value)
-{
-	radix_compressed_t *tree = (radix_compressed_t *)st;
-	rspamd_mempool_t *pool;
-	gpointer nvalue;
-
-	pool = radix_get_pool (tree);
-	nvalue = rspamd_mempool_strdup (pool, value);
-	rspamd_radix_add_iplist (key, ",", tree, nvalue, FALSE);
-}
-
-static void
-hash_insert_helper (gpointer st, gconstpointer key, gconstpointer value)
-{
-	GHashTable *ht = st;
-	gpointer k, v;
-
-	k = g_strdup (key);
-	v = g_strdup (value);
-	g_hash_table_replace (ht, k, v);
-}
-
-/* Helpers */
-gchar *
-rspamd_hosts_read (
-	gchar * chunk,
-	gint len,
-	struct map_cb_data *data,
-	gboolean final)
-{
-	if (data->cur_data == NULL) {
-		data->cur_data = g_hash_table_new_full (rspamd_strcase_hash,
-				rspamd_strcase_equal, g_free, g_free);
-	}
-	return rspamd_parse_kv_list (
-			   chunk,
-			   len,
-			   data,
-			   hash_insert_helper,
-			   hash_fill,
-			   final);
-}
-
-void
-rspamd_hosts_fin (struct map_cb_data *data)
-{
-	struct rspamd_map *map = data->map;
-
-	if (data->prev_data) {
-		g_hash_table_unref (data->prev_data);
-	}
-	if (data->cur_data) {
-		msg_info_map ("read hash of %d elements", g_hash_table_size
-				(data->cur_data));
-	}
-}
-
-gchar *
-rspamd_kv_list_read (
-	gchar * chunk,
-	gint len,
-	struct map_cb_data *data,
-	gboolean final)
-{
-	if (data->cur_data == NULL) {
-		data->cur_data = g_hash_table_new_full (rspamd_strcase_hash,
-				rspamd_strcase_equal, g_free, g_free);
-	}
-	return rspamd_parse_kv_list (
-			   chunk,
-			   len,
-			   data,
-			   hash_insert_helper,
-			   "",
-			   final);
-}
-
-void
-rspamd_kv_list_fin (struct map_cb_data *data)
-{
-	struct rspamd_map *map = data->map;
-
-	if (data->prev_data) {
-		g_hash_table_unref (data->prev_data);
-	}
-	if (data->cur_data) {
-		msg_info_map ("read hash of %d elements", g_hash_table_size
-				(data->cur_data));
-	}
-}
-
-gchar *
-rspamd_radix_read (
-	gchar * chunk,
-	gint len,
-	struct map_cb_data *data,
-	gboolean final)
-{
-	radix_compressed_t *tree;
-	rspamd_mempool_t *rpool;
-	struct rspamd_map *map = data->map;
-
-	if (data->cur_data == NULL) {
-		tree = radix_create_compressed ();
-		rpool = radix_get_pool (tree);
-		memcpy (rpool->tag.uid, map->tag, sizeof (rpool->tag.uid));
-		data->cur_data = tree;
-	}
-	return rspamd_parse_kv_list (
-			   chunk,
-			   len,
-			   data,
-			   radix_tree_insert_helper,
-			   hash_fill,
-			   final);
-}
-
-void
-rspamd_radix_fin (struct map_cb_data *data)
-{
-	struct rspamd_map *map = data->map;
-
-	if (data->prev_data) {
-		radix_destroy_compressed (data->prev_data);
-	}
-	if (data->cur_data) {
-		msg_info_map ("read radix trie of %z elements: %s",
-				radix_get_size (data->cur_data), radix_get_info (data->cur_data));
-	}
-}
-
-enum rspamd_regexp_map_flags {
-	RSPAMD_REGEXP_FLAG_UTF = (1 << 0),
-	RSPAMD_REGEXP_FLAG_MULTIPLE = (1 << 1)
-};
-
-struct rspamd_regexp_map {
-	struct rspamd_map *map;
-	GPtrArray *regexps;
-	GPtrArray *values;
-	enum rspamd_regexp_map_flags map_flags;
-#ifdef WITH_HYPERSCAN
-	hs_database_t *hs_db;
-	hs_scratch_t *hs_scratch;
-	const gchar **patterns;
-	gint *flags;
-	gint *ids;
-#endif
-};
-
-static struct rspamd_regexp_map *
-rspamd_regexp_map_create (struct rspamd_map *map,
-		enum rspamd_regexp_map_flags flags)
-{
-	struct rspamd_regexp_map *re_map;
-
-	re_map = g_slice_alloc0 (sizeof (*re_map));
-	re_map->values = g_ptr_array_new ();
-	re_map->regexps = g_ptr_array_new ();
-	re_map->map = map;
-	re_map->map_flags = flags;
-
-	return re_map;
-}
-
-
-static void
-rspamd_regexp_map_destroy (struct rspamd_regexp_map *re_map)
-{
-	rspamd_regexp_t *re;
-	guint i;
-
-	for (i = 0; i < re_map->regexps->len; i ++) {
-		re = g_ptr_array_index (re_map->regexps, i);
-		rspamd_regexp_unref (re);
-	}
-
-	for (i = 0; i < re_map->values->len; i ++) {
-		g_free (g_ptr_array_index (re_map->values, i));
-	}
-
-	g_ptr_array_free (re_map->regexps, TRUE);
-	g_ptr_array_free (re_map->values, TRUE);
-
-#ifdef WITH_HYPERSCAN
-	if (re_map->hs_scratch) {
-		hs_free_scratch (re_map->hs_scratch);
-	}
-	if (re_map->hs_db) {
-		hs_free_database (re_map->hs_db);
-	}
-	if (re_map->patterns) {
-		g_free (re_map->patterns);
-	}
-	if (re_map->flags) {
-		g_free (re_map->flags);
-	}
-	if (re_map->ids) {
-		g_free (re_map->ids);
-	}
-#endif
-
-	g_slice_free1 (sizeof (*re_map), re_map);
-}
-
-static void
-rspamd_re_map_insert_helper (gpointer st, gconstpointer key, gconstpointer value)
-{
-	struct rspamd_regexp_map *re_map = st;
-	struct rspamd_map *map;
-	rspamd_regexp_t *re;
-	GError *err = NULL;
-	gint pcre_flags;
-
-	map = re_map->map;
-	re = rspamd_regexp_new (key, NULL, &err);
-
-	if (re == NULL) {
-		msg_err_map ("cannot parse regexp %s: %e", key, err);
-
-		if (err) {
-			g_error_free (err);
-		}
-
-		return;
-	}
-
-	pcre_flags = rspamd_regexp_get_pcre_flags (re);
-
-#ifndef WITH_PCRE2
-	if (pcre_flags & PCRE_FLAG(UTF8)) {
-		re_map->map_flags |= RSPAMD_REGEXP_FLAG_UTF;
-	}
-#else
-	if (pcre_flags & PCRE_FLAG(UTF)) {
-		re_map->map_flags |= RSPAMD_REGEXP_FLAG_UTF;
-	}
-#endif
-
-	g_ptr_array_add (re_map->regexps, re);
-	g_ptr_array_add (re_map->values, g_strdup (value));
-}
-
-static void
-rspamd_re_map_finalize (struct rspamd_regexp_map *re_map)
-{
-#ifdef WITH_HYPERSCAN
-	guint i;
-	hs_platform_info_t plt;
-	hs_compile_error_t *err;
-	struct rspamd_map *map;
-	rspamd_regexp_t *re;
-	gint pcre_flags;
-
-	map = re_map->map;
-
-	if (!(map->cfg->libs_ctx->crypto_ctx->cpu_config & CPUID_SSSE3)) {
-		msg_info_map ("disable hyperscan for map %s, ssse3 instructons are not supported by CPU",
-				map->name);
-		return;
-	}
-
-	if (hs_populate_platform (&plt) != HS_SUCCESS) {
-		msg_err_map ("cannot populate hyperscan platform");
-		return;
-	}
-
-	re_map->patterns = g_new (const gchar *, re_map->regexps->len);
-	re_map->flags = g_new (gint, re_map->regexps->len);
-	re_map->ids = g_new (gint, re_map->regexps->len);
-
-	for (i = 0; i < re_map->regexps->len; i ++) {
-		re = g_ptr_array_index (re_map->regexps, i);
-		re_map->patterns[i] = rspamd_regexp_get_pattern (re);
-		re_map->flags[i] = HS_FLAG_SINGLEMATCH;
-		pcre_flags = rspamd_regexp_get_pcre_flags (re);
-
-#ifndef WITH_PCRE2
-		if (pcre_flags & PCRE_FLAG(UTF8)) {
-			re_map->flags[i] |= HS_FLAG_UTF8;
-		}
-#else
-		if (pcre_flags & PCRE_FLAG(UTF)) {
-			re_map->flags[i] |= HS_FLAG_UTF8;
-		}
-#endif
-		if (pcre_flags & PCRE_FLAG(CASELESS)) {
-			re_map->flags[i] |= HS_FLAG_CASELESS;
-		}
-		if (pcre_flags & PCRE_FLAG(MULTILINE)) {
-			re_map->flags[i] |= HS_FLAG_MULTILINE;
-		}
-		if (pcre_flags & PCRE_FLAG(DOTALL)) {
-			re_map->flags[i] |= HS_FLAG_DOTALL;
-		}
-		if (rspamd_regexp_get_maxhits (re) == 1) {
-			re_map->flags[i] |= HS_FLAG_SINGLEMATCH;
-		}
-
-		re_map->ids[i] = i;
-	}
-
-	if (re_map->regexps->len > 0 && re_map->patterns) {
-		if (hs_compile_multi (re_map->patterns,
-				re_map->flags,
-				re_map->ids,
-				re_map->regexps->len,
-				HS_MODE_BLOCK,
-				&plt,
-				&re_map->hs_db,
-				&err) != HS_SUCCESS) {
-
-			msg_err_map ("cannot create tree of regexp when processing '%s': %s",
-					err->expression >= 0 ?
-							re_map->patterns[err->expression] :
-							"unknown regexp", err->message);
-			re_map->hs_db = NULL;
-			hs_free_compile_error (err);
-
-			return;
-		}
-
-		if (hs_alloc_scratch (re_map->hs_db, &re_map->hs_scratch) != HS_SUCCESS) {
-			msg_err_map ("cannot allocate scratch space for hyperscan");
-			hs_free_database (re_map->hs_db);
-			re_map->hs_db = NULL;
-		}
-	}
-	else {
-		msg_err_map ("regexp map is empty");
-	}
-#endif
-}
-
-gchar *
-rspamd_regexp_list_read_single (
-		gchar *chunk,
-		gint len,
-		struct map_cb_data *data,
-		gboolean final)
-{
-	struct rspamd_regexp_map *re_map;
-
-	if (data->cur_data == NULL) {
-		re_map = rspamd_regexp_map_create (data->map, 0);
-		data->cur_data = re_map;
-	}
-
-	return rspamd_parse_kv_list (
-			chunk,
-			len,
-			data,
-			rspamd_re_map_insert_helper,
-			hash_fill,
-			final);
-}
-
-gchar *
-rspamd_regexp_list_read_multiple (
-		gchar *chunk,
-		gint len,
-		struct map_cb_data *data,
-		gboolean final)
-{
-	struct rspamd_regexp_map *re_map;
-
-	if (data->cur_data == NULL) {
-		re_map = rspamd_regexp_map_create (data->map, RSPAMD_REGEXP_FLAG_MULTIPLE);
-		data->cur_data = re_map;
-	}
-
-	return rspamd_parse_kv_list (
-			chunk,
-			len,
-			data,
-			rspamd_re_map_insert_helper,
-			hash_fill,
-			final);
-}
-
-void
-rspamd_regexp_list_fin (struct map_cb_data *data)
-{
-	struct rspamd_regexp_map *re_map;
-	struct rspamd_map *map = data->map;
-
-	if (data->prev_data) {
-		rspamd_regexp_map_destroy (data->prev_data);
-	}
-	if (data->cur_data) {
-		re_map = data->cur_data;
-		rspamd_re_map_finalize (re_map);
-		msg_info_map ("read regexp list of %ud elements",
-				re_map->regexps->len);
-	}
-}
-
-#ifdef WITH_HYPERSCAN
-static int
-rspamd_match_hs_single_handler (unsigned int id, unsigned long long from,
-		unsigned long long to,
-		unsigned int flags, void *context)
-{
-	guint *i = context;
-	/* Always return non-zero as we need a single match here */
-
-	*i = id;
-
-	return 1;
-}
-#endif
-
-gpointer
-rspamd_match_regexp_map_single (struct rspamd_regexp_map *map,
-		const gchar *in, gsize len)
-{
-	guint i;
-	rspamd_regexp_t *re;
-	gint res = 0;
-	gpointer ret = NULL;
-	gboolean validated = FALSE;
-
-	g_assert (in != NULL);
-
-	if (map == NULL || len == 0) {
-		return NULL;
-	}
-
-	if (map->map_flags & RSPAMD_REGEXP_FLAG_UTF) {
-		if (g_utf8_validate (in, len, NULL)) {
-			validated = TRUE;
-		}
-	}
-	else {
-		validated = TRUE;
-	}
-
-#ifdef WITH_HYPERSCAN
-	if (map->hs_db && map->hs_scratch) {
-
-		if (validated) {
-
-			res = hs_scan (map->hs_db, in, len, 0, map->hs_scratch,
-					rspamd_match_hs_single_handler, (void *)&i);
-
-			if (res == HS_SCAN_TERMINATED) {
-				res = 1;
-				ret = g_ptr_array_index (map->values, i);
-			}
-
-			return ret;
-		}
-	}
-#endif
-
-	if (!res) {
-		/* PCRE version */
-		for (i = 0; i < map->regexps->len; i ++) {
-			re = g_ptr_array_index (map->regexps, i);
-
-			if (rspamd_regexp_search (re, in, len, NULL, NULL, !validated, NULL)) {
-				ret = g_ptr_array_index (map->values, i);
-				break;
-			}
-		}
-	}
-
-	return ret;
-}
-
-#ifdef WITH_HYPERSCAN
-struct rspamd_multiple_cbdata {
-	GPtrArray *ar;
-	struct rspamd_regexp_map *map;
-};
-
-static int
-rspamd_match_hs_multiple_handler (unsigned int id, unsigned long long from,
-		unsigned long long to,
-		unsigned int flags, void *context)
-{
-	struct rspamd_multiple_cbdata *cbd = context;
-
-	if (id < cbd->map->values->len) {
-		g_ptr_array_add (cbd->ar, g_ptr_array_index (cbd->map->values, id));
-	}
-
-	/* Always return zero as we need all matches here */
-	return 0;
-}
-#endif
-
-gpointer
-rspamd_match_regexp_map_all (struct rspamd_regexp_map *map,
-		const gchar *in, gsize len)
-{
-	guint i;
-	rspamd_regexp_t *re;
-	GPtrArray *ret;
-	gint res = 0;
-	gboolean validated = FALSE;
-
-	g_assert (in != NULL);
-
-	if (map == NULL || len == 0) {
-		return NULL;
-	}
-
-	if (map->map_flags & RSPAMD_REGEXP_FLAG_UTF) {
-		if (g_utf8_validate (in, len, NULL)) {
-			validated = TRUE;
-		}
-	}
-	else {
-		validated = TRUE;
-	}
-
-	ret = g_ptr_array_new ();
-
-#ifdef WITH_HYPERSCAN
-	if (map->hs_db && map->hs_scratch) {
-
-		if (validated) {
-			struct rspamd_multiple_cbdata cbd;
-
-			cbd.ar = ret;
-			cbd.map = map;
-
-			if (hs_scan (map->hs_db, in, len, 0, map->hs_scratch,
-					rspamd_match_hs_multiple_handler, &cbd) == HS_SUCCESS) {
-				res = 1;
-			}
-		}
-	}
-#endif
-
-	if (!res) {
-		/* PCRE version */
-		for (i = 0; i < map->regexps->len; i ++) {
-			re = g_ptr_array_index (map->regexps, i);
-
-			if (rspamd_regexp_search (re, in, len, NULL, NULL,
-					!validated, NULL)) {
-				g_ptr_array_add (ret, g_ptr_array_index (map->values, i));
-			}
-		}
-	}
-
-	if (ret->len > 0) {
-		return ret;
-	}
-
-	g_ptr_array_free (ret, TRUE);
-
 	return NULL;
 }
+
+void
+rspamd_map_traverse (struct rspamd_map *map, rspamd_map_traverse_cb cb,
+		gpointer cbdata, gboolean reset_hits)
+{
+	if (*map->user_data && map->traverse_function) {
+		map->traverse_function (*map->user_data, cb, cbdata, reset_hits);
+	}
+}
