@@ -188,6 +188,7 @@ struct lua_tcp_dtor {
 struct lua_tcp_cbdata {
 	lua_State *L;
 	struct rspamd_async_session *session;
+	struct rspamd_async_event *async_ev;
 	struct event_base *ev_base;
 	struct timeval tv;
 	rspamd_inet_addr_t *addr;
@@ -204,11 +205,13 @@ struct lua_tcp_cbdata {
 	ref_entry_t ref;
 };
 
-#define msg_debug_tcp(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        "lua_tcp", cbd->tag, \
+#define msg_debug_tcp(...)  rspamd_conditional_debug_fast (NULL, cbd->addr, \
+        rspamd_lua_tcp_log_id, "lua_tcp", cbd->tag, \
         G_STRFUNC, \
         __VA_ARGS__)
 
+INIT_LOG_MODULE(lua_tcp)
+
 static void lua_tcp_handler (int fd, short what, gpointer ud);
 static void lua_tcp_plan_handler_event (struct lua_tcp_cbdata *cbd,
 		gboolean can_read, gboolean can_write);
@@ -263,7 +266,7 @@ lua_tcp_shift_handler (struct lua_tcp_cbdata *cbd)
 		}
 	}
 
-	g_slice_free1 (sizeof (*hdl), hdl);
+	g_free (hdl);
 
 	return TRUE;
 }
@@ -294,11 +297,11 @@ lua_tcp_fin (gpointer arg)
 
 	LL_FOREACH_SAFE (cbd->dtors, dtor, dttmp) {
 		dtor->dtor (dtor->data);
-		g_slice_free1 (sizeof (*dtor), dtor);
+		g_free (dtor);
 	}
 
 	g_byte_array_unref (cbd->in);
-	g_slice_free1 (sizeof (struct lua_tcp_cbdata), cbd);
+	g_free (cbd);
 }
 
 static struct lua_tcp_cbdata *
@@ -312,7 +315,7 @@ lua_check_tcp (lua_State *L, gint pos)
 static void
 lua_tcp_maybe_free (struct lua_tcp_cbdata *cbd)
 {
-	if (cbd->session) {
+	if (cbd->async_ev) {
 		rspamd_session_watcher_pop (cbd->session, cbd->w);
 		rspamd_session_remove_event (cbd->session, lua_tcp_fin, cbd);
 	}
@@ -321,6 +324,12 @@ lua_tcp_maybe_free (struct lua_tcp_cbdata *cbd)
 	}
 }
 
+#ifdef __GNUC__
+static void
+lua_tcp_push_error (struct lua_tcp_cbdata *cbd, gboolean is_fatal,
+		const char *err, ...) __attribute__ ((format(printf, 3, 4)));
+#endif
+
 static void
 lua_tcp_push_error (struct lua_tcp_cbdata *cbd, gboolean is_fatal,
 		const char *err, ...)
@@ -328,20 +337,14 @@ lua_tcp_push_error (struct lua_tcp_cbdata *cbd, gboolean is_fatal,
 	va_list ap, ap_copy;
 	struct lua_tcp_cbdata **pcbd;
 	struct lua_tcp_handler *hdl;
-	gint cbref;
+	gint cbref, top;
 
 	va_start (ap, err);
 
 	for (;;) {
-		if (is_fatal) {
-			hdl = g_queue_pop_head (cbd->handlers);
-		}
-		else {
-			hdl = g_queue_peek_head (cbd->handlers);
-		}
+		hdl = g_queue_peek_head (cbd->handlers);
 
 		if (hdl == NULL) {
-			va_end (ap_copy);
 			break;
 		}
 
@@ -353,6 +356,7 @@ lua_tcp_push_error (struct lua_tcp_cbdata *cbd, gboolean is_fatal,
 		}
 
 		if (cbref != -1) {
+			top = lua_gettop (cbd->L);
 			lua_rawgeti (cbd->L, LUA_REGISTRYINDEX, cbref);
 
 			/* Error message */
@@ -370,9 +374,10 @@ lua_tcp_push_error (struct lua_tcp_cbdata *cbd, gboolean is_fatal,
 
 			if (lua_pcall (cbd->L, 3, 0, 0) != 0) {
 				msg_info ("callback call failed: %s", lua_tostring (cbd->L, -1));
-				lua_pop (cbd->L, 1);
 			}
 
+			lua_settop (cbd->L, top);
+
 			REF_RELEASE (cbd);
 		}
 
@@ -380,6 +385,9 @@ lua_tcp_push_error (struct lua_tcp_cbdata *cbd, gboolean is_fatal,
 			/* Stop on the first callback found */
 			break;
 		}
+		else {
+			lua_tcp_shift_handler (cbd);
+		}
 	}
 
 	va_end (ap);
@@ -391,7 +399,7 @@ lua_tcp_push_data (struct lua_tcp_cbdata *cbd, const guint8 *str, gsize len)
 	struct rspamd_lua_text *t;
 	struct lua_tcp_cbdata **pcbd;
 	struct lua_tcp_handler *hdl;
-	gint cbref, arg_cnt;
+	gint cbref, arg_cnt, top;
 
 	hdl = g_queue_peek_head (cbd->handlers);
 
@@ -405,6 +413,7 @@ lua_tcp_push_data (struct lua_tcp_cbdata *cbd, const guint8 *str, gsize len)
 	}
 
 	if (cbref != -1) {
+		top = lua_gettop (cbd->L);
 		lua_rawgeti (cbd->L, LUA_REGISTRYINDEX, cbref);
 		/* Error */
 		lua_pushnil (cbd->L);
@@ -430,9 +439,9 @@ lua_tcp_push_data (struct lua_tcp_cbdata *cbd, const guint8 *str, gsize len)
 
 		if (lua_pcall (cbd->L, arg_cnt, 0, 0) != 0) {
 			msg_info ("callback call failed: %s", lua_tostring (cbd->L, -1));
-			lua_pop (cbd->L, 1);
 		}
 
+		lua_settop (cbd->L, top);
 		REF_RELEASE (cbd);
 	}
 }
@@ -686,7 +695,9 @@ lua_tcp_handler (int fd, short what, gpointer ud)
 
 				if (cbd->connect_cb != -1) {
 					struct lua_tcp_cbdata **pcbd;
+					gint top;
 
+					top = lua_gettop (cbd->L);
 					lua_rawgeti (cbd->L, LUA_REGISTRYINDEX, cbd->connect_cb);
 					pcbd = lua_newuserdata (cbd->L, sizeof (*pcbd));
 					*pcbd = cbd;
@@ -695,9 +706,10 @@ lua_tcp_handler (int fd, short what, gpointer ud)
 
 					if (lua_pcall (cbd->L, 1, 0, 0) != 0) {
 						msg_info ("callback call failed: %s", lua_tostring (cbd->L, -1));
-						lua_pop (cbd->L, 1);
 					}
 
+					lua_settop (cbd->L, top);
+
 					REF_RELEASE (cbd);
 				}
 			}
@@ -792,6 +804,19 @@ lua_tcp_plan_handler_event (struct lua_tcp_cbdata *cbd, gboolean can_read,
 	}
 }
 
+static void
+lua_tcp_register_event (struct lua_tcp_cbdata *cbd)
+{
+	if (cbd->session) {
+		cbd->async_ev = rspamd_session_add_event (cbd->session,
+				(event_finalizer_t) lua_tcp_fin,
+				cbd,
+				g_quark_from_static_string ("lua tcp"));
+		cbd->w = rspamd_session_get_watcher (cbd->session);
+		rspamd_session_watcher_push (cbd->session);
+	}
+}
+
 static gboolean
 lua_tcp_make_connection (struct lua_tcp_cbdata *cbd)
 {
@@ -801,7 +826,18 @@ lua_tcp_make_connection (struct lua_tcp_cbdata *cbd)
 	fd = rspamd_inet_address_connect (cbd->addr, SOCK_STREAM, TRUE);
 
 	if (fd == -1) {
-		msg_info ("cannot connect to %s", rspamd_inet_address_to_string (cbd->addr));
+		if (cbd->session) {
+			rspamd_mempool_t *pool = rspamd_session_mempool (cbd->session);
+			msg_info_pool ("cannot connect to %s: %s",
+					rspamd_inet_address_to_string (cbd->addr),
+					strerror (errno));
+		}
+		else {
+			msg_info ("cannot connect to %s: %s",
+					rspamd_inet_address_to_string (cbd->addr),
+					strerror (errno));
+		}
+
 		return FALSE;
 	}
 
@@ -862,7 +898,7 @@ lua_tcp_arg_toiovec (lua_State *L, gint pos, struct lua_tcp_cbdata *cbd,
 			if (t->flags & RSPAMD_TEXT_FLAG_OWN) {
 				/* Steal ownership */
 				t->flags = 0;
-				dtor = g_slice_alloc0 (sizeof (*dtor));
+				dtor = g_malloc0 (sizeof (*dtor));
 				dtor->dtor = g_free;
 				dtor->data = (void *)t->start;
 				LL_PREPEND (cbd->dtors, dtor);
@@ -876,7 +912,7 @@ lua_tcp_arg_toiovec (lua_State *L, gint pos, struct lua_tcp_cbdata *cbd,
 	else if (lua_type (L, pos) == LUA_TSTRING) {
 		str = luaL_checklstring (L, pos, &len);
 		vec->iov_base = g_malloc (len);
-		dtor = g_slice_alloc0 (sizeof (*dtor));
+		dtor = g_malloc0 (sizeof (*dtor));
 		dtor->dtor = g_free;
 		dtor->data = vec->iov_base;
 		LL_PREPEND (cbd->dtors, dtor);
@@ -926,7 +962,7 @@ lua_tcp_request (lua_State *L)
 	struct event_base *ev_base;
 	struct lua_tcp_cbdata *cbd;
 	struct rspamd_dns_resolver *resolver;
-	struct rspamd_async_session *session;
+	struct rspamd_async_session *session = NULL;
 	struct rspamd_task *task = NULL;
 	struct rspamd_config *cfg;
 	struct iovec *iov = NULL;
@@ -963,7 +999,7 @@ lua_tcp_request (lua_State *L)
 		}
 		cbref = luaL_ref (L, LUA_REGISTRYINDEX);
 
-		cbd = g_slice_alloc0 (sizeof (*cbd));
+		cbd = g_malloc0 (sizeof (*cbd));
 
 		lua_pushstring (L, "task");
 		lua_gettable (L, -2);
@@ -1083,7 +1119,7 @@ lua_tcp_request (lua_State *L)
 				msg_err ("tcp request has bad data argument");
 				lua_pushboolean (L, FALSE);
 				g_free (iov);
-				g_slice_free1 (sizeof (*cbd), cbd);
+				g_free (cbd);
 
 				return 1;
 			}
@@ -1108,7 +1144,7 @@ lua_tcp_request (lua_State *L)
 					msg_err ("tcp request has bad data argument at pos %d", niov);
 					lua_pushboolean (L, FALSE);
 					g_free (iov);
-					g_slice_free1 (sizeof (*cbd), cbd);
+					g_free (cbd);
 
 					return 1;
 				}
@@ -1137,7 +1173,7 @@ lua_tcp_request (lua_State *L)
 	if (total_out > 0) {
 		struct lua_tcp_handler *wh;
 
-		wh = g_slice_alloc0 (sizeof (*wh));
+		wh = g_malloc0 (sizeof (*wh));
 		wh->type = LUA_WANT_WRITE;
 		wh->h.w.iov = iov;
 		wh->h.w.iovlen = niov;
@@ -1182,7 +1218,7 @@ lua_tcp_request (lua_State *L)
 	if (do_read) {
 		struct lua_tcp_handler *rh;
 
-		rh = g_slice_alloc0 (sizeof (*rh));
+		rh = g_malloc0 (sizeof (*rh));
 		rh->type = LUA_WANT_READ;
 		rh->h.r.cbref = cbref;
 		rh->h.r.stop_pattern = stop_pattern;
@@ -1195,12 +1231,6 @@ lua_tcp_request (lua_State *L)
 
 	if (session) {
 		cbd->session = session;
-		rspamd_session_add_event (session,
-				(event_finalizer_t)lua_tcp_fin,
-				cbd,
-				g_quark_from_static_string ("lua tcp"));
-		cbd->w = rspamd_session_get_watcher (session);
-		rspamd_session_watcher_push (session);
 	}
 
 	if (rspamd_parse_inet_address (&cbd->addr, host, 0)) {
@@ -1212,6 +1242,9 @@ lua_tcp_request (lua_State *L)
 
 			return 1;
 		}
+		else {
+			lua_tcp_register_event (cbd);
+		}
 	}
 	else {
 		if (task == NULL) {
@@ -1219,6 +1252,11 @@ lua_tcp_request (lua_State *L)
 					RDNS_REQUEST_A, host)) {
 				lua_tcp_push_error (cbd, TRUE, "cannot resolve host: %s", host);
 				REF_RELEASE (cbd);
+				lua_pushboolean (L, FALSE);
+				return 1;
+			}
+			else {
+				lua_tcp_register_event (cbd);
 			}
 		}
 		else {
@@ -1226,6 +1264,11 @@ lua_tcp_request (lua_State *L)
 					RDNS_REQUEST_A, host)) {
 				lua_tcp_push_error (cbd, TRUE, "cannot resolve host: %s", host);
 				REF_RELEASE (cbd);
+				lua_pushboolean (L, FALSE);
+				return 1;
+			}
+			else {
+				lua_tcp_register_event (cbd);
 			}
 		}
 	}
@@ -1293,7 +1336,7 @@ lua_tcp_add_read (lua_State *L)
 		}
 	}
 
-	rh = g_slice_alloc0 (sizeof (*rh));
+	rh = g_malloc0 (sizeof (*rh));
 	rh->type = LUA_WANT_READ;
 	rh->h.r.cbref = cbref;
 	rh->h.r.stop_pattern = stop_pattern;
@@ -1358,7 +1401,7 @@ lua_tcp_add_write (lua_State *L)
 				msg_err ("tcp request has bad data argument at pos %d", niov);
 				lua_pushboolean (L, FALSE);
 				g_free (iov);
-				g_slice_free1 (sizeof (*cbd), cbd);
+				g_free (cbd);
 
 				return 1;
 			}
@@ -1372,7 +1415,7 @@ lua_tcp_add_write (lua_State *L)
 		lua_pop (L, 1);
 	}
 
-	wh = g_slice_alloc0 (sizeof (*wh));
+	wh = g_malloc0 (sizeof (*wh));
 	wh->type = LUA_WANT_WRITE;
 	wh->h.w.iov = iov;
 	wh->h.w.iovlen = niov;
