@@ -35,9 +35,11 @@
 #include "config.h"
 #include "libmime/message.h"
 #include "libutil/map.h"
+#include "libutil/map_helpers.h"
 #include "rspamd.h"
-#include "surbl.h"
 #include "utlist.h"
+#include "multipattern.h"
+#include "monitored.h"
 #include "libserver/html.h"
 #include "libutil/http_private.h"
 #include "unix-std.h"
@@ -55,16 +57,84 @@
         "surbl", task->task_pool->tag.uid, \
         G_STRFUNC, \
         __VA_ARGS__)
-#define msg_debug_surbl(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        "surbl", task->task_pool->tag.uid, \
+#define msg_debug_surbl(...)  rspamd_conditional_debug_fast (NULL, task->from_addr, \
+        rspamd_surbl_log_id, "surbl", task->task_pool->tag.uid, \
         G_STRFUNC, \
         __VA_ARGS__)
 
+INIT_LOG_MODULE(surbl)
+
+#define DEFAULT_SURBL_WEIGHT 10
+#define DEFAULT_REDIRECTOR_READ_TIMEOUT 5.0
+#define DEFAULT_SURBL_SYMBOL "SURBL_DNS"
+#define SURBL_OPTION_NOIP (1 << 0)
+#define SURBL_OPTION_RESOLVEIP (1 << 1)
+#define SURBL_OPTION_CHECKIMAGES (1 << 2)
+#define MAX_LEVELS 10
+
+struct surbl_ctx {
+	struct module_ctx ctx;
+	guint16 weight;
+	gdouble read_timeout;
+	gboolean use_tags;
+	GList *suffixes;
+	gchar *metric;
+	const gchar *redirector_symbol;
+	GHashTable **exceptions;
+	struct rspamd_hash_map_helper *whitelist;
+	void *redirector_map_data;
+	GHashTable *redirector_tlds;
+	guint use_redirector;
+	guint max_redirected_urls;
+	gint redirector_cbid;
+	struct upstream_list *redirectors;
+	rspamd_mempool_t *surbl_pool;
+};
+
+struct suffix_item {
+	guint64 magic;
+	const gchar *monitored_domain;
+	const gchar *suffix;
+	const gchar *symbol;
+	guint32 options;
+	GArray *bits;
+	GHashTable *ips;
+	struct rspamd_monitored *m;
+	gint callback_id;
+	gint url_process_cbref;
+};
+
+struct dns_param {
+	struct rspamd_url *url;
+	struct rspamd_task *task;
+	gchar *host_resolve;
+	struct suffix_item *suffix;
+	struct rspamd_async_watcher *w;
+};
+
+struct redirector_param {
+	struct rspamd_url *url;
+	struct rspamd_task *task;
+	struct upstream *redirector;
+	struct rspamd_http_connection *conn;
+	GHashTable *tree;
+	struct suffix_item *suffix;
+	struct rspamd_async_watcher *w;
+	gint sock;
+	guint redirector_requests;
+};
+
+struct surbl_bit_item {
+	guint32 bit;
+	gchar *symbol;
+};
+
 #define SURBL_REDIRECTOR_CALLBACK "SURBL_REDIRECTOR_CALLBACK"
 
 static struct surbl_ctx *surbl_module_ctx = NULL;
 static const guint64 rspamd_surbl_cb_magic = 0xe09b8536f80de0d1ULL;
 static const gchar *rspamd_surbl_default_monitored = "facebook.com";
+static const guint default_max_redirected_urls = 10;
 
 static void surbl_test_url (struct rspamd_task *task, void *user_data);
 static void surbl_test_redirector (struct rspamd_task *task, void *user_data);
@@ -110,7 +180,7 @@ exceptions_free_value (gpointer v)
 	rspamd_ftok_t *val = v;
 
 	g_free ((gpointer)val->begin);
-	g_slice_free1 (sizeof (*val), val);
+	g_free (val);
 }
 
 static void
@@ -134,7 +204,7 @@ exception_insert (gpointer st, gconstpointer key, gconstpointer value)
 		return;
 	}
 
-	val = g_slice_alloc (sizeof (rspamd_ftok_t));
+	val = g_malloc (sizeof (rspamd_ftok_t));
 	val->begin = g_strdup (key);
 	val->len = strlen (key);
 
@@ -214,7 +284,7 @@ redirector_insert (gpointer st, gconstpointer key, gconstpointer value)
 	}
 
 	pat = rspamd_fstring_new_init (begin, p - begin);
-	tok = g_slice_alloc (sizeof (*tok));
+	tok = g_malloc0 (sizeof (*tok));
 	tok->begin = pat->str;
 	tok->len = pat->len;
 
@@ -301,10 +371,9 @@ surbl_module_init (struct rspamd_config *cfg, struct module_ctx **ctx)
 	surbl_module_ctx->surbl_pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), NULL);
 
 	surbl_module_ctx->redirectors = NULL;
-	surbl_module_ctx->whitelist = g_hash_table_new (rspamd_strcase_hash,
-			rspamd_strcase_equal);
+	surbl_module_ctx->whitelist = NULL;
 	rspamd_mempool_add_destructor (surbl_module_ctx->surbl_pool,
-			(rspamd_mempool_destruct_t) g_hash_table_destroy,
+			(rspamd_mempool_destruct_t) rspamd_map_helper_destroy_hash,
 			surbl_module_ctx->whitelist);
 	surbl_module_ctx->exceptions = rspamd_mempool_alloc0 (
 			surbl_module_ctx->surbl_pool, MAX_LEVELS * sizeof (GHashTable *));
@@ -522,6 +591,8 @@ surbl_module_parse_rule (const ucl_object_t* value, struct rspamd_config* cfg)
 	ucl_object_t *ropts;
 
 	LL_FOREACH(value, cur_rule) {
+		monitored_domain = NULL;
+
 		cur = ucl_object_lookup (cur_rule, "enabled");
 		if (cur != NULL && cur->type == UCL_BOOLEAN) {
 			if (!ucl_object_toboolean (cur)) {
@@ -728,7 +799,7 @@ surbl_module_parse_rule (const ucl_object_t* value, struct rspamd_config* cfg)
 
 			if (luaL_loadstring (L, tb->str) != 0) {
 				/* Reset stack */
-				lua_settop (L, 0);
+				lua_settop (L, err_idx - 1);
 				lua_pushcfunction (L, &rspamd_lua_traceback);
 				err_idx = lua_gettop (L);
 				/* Try with no return */
@@ -791,13 +862,17 @@ surbl_module_config (struct rspamd_config *cfg)
 	gint nrules = 0;
 	lua_State *L;
 
+	if (!rspamd_config_is_module_enabled (cfg, "surbl")) {
+		return TRUE;
+	}
+
 	/* Register global methods */
 	L = cfg->lua_state;
 	lua_getglobal (L, "rspamd_plugins");
 
 	if (lua_type (L, -1) == LUA_TTABLE) {
 		lua_pushstring (L, "surbl");
-		lua_createtable (L, 0, 2);
+		lua_createtable (L, 0, 3);
 		/* Set methods */
 		lua_pushstring (L, "register_redirect");
 		lua_pushcfunction (L, surbl_register_redirect_handler);
@@ -808,16 +883,12 @@ surbl_module_config (struct rspamd_config *cfg)
 		lua_pushstring (L, "is_redirector");
 		lua_pushcfunction (L, surbl_is_redirector_handler);
 		lua_settable (L, -3);
-		/* Finish fuzzy_check key */
+		/* Finish surbl key */
 		lua_settable (L, -3);
 	}
 
 	lua_pop (L, 1); /* Remove global function */
 
-	if (!rspamd_config_is_module_enabled (cfg, "surbl")) {
-		return TRUE;
-	}
-
 	(void)rspamd_symbols_cache_add_symbol (cfg->cache, SURBL_REDIRECTOR_CALLBACK,
 			0, surbl_test_redirector, NULL,
 			SYMBOL_TYPE_CALLBACK, -1);
@@ -896,7 +967,7 @@ surbl_module_config (struct rspamd_config *cfg)
 	if ((value =
 			rspamd_config_get_module_opt (cfg, "surbl", "whitelist")) != NULL) {
 		rspamd_map_add_from_ucl (cfg, value,
-				"SURBL whitelist", rspamd_hosts_read, rspamd_hosts_fin,
+				"SURBL whitelist", rspamd_kv_list_read, rspamd_kv_list_fin,
 				(void **)&surbl_module_ctx->whitelist);
 	}
 
@@ -950,6 +1021,13 @@ surbl_module_config (struct rspamd_config *cfg)
 		cur_opt = g_list_next (cur_opt);
 	}
 
+	surbl_module_ctx->max_redirected_urls = default_max_redirected_urls;
+
+	if ((value =
+			rspamd_config_get_module_opt (cfg, "surbl", "max_redirected_urls")) != NULL) {
+		surbl_module_ctx->max_redirected_urls = ucl_obj_toint (value);
+	}
+
 	msg_info_config ("init internal surbls module, %d uribl rules loaded",
 			nrules);
 
@@ -967,15 +1045,14 @@ surbl_module_reconfig (struct rspamd_config *cfg)
 	surbl_module_ctx->surbl_pool = rspamd_mempool_new (rspamd_mempool_suggest_size (), NULL);
 
 	surbl_module_ctx->redirectors = NULL;
-	surbl_module_ctx->whitelist = g_hash_table_new (rspamd_strcase_hash,
-			rspamd_strcase_equal);
+	surbl_module_ctx->whitelist = NULL;
 	/* Zero exceptions hashes */
 	surbl_module_ctx->exceptions = rspamd_mempool_alloc0 (
 		surbl_module_ctx->surbl_pool,
 		MAX_LEVELS * sizeof (GHashTable *));
 	/* Register destructors */
 	rspamd_mempool_add_destructor (surbl_module_ctx->surbl_pool,
-		(rspamd_mempool_destruct_t) g_hash_table_destroy,
+		(rspamd_mempool_destruct_t) rspamd_map_helper_destroy_hash,
 		surbl_module_ctx->whitelist);
 	rspamd_mempool_add_destructor (surbl_module_ctx->surbl_pool,
 		(rspamd_mempool_destruct_t) g_hash_table_destroy,
@@ -1124,7 +1201,7 @@ format_surbl_request (rspamd_mempool_t * pool,
 	url->surbllen = r;
 
 	if (!forced &&
-			g_hash_table_lookup (surbl_module_ctx->whitelist, result) != NULL) {
+			rspamd_match_hash_map (surbl_module_ctx->whitelist, result) != NULL) {
 		msg_debug_pool ("url %s is whitelisted", result);
 		g_set_error (err, SURBL_ERROR,
 				WHITELIST_ERROR,
@@ -1637,6 +1714,16 @@ surbl_tree_redirector_callback (gpointer key, gpointer value, void *data)
 						found_tld);
 			}
 
+			if (param->redirector_requests >= surbl_module_ctx->max_redirected_urls) {
+				msg_info_surbl ("cannot register redirector request for url domain: "
+						"%s, max_redirected_urls is reached: %d",
+						found_tld, surbl_module_ctx->max_redirected_urls);
+
+				return;
+			}
+
+			param->redirector_requests ++;
+
 			if (surbl_module_ctx->redirector_cbid != -1) {
 				nparam = rspamd_mempool_alloc (task->task_pool,
 						sizeof (*nparam));
@@ -1767,6 +1854,7 @@ surbl_test_redirector (struct rspamd_task *task, void *user_data)
 	param = rspamd_mempool_alloc0 (task->task_pool, sizeof (*param));
 	param->task = task;
 	param->suffix = NULL;
+	param->redirector_requests = 0;
 	g_hash_table_foreach (task->urls, surbl_tree_redirector_callback, param);
 
 	/* We also need to check and process img URLs */
