@@ -52,11 +52,23 @@
 #define CR '\r'
 #define LF '\n'
 
+enum rspamd_worker_flags {
+	RSPAMD_WORKER_HAS_SOCKET = (1 << 0),
+	RSPAMD_WORKER_UNIQUE = (1 << 1),
+	RSPAMD_WORKER_THREADED = (1 << 2),
+	RSPAMD_WORKER_KILLABLE = (1 << 3),
+	RSPAMD_WORKER_ALWAYS_START = (1 << 4),
+	RSPAMD_WORKER_SCANNER = (1 << 5),
+	RSPAMD_WORKER_CONTROLLER = (1 << 6),
+};
+
+
 /**
  * Worker process structure
  */
 struct rspamd_worker {
 	pid_t pid;                      /**< pid of worker									*/
+	pid_t ppid;                     /**< pid of parent									*/
 	guint index;                    /**< index number									*/
 	guint nconns;                   /**< current connections count						*/
 	gboolean wanna_die;             /**< worker is terminating							*/
@@ -67,6 +79,7 @@ struct rspamd_worker {
 	GList *accept_events;           /**< socket events									*/
 	struct rspamd_worker_conf *cf;  /**< worker config data								*/
 	gpointer ctx;                   /**< worker's specific data							*/
+	enum rspamd_worker_flags flags; /**< worker's flags									*/
 	gint control_pipe[2];           /**< control pipe. [0] is used by main process,
 	                                                   [1] is used by a worker			*/
 	gint srv_pipe[2];               /**< used by workers to request something from the
@@ -88,9 +101,11 @@ struct rspamd_abstract_worker_ctx {
 };
 
 struct rspamd_worker_signal_handler;
+typedef gboolean (*rspamd_worker_signal_handler) (
+		struct rspamd_worker_signal_handler *, void *ud);
 
 struct rspamd_worker_signal_cb {
-	void (*handler) (struct rspamd_worker_signal_handler *, void *ud);
+	rspamd_worker_signal_handler handler;
 	void *handler_data;
 	struct rspamd_worker_signal_cb *next, *prev;
 };
@@ -177,14 +192,6 @@ typedef struct module_s {
 	const gchar *rspamd_features;
 } module_t;
 
-enum rspamd_worker_flags {
-	RSPAMD_WORKER_HAS_SOCKET = (1 << 0),
-	RSPAMD_WORKER_UNIQUE = (1 << 1),
-	RSPAMD_WORKER_THREADED = (1 << 2),
-	RSPAMD_WORKER_KILLABLE = (1 << 3),
-	RSPAMD_WORKER_ALWAYS_START = (1 << 4),
-};
-
 enum rspamd_worker_socket_type {
 	RSPAMD_WORKER_SOCKET_NONE = 0,
 	RSPAMD_WORKER_SOCKET_TCP = (1 << 0),
@@ -307,9 +314,11 @@ struct zstd_dictionary {
 	guint id;
 };
 
+struct rspamd_radix_map_helper;
+
 struct rspamd_external_libs_ctx {
 	magic_t libmagic;
-	radix_compressed_t **local_addrs;
+	struct rspamd_radix_map_helper **local_addrs;
 	struct rspamd_cryptobox_library_ctx *crypto_ctx;
 	struct ottery_config *ottery_cfg;
 	SSL_CTX *ssl_ctx;
