@@ -20,6 +20,7 @@
 #include "libutil/http.h"
 #include "libutil/http_private.h"
 #include "libserver/protocol.h"
+#include "libserver/protocol_internal.h"
 #include "libserver/cfg_file.h"
 #include "libserver/url.h"
 #include "libserver/dns.h"
@@ -33,8 +34,14 @@
 #include "ottery.h"
 #include "unix-std.h"
 #include "libserver/milter.h"
+#include "libserver/milter_internal.h"
+#include "libmime/lang_detection.h"
 #include "contrib/zstd/zstd.h"
 
+#ifdef HAVE_NETINET_TCP_H
+#include <netinet/tcp.h> /* for TCP_NODELAY */
+#endif
+
 /* Rotate keys each minute by default */
 #define DEFAULT_ROTATION_TIME 60.0
 #define DEFAULT_RETRIES 5
@@ -51,11 +58,14 @@
         session->pool->tag.tagname, session->pool->tag.uid, \
         G_STRFUNC, \
         __VA_ARGS__)
-#define msg_debug_session(...)  rspamd_default_log_function (G_LOG_LEVEL_DEBUG, \
-        session->pool->tag.tagname, session->pool->tag.uid, \
+
+#define msg_debug_session(...)  rspamd_conditional_debug_fast (NULL, session->client_addr, \
+        rspamd_proxy_log_id, "proxy", session->pool->tag.uid, \
         G_STRFUNC, \
         __VA_ARGS__)
 
+INIT_LOG_MODULE(proxy)
+
 gpointer init_rspamd_proxy (struct rspamd_config *cfg);
 void start_rspamd_proxy (struct rspamd_worker *worker);
 
@@ -63,13 +73,14 @@ worker_t rspamd_proxy_worker = {
 	"rspamd_proxy",               /* Name */
 	init_rspamd_proxy,            /* Init function */
 	start_rspamd_proxy,           /* Start function */
-	RSPAMD_WORKER_HAS_SOCKET | RSPAMD_WORKER_KILLABLE,
+	RSPAMD_WORKER_HAS_SOCKET | RSPAMD_WORKER_KILLABLE | RSPAMD_WORKER_SCANNER,
 	RSPAMD_WORKER_SOCKET_TCP,    /* TCP socket */
 	RSPAMD_WORKER_VER
 };
 
 struct rspamd_http_upstream {
 	gchar *name;
+	gchar *settings_id;
 	struct upstream_list *u;
 	struct rspamd_cryptobox_pubkey *key;
 	gdouble timeout;
@@ -133,10 +144,19 @@ struct rspamd_proxy_ctx {
 	gboolean milter;
 	/* Discard messages instead of rejecting them */
 	gboolean discard_on_reject;
+	/* Quarantine messages instead of rejecting them */
+	gboolean quarantine_on_reject;
 	/* Milter spam header */
 	gchar *spam_header;
+	/* CA name that can be used for client certificates */
+	gchar *client_ca_name;
+	/* Milter rejection message */
+	gchar *reject_message;
 	/* Sessions cache */
 	void *sessions_cache;
+	struct rspamd_milter_context milter_ctx;
+	/* Language detector */
+	struct rspamd_lang_detector *lang_det;
 };
 
 enum rspamd_backend_flags {
@@ -177,6 +197,7 @@ struct rspamd_proxy_session {
 	rspamd_inet_addr_t *client_addr;
 	struct rspamd_http_connection *client_conn;
 	struct rspamd_milter_session *client_milter_conn;
+	struct rspamd_http_upstream *backend;
 	gpointer map;
 	gchar *fname;
 	gpointer shmem_ref;
@@ -427,6 +448,11 @@ rspamd_proxy_parse_upstream (rspamd_mempool_t *pool,
 		ucl_object_todouble_safe (elt, &up->timeout);
 	}
 
+	elt = ucl_object_lookup_any (obj, "settings", "settings_id", NULL);
+	if (elt && ucl_object_type (elt) == UCL_STRING) {
+		up->settings_id = rspamd_mempool_strdup (pool, ucl_object_tostring (elt));
+	}
+
 	/*
 	 * Accept lua function here in form
 	 * fun :: String -> UCL
@@ -724,6 +750,7 @@ init_rspamd_proxy (struct rspamd_config *cfg)
 	rspamd_mempool_add_destructor (cfg->cfg_pool,
 			(rspamd_mempool_destruct_t)rspamd_array_free_hard, ctx->cmp_refs);
 	ctx->max_retries = DEFAULT_RETRIES;
+	ctx->spam_header = RSPAMD_MILTER_SPAM_HEADER;
 
 	rspamd_rcl_register_worker_option (cfg,
 			type,
@@ -799,6 +826,14 @@ init_rspamd_proxy (struct rspamd_config *cfg)
 			G_STRUCT_OFFSET (struct rspamd_proxy_ctx, discard_on_reject),
 			0,
 			"Tell MTA to discard rejected messages silently");
+	rspamd_rcl_register_worker_option (cfg,
+			type,
+			"quarantine_on_reject",
+			rspamd_rcl_parse_struct_boolean,
+			ctx,
+			G_STRUCT_OFFSET (struct rspamd_proxy_ctx, quarantine_on_reject),
+			0,
+			"Tell MTA to quarantine rejected messages");
 	rspamd_rcl_register_worker_option (cfg,
 			type,
 			"spam_header",
@@ -806,7 +841,23 @@ init_rspamd_proxy (struct rspamd_config *cfg)
 			ctx,
 			G_STRUCT_OFFSET (struct rspamd_proxy_ctx, spam_header),
 			0,
-			"Use the specific spam header instead of X-Spam");
+			"Use the specific spam header (default: X-Spam)");
+	rspamd_rcl_register_worker_option (cfg,
+			type,
+			"client_ca_name",
+			rspamd_rcl_parse_struct_string,
+			ctx,
+			G_STRUCT_OFFSET (struct rspamd_proxy_ctx, client_ca_name),
+			0,
+			"Allow certificates issued by this CA to be treated as client certificates");
+	rspamd_rcl_register_worker_option (cfg,
+			type,
+			"reject_message",
+			rspamd_rcl_parse_struct_string,
+			ctx,
+			G_STRUCT_OFFSET (struct rspamd_proxy_ctx, reject_message),
+			0,
+			"Use custom rejection message");
 
 	return ctx;
 }
@@ -1007,7 +1058,7 @@ proxy_session_dtor (struct rspamd_proxy_session *session)
 		rspamd_mempool_delete (session->pool);
 	}
 
-	g_slice_free1 (sizeof (*session), session);
+	g_free (session);
 }
 
 static void
@@ -1120,7 +1171,7 @@ proxy_session_refresh (struct rspamd_proxy_session *session)
 {
 	struct rspamd_proxy_session *nsession;
 
-	nsession = g_slice_alloc0 (sizeof (*nsession));
+	nsession = g_malloc0 (sizeof (*nsession));
 	nsession->client_milter_conn = session->client_milter_conn;
 	session->client_milter_conn = NULL;
 	rspamd_milter_update_userdata (nsession->client_milter_conn,
@@ -1165,9 +1216,12 @@ proxy_check_file (struct rspamd_http_message *msg,
 				TRUE);
 
 		if (session->map == NULL) {
-			msg_err_session ("cannot map %s: %s", file_str, strerror (errno));
+			if (session->map_len != 0) {
+				msg_err_session ("cannot map %s: %s", file_str,
+						strerror (errno));
 
-			return FALSE;
+				return FALSE;
+			}
 		}
 		/* Remove header after processing */
 		rspamd_http_message_remove_header (msg, "File");
@@ -1175,7 +1229,8 @@ proxy_check_file (struct rspamd_http_message *msg,
 	}
 	else {
 		/* Need to parse query URL */
-		if (http_parser_parse_url (msg->url->str, msg->url->len, 0, &u) != 0) {
+		if (http_parser_parse_url (RSPAMD_FSTRING_DATA (msg->url),
+				RSPAMD_FSTRING_LEN (msg->url), 0, &u) != 0) {
 			msg_err_session ("bad request url: %V", msg->url);
 
 			return FALSE;
@@ -1194,14 +1249,17 @@ proxy_check_file (struct rspamd_http_message *msg,
 						&session->map_len, TRUE);
 
 				if (session->map == NULL) {
-					msg_err_session ("cannot map %s: %s", file_str, strerror (errno));
-					g_hash_table_unref (query_args);
+					if (session->map_len != 0) {
+						msg_err_session ("cannot map %s: %s", file_str,
+								strerror (errno));
+						g_hash_table_unref (query_args);
 
-					return FALSE;
+						return FALSE;
+					}
 				}
 
 				/* We need to create a new URL with file attribute removed */
-				new_url = rspamd_fstring_new_init (msg->url->str,
+				new_url = rspamd_fstring_new_init (RSPAMD_FSTRING_DATA (msg->url),
 						u.field_data[UF_QUERY].off);
 				new_url = rspamd_fstring_append (new_url, "?", 1);
 
@@ -1292,6 +1350,7 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 	guint i;
 	struct rspamd_proxy_backend_connection *bk_conn;
 	struct rspamd_http_message *msg;
+	GError *err = NULL;
 
 	coin = rspamd_random_double ();
 
@@ -1329,16 +1388,17 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 			continue;
 		}
 
-		msg = rspamd_http_connection_copy_msg (session->client_message);
+		msg = rspamd_http_connection_copy_msg (session->client_message, &err);
 
 		if (msg == NULL) {
-			msg_err_session ("cannot copy message to send to a mirror %s: %s",
-					m->name, strerror (errno));
+			msg_err_session ("cannot copy message to send to a mirror %s: %e",
+					m->name, err);
+			if (err) {
+				g_error_free (err);
+			}
 			continue;
 		}
 
-		msg->method = HTTP_GET;
-
 		if (msg->url->len == 0) {
 			msg->url = rspamd_fstring_append (msg->url, "/check", strlen ("/check"));
 		}
@@ -1370,6 +1430,7 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 				rspamd_http_message_add_header (msg, "File", session->fname);
 			}
 
+			msg->method = HTTP_GET;
 			rspamd_http_connection_write_message_shared (bk_conn->backend_conn,
 					msg, NULL, NULL, bk_conn,
 					bk_conn->backend_sock,
@@ -1377,9 +1438,12 @@ proxy_open_mirror_connections (struct rspamd_proxy_session *session)
 		}
 		else {
 			if (session->fname) {
+				msg->flags &= ~RSPAMD_HTTP_FLAG_SHMEM;
 				rspamd_http_message_set_body (msg, session->map, session->map_len);
 			}
 
+			msg->method = HTTP_POST;
+
 			if (m->compress) {
 				proxy_request_compress (msg);
 
@@ -1547,6 +1611,7 @@ rspamd_proxy_scan_self_reply (struct rspamd_task *task)
 	case CMD_PROCESS:
 	case CMD_SKIP:
 	case CMD_CHECK_V2:
+		rspamd_task_set_finish_time (task);
 		rspamd_protocol_http_reply (msg, task, &rep);
 		rspamd_protocol_write_log_pipe (task);
 		break;
@@ -1621,7 +1686,7 @@ rspamd_proxy_self_scan (struct rspamd_proxy_session *session)
 
 	msg = session->client_message;
 	task = rspamd_task_new (session->worker, session->ctx->cfg,
-			session->pool);
+			session->pool, session->ctx->lang_det);
 	task->flags |= RSPAMD_TASK_FLAG_MIME;
 	task->sock = -1;
 
@@ -1642,6 +1707,12 @@ rspamd_proxy_self_scan (struct rspamd_proxy_session *session)
 			NULL, (event_finalizer_t )rspamd_task_free, task);
 	data = rspamd_http_message_get_body (msg, &len);
 
+	if (session->backend->settings_id) {
+		rspamd_http_message_remove_header (msg, "Settings-ID");
+		rspamd_http_message_add_header (msg, "Settings-ID",
+				session->backend->settings_id);
+	}
+
 	/* Process message */
 	if (!rspamd_protocol_handle_request (task, msg)) {
 		msg_err_task ("cannot handle request: %e", task->err);
@@ -1684,6 +1755,7 @@ proxy_send_master_message (struct rspamd_proxy_session *session)
 	struct rspamd_http_message *msg;
 	struct rspamd_http_upstream *backend = NULL;
 	const rspamd_ftok_t *host;
+	GError *err = NULL;
 	gchar hostbuf[512];
 
 	host = rspamd_http_message_find_header (session->client_message, "Host");
@@ -1706,6 +1778,8 @@ proxy_send_master_message (struct rspamd_proxy_session *session)
 		goto err;
 	}
 	else {
+		session->backend = backend;
+
 		if (backend->self_scan) {
 			return rspamd_proxy_self_scan (session);
 		}
@@ -1741,6 +1815,18 @@ retry:
 			goto retry;
 		}
 
+		msg = rspamd_http_connection_copy_msg (session->client_message, &err);
+		if (msg == NULL) {
+			msg_err_session ("cannot copy message to send it to the upstream: %e",
+					err);
+
+			if (err) {
+				g_error_free (err);
+			}
+
+			goto err; /* No fallback here */
+		}
+
 		session->master_conn->backend_conn = rspamd_http_connection_new (
 				NULL,
 				proxy_backend_master_error_handler,
@@ -1753,14 +1839,18 @@ retry:
 		session->master_conn->parser_from_ref = backend->parser_from_ref;
 		session->master_conn->parser_to_ref = backend->parser_to_ref;
 
-		msg = rspamd_http_connection_copy_msg (session->client_message);
-
 		if (backend->key) {
 			msg->peer_key = rspamd_pubkey_ref (backend->key);
 			rspamd_http_connection_set_key (session->master_conn->backend_conn,
 					session->ctx->local_key);
 		}
 
+		if (backend->settings_id != NULL) {
+			rspamd_http_message_remove_header (msg, "Settings-ID");
+			rspamd_http_message_add_header (msg, "Settings-ID",
+					backend->settings_id);
+		}
+
 		if (backend->local ||
 				rspamd_inet_address_is_local (
 						rspamd_upstream_addr (session->master_conn->up), FALSE)) {
@@ -1769,6 +1859,8 @@ retry:
 				rspamd_http_message_add_header (msg, "File", session->fname);
 			}
 
+			msg->method = HTTP_GET;
+
 			rspamd_http_connection_write_message_shared (
 					session->master_conn->backend_conn,
 					msg, NULL, NULL, session->master_conn,
@@ -1777,10 +1869,13 @@ retry:
 		}
 		else {
 			if (session->fname) {
+				msg->flags &= ~RSPAMD_HTTP_FLAG_SHMEM;
 				rspamd_http_message_set_body (msg,
 						session->map, session->map_len);
 			}
 
+			msg->method = HTTP_POST;
+
 			if (backend->compress) {
 				proxy_request_compress (msg);
 				if (session->client_milter_conn) {
@@ -1850,16 +1945,17 @@ proxy_client_finish_handler (struct rspamd_http_connection *conn,
 
 			if (msg->flags & RSPAMD_HTTP_FLAG_SPAMC) {
 				session->legacy_support = LEGACY_SUPPORT_SPAMC;
+				msg_info_session ("enabling legacy spamc mode for session");
 			}
 			else {
-				session->legacy_support = LEGACY_SUPPORT_SPAMC;
+				session->legacy_support = LEGACY_SUPPORT_RSPAMC;
+				msg_info_session ("enabling legacy rspamc mode for session");
 			}
-
-			msg_info_session ("enabling legacy rspamc mode for session");
 		}
 
 		if (msg->url->len == 0) {
-			msg->url = rspamd_fstring_append (msg->url, "/check", strlen ("/check"));
+			msg->url = rspamd_fstring_append (msg->url,
+					"/" MSG_CMD_CHECK_V2, strlen ("/" MSG_CMD_CHECK_V2));
 		}
 
 		if (!proxy_check_file (msg, session)) {
@@ -1870,6 +1966,9 @@ proxy_client_finish_handler (struct rspamd_http_connection *conn,
 				session->client_conn);
 		session->shmem_ref = rspamd_http_message_shmem_ref (session->client_message);
 		rspamd_http_message_remove_header (msg, "Content-Length");
+		rspamd_http_message_remove_header (msg, "Transfer-Encoding");
+		rspamd_http_message_remove_header (msg, "Keep-Alive");
+		rspamd_http_message_remove_header (msg, "Connection");
 		rspamd_http_message_remove_header (msg, "Key");
 
 		proxy_open_mirror_connections (session);
@@ -1889,6 +1988,9 @@ err:
 	rspamd_http_connection_steal_msg (session->client_conn);
 	rspamd_http_message_remove_header (msg, "Content-Length");
 	rspamd_http_message_remove_header (msg, "Key");
+	rspamd_http_message_remove_header (msg, "Transfer-Encoding");
+	rspamd_http_message_remove_header (msg, "Keep-Alive");
+	rspamd_http_message_remove_header (msg, "Connection");
 	rspamd_http_connection_reset (session->client_conn);
 	proxy_client_write_error (session, 404, "Backend not found");
 
@@ -1962,7 +2064,7 @@ proxy_accept_socket (gint fd, short what, void *arg)
 		return;
 	}
 
-	session = g_slice_alloc0 (sizeof (*session));
+	session = g_malloc0 (sizeof (*session));
 	REF_INIT_RETAIN (session, proxy_session_dtor);
 	session->client_sock = nfd;
 	session->client_addr = addr;
@@ -2006,6 +2108,23 @@ proxy_accept_socket (gint fd, short what, void *arg)
 				rspamd_inet_address_to_string (addr),
 				rspamd_inet_address_get_port (addr));
 
+#ifdef TCP_NODELAY
+
+	#ifndef SOL_TCP
+	#define SOL_TCP IPPROTO_TCP
+	#endif
+
+		if (rspamd_inet_address_get_af (addr) != AF_UNIX) {
+			gint sopt = 1;
+
+			if (setsockopt (nfd, SOL_TCP, TCP_NODELAY, &sopt, sizeof (sopt)) ==
+					-1) {
+				msg_warn_session ("cannot set TCP_NODELAY: %s",
+						strerror (errno));
+			}
+		}
+#endif
+
 		rspamd_milter_handle_socket (nfd, NULL,
 				session->pool,
 				ctx->ev_base,
@@ -2040,8 +2159,7 @@ start_rspamd_proxy (struct rspamd_worker *worker) {
 
 	ctx->cfg = worker->srv->cfg;
 	ctx->ev_base = rspamd_prepare_worker (worker, "rspamd_proxy",
-			proxy_accept_socket,
-			TRUE);
+			proxy_accept_socket);
 
 	ctx->resolver = dns_resolver_init (worker->srv->logger,
 			ctx->ev_base,
@@ -2065,7 +2183,8 @@ start_rspamd_proxy (struct rspamd_worker *worker) {
 
 	if (ctx->has_self_scan) {
 		/* Additional initialisation needed */
-		rspamd_worker_init_scanner (worker, ctx->ev_base, ctx->resolver);
+		rspamd_worker_init_scanner (worker, ctx->ev_base, ctx->resolver,
+				&ctx->lang_det);
 	}
 
 	if (worker->srv->cfg->enable_sessions_cache) {
@@ -2073,8 +2192,16 @@ start_rspamd_proxy (struct rspamd_worker *worker) {
 				ctx->ev_base);
 	}
 
-	rspamd_milter_init_library (ctx->spam_header, ctx->sessions_cache,
-			ctx->discard_on_reject);
+	ctx->milter_ctx.spam_header = ctx->spam_header;
+	ctx->milter_ctx.discard_on_reject = ctx->discard_on_reject;
+	ctx->milter_ctx.quarantine_on_reject = ctx->quarantine_on_reject;
+	ctx->milter_ctx.sessions_cache = ctx->sessions_cache;
+	ctx->milter_ctx.client_ca_name = ctx->client_ca_name;
+	ctx->milter_ctx.reject_message = ctx->reject_message;
+	rspamd_milter_init_library (&ctx->milter_ctx);
+
+	rspamd_lua_run_postloads (ctx->cfg->lua_state, ctx->cfg, ctx->ev_base,
+			worker);
 
 	event_base_loop (ctx->ev_base, 0);
 	rspamd_worker_block_signals ();
